/*----------------------------------------------------------------------
  File    : dti.c
  Contents: decision and regression tree induction
  Authors : Christian Borgelt
  History : 08.08.1997 file created
            08.09.1997 class checks corrected
            17.09.1997 option '-a' (aligned output) added
            11.01.1998 unknown value characters (option -u) added
            08.02.1998 adapted to changed parse functions
            09.02.1998 adapted to changed order of evaluation measures
            30.03.1998 bug in class check removed
            23.06.1998 adapted to modified attset functions
            29.09.1998 table reading simplified
            20.10.1998 output of relative class frequencies added
            09.02.1999 input from stdin, output to stdout added
            17.04.1999 simplified using the new module 'io'
            30.04.1999 log messages improved
            25.10.1999 bug in initialization of 'minval' fixed
            29.10.1999 evaluation measure FEM_WDIFF added
            18.12.2000 extended to regression trees
            02.03.2001 evaluation measure FEM_INFGBAL added
            23.07.2001 adapted to modified module scan
            11.02.2002 evaluation measures coded with names
            02.02.2002 adapted to modified list of evaluation measures
            04.02.2002 quadratic information measures added
            16.08.2003 slight changes in error message output
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <assert.h>
#ifndef AS_RDWR
#define AS_RDWR
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#ifndef TAB_RDWR
#define TAB_RDWR
#endif
#ifndef DT_GROW
#define DT_GROW
#endif
#include "io.h"
#include "dtree.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "dti"
#define DESCRIPTION "decision and regression tree induction"
#define VERSION     "version 3.10 (2003.08.16)        " \
                    "(c) 1997-2003   Christian Borgelt"

/* --- error codes --- */
#define OK            0         /* no error */
#define E_NONE        0         /* no error */
#define E_NOMEM     (-1)        /* not enough memory */
#define E_FOPEN     (-2)        /* cannot open file */
#define E_FREAD     (-3)        /* read error on file */
#define E_FWRITE    (-4)        /* write error on file */
#define E_OPTION    (-5)        /* unknown option */
#define E_OPTARG    (-6)        /* missing option argument */
#define E_ARGCNT    (-7)        /* wrong number of arguments */
#define E_STDIN     (-8)        /* double assignment of stdin */
#define E_PARSE     (-9)        /* unknown balancing mode */
#define E_BALANCE  (-10)        /* double assignment of stdin */
#define E_TARGET   (-11)        /* missing a target attribute */
#define E_MEASURE  (-12)        /* unknown selection measure */
#define E_MINCNT   (-13)        /* illegal minimal number of tuples */
#define E_UNKNOWN  (-14)        /* unknown error */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- measure information --- */
  int  code;                    /* measure code */
  char *name;                   /* name of the measure */
} MINFO;                        /* (measure information) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
/* --- measures for symbolic targets --- */
static const MINFO symtab[] = {
  { FEM_NONE,    "none"    },   /* no measure */
  { FEM_INFGAIN, "infgain" },   /* information gain */
  { FEM_INFGBAL, "infgbal" },   /* balanced information gain */
  { FEM_INFGR,   "infgr"   },   /* information gain ratio */
  { FEM_INFSGR1, "infsgr1" },   /* sym. information gain ratio 1 */
  { FEM_INFSGR2, "infsgr2" },   /* sym. information gain ratio 2 */
  { FEM_QIGAIN,  "qigain"  },   /* quadratic information gain */
  { FEM_QIGBAL,  "qigbal"  },   /* balanced quad. information gain */
  { FEM_QIGR,    "qigr"    },   /* quadratic information gain ratio */
  { FEM_QISGR1,  "qisgr1"  },   /* sym. quad. info. gain ratio 1 */
  { FEM_QISGR2,  "qisgr2"  },   /* sym. quad. info. gain ratio 2 */
  { FEM_GINI,    "gini"    },   /* gini index */
  { FEM_GINISYM, "ginisym" },   /* symmetric gini index */
  { FEM_GINIMOD, "ginimod" },   /* modified gini index */
  { FEM_RELIEF,  "relief"  },   /* relief measure */
  { FEM_WDIFF,   "wdiff"   },   /* weighted differences */
  { FEM_CHI2,    "chi2"    },   /* chi^2 measure */
  { FEM_CHI2NRM, "chi2nrm" },   /* normalized chi^2 measure */
  { FEM_WEVID,   "wevid"   },   /* weight of evidence */
  { FEM_RELEV,   "relev"   },   /* relevance */
  { FEM_BDM,     "bdm"     },   /* Bayesian-Dirichlet / K2 metric */
  { FEM_BDMOD,   "bdmod"   },   /* modified BD / K2 metric */
  { FEM_RDLREL,  "rdlrel"  },   /* red. of description length 1 */
  { FEM_RDLABS,  "rdlabs"  },   /* red. of description length 2 */
  { FEM_STOCO,   "stoco"   },   /* stochastic complexity */
  { FEM_SPCGAIN, "spcgain" },   /* specificity gain */
  { FEM_SPCGBAL, "spcgbal" },   /* balanced specificity gain */
  { FEM_SPCGR,   "spcgr"   },   /* specificity gain ratio */
  { FEM_SPCSGR1, "spcsgr1" },   /* sym. specificity gain ratio 1 */
  { FEM_SPCSGR2, "spcsgr2" },   /* sym. specificity gain ratio 2 */
  { -1,          NULL      }    /* sentinel */
};

/* --- measures for numeric targets --- */
static const MINFO numtab[] = {
  { VEM_NONE,    "none"    },   /* no measure */
  { VEM_SSE,     "sse"     },   /* sum of squared errors */
  { VEM_MSE,     "mse"     },   /* mean squared error */
  { VEM_RMSE,    "rmse"    },   /* square root of mean squared error */
  { VEM_VAR,     "var"     },   /* variance (unbiased estimator) */
  { VEM_SDEV,    "sd"      },   /* standard deviation (from variance) */
  { -1,          NULL      }    /* sentinel */
};

/* --- error messages --- */
static const char *errmsgs[] = {
  /* E_NONE      0 */  "no error\n",
  /* E_NOMEM    -1 */  "not enough memory\n",
  /* E_FOPEN    -2 */  "cannot open file %s\n",
  /* E_FREAD    -3 */  "read error on file %s\n",
  /* E_FWRITE   -4 */  "write error on file %s\n",
  /* E_OPTION   -5 */  "unknown option -%c\n",
  /* E_OPTARG   -6 */  "missing option argument\n",
  /* E_ARGCNT   -7 */  "wrong number of arguments\n",
  /* E_STDIN    -8 */  "double assignment of standard input\n",
  /* E_PARSE    -9 */  "parse error(s) on file %s\n",
  /* E_BALANCE -10 */  "unknown balancing mode %c\n",
  /* E_TARGET  -11 */  "missing target \"%s\" in file %s\n",
  /* E_MEASURE -12 */  "unknown attribute selection measure %s\n",
  /* E_MINCNT  -13 */  "illegal minimal number of tuples %g\n",
  /* E_UNKNOWN -14 */  "unknown error\n"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
const  char   *prgname = NULL;  /* program name for error messages */
static SCAN   *scan    = NULL;  /* scanner */
static ATTSET *attset  = NULL;  /* attribute set */
static TABLE  *table   = NULL;  /* table */
static DTREE  *dtree   = NULL;  /* decision/regression tree */
static FILE   *out     = NULL;  /* output file */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- print error message */
  va_list    args;              /* list of variable arguments */
  const char *msg;              /* error message */

  assert(prgname);              /* check the program name */
  if (code < E_UNKNOWN) code = E_UNKNOWN;
  if (code < 0) {               /* if to report an error, */
    msg = errmsgs[-code];       /* get the error message */
    if (!msg) msg = errmsgs[-E_UNKNOWN];
    fprintf(stderr, "\n%s: ", prgname);
    va_start(args, code);       /* get variable arguments */
    vfprintf(stderr, msg, args);/* print error message */
    va_end(args);               /* end argument evaluation */
  }
  #ifndef NDEBUG
  if (dtree)  dt_delete(dtree, 0);
  if (table)  tab_delete(table, 0);
  if (attset) as_delete(attset);   /* clean up memory */
  if (scan)   sc_delete(scan);     /* and close files */
  if (out && (out != stdout)) fclose(out);
  #endif
  #ifdef STORAGE
  showmem("at end of program"); /* check memory usage */
  #endif
  exit(code);                   /* abort the program */
}  /* error() */

/*--------------------------------------------------------------------*/

static void help (void)
{                               /* --- print help on sel. measures */
  int i;                        /* loop variable */

  fprintf(stderr, "\n");        /* terminate startup message */
  printf("\nList of attribute selection measures (option -e#)\n");
  printf("Measures for symbolic target attributes:\n");
  printf("  name       measure\n");
  for (i = 0; symtab[i].name; i++)
    printf("  %-9s  %s\n", symtab[i].name, ft_mname(symtab[i].code));
  printf("\nMeasures wdiff, bdm, bdmod, rdlrel, rdlabs "
         "take a sensitivity\n"
         "parameter (-z#, default: 0, i.e. normal sensitivity)\n");
  printf("Measures bdm and bdmod take a prior (-p#, positive number)\n"
         "or an equivalent sample size (-p#, negative number).\n");
  printf("\nMeasures for numeric target attributes:\n");
  printf("  name       measure\n");
  for (i = 0; numtab[i].name; i++)
    printf("  %-9s  %s\n", numtab[i].name, vt_mname(numtab[i].code));
  exit(0);                      /* print a list of selection measures */
}  /* help() */                 /* and abort the program */

/*--------------------------------------------------------------------*/

static int code (const MINFO *tab, const char *name)
{                               /* --- get measure code */
  for ( ; tab->name; tab++)     /* look up name in table */
    if (strcmp(tab->name, name) == 0)
      return tab->code;         /* return the measure code */
  return -1;                    /* or an error indicator */
}  /* code() */

/*--------------------------------------------------------------------*/

int main (int argc, char* argv[])
{                               /* --- main function */
  int    i, k = 0;              /* loop variables, counter */
  char   *s;                    /* to traverse options */
  char   **optarg  = NULL;      /* option argument */
  char   *fn_dom   = NULL;      /* name of domain file */
  char   *fn_hdr   = NULL;      /* name of table header file */
  char   *fn_tab   = NULL;      /* name of table file */
  char   *fn_dt    = NULL;      /* name of dec./reg. tree file */
  char   *blanks   = NULL;      /* blanks */
  char   *fldseps  = NULL;      /* field  separators */
  char   *recseps  = NULL;      /* record separators */
  char   *uvchars  = NULL;      /* unknown value characters */
  char   *trgname  = NULL;      /* target attribute name */
  char   *mname    = NULL;      /* name of att. selection measure */
  int    measure   = 3;         /* attribute selection measure */
  int    wgtd      = FEF_WGTD;  /* flag for weighted measure */
  double params[2] = { 0, 0 };  /* selection measure parameters */
  double minval    = -DBL_MAX;  /* minimal value of selection measure */
  float  mincnt    = 2.0F;      /* minimal number of cases */
  int    maxht     = INT_MAX;   /* maximal height of tree */
  int    inflags   = AS_NOXATT; /* table file read flags */
  int    dtflags   = 0;         /* flags, e.g. DF_SUBSET */
  int    balance   = 0;         /* flag for balancing class freqs. */
  int    maxlen    = 0;         /* maximal output line length */
  int    dmode     = 0;         /* description mode */
  int    attcnt    = 0;         /* number of attributes */
  int    nodecnt;               /* number of nodes in dec./reg. tree */
  double tplwgt;                /* sum of the tuple weights */
  int    trgid;                 /* id of target column */
  ATT    *att;                  /* to traverse attributes */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument is given */
    printf("usage: %s [options] domfile "
                     "[-d|-h hdrfile] tabfile dtfile\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-c#      target attribute name "
                    "(default: last attribute)\n");
    printf("-q#      balance class frequencies (weight tuples)\n");
    printf("         l: lower, b: boost, s: shift tuple weights\n");
    printf("-e#      attribute selection measure "
                    "(default: infgr/rmse)\n");
    printf("-!       print a list of available "
                    "attribute selection measures\n");
    printf("-z#      sensitivity parameter (default: %g)\n", params[0]);
    printf("         (for measures wdiff, bdm, bdmod, "
                    "rdlen1, rdlen2)\n");
    printf("-p#      prior (positive value) "
                    "or equivalent sample size (negative value)\n");
    printf("         (for measures bdm, bdmod)\n");
    printf("-i#      minimal value of the selection measure "
                    "(default: no limit)\n");
    printf("-w       do not weight measure with fraction "
                    "of known values\n");
    printf("-t#      maximal height of the tree (default: no limit)\n");
    printf("-m#      minimal number of tuples in two branches "
                    "(default: %g)\n", mincnt);
    printf("-s       try to form subsets on symbolic attributes\n");
    printf("-l#      output line length (default: no limit)\n");
    printf("-a       align values of test attributes "
                    "(default: do not align)\n");
    printf("-v       print relative frequencies (in percent)\n");
    printf("-b/f/r#  blank characters, field and record separators\n"
           "         (default: \" \\t\\r\", \" \\t\", \"\\n\")\n");
    printf("-u#      unknown value characters (default: \"?\")\n");
    printf("-n       number of tuple occurrences in last field\n");
    printf("domfile  file containing domain descriptions\n");
    printf("-d       use default header "
                    "(field names = field numbers)\n");
    printf("-h       read table header (field names) from hdrfile\n");
    printf("hdrfile  file containing table header (field names)\n");
    printf("tabfile  table file to read "
                    "(field names in first record)\n");
    printf("dtfile   file to write induced "
                    "decision/regression tree to\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case '!': help();                              break;
          case 'c': optarg    = &trgname;                break;
          case 'q': balance   = (*s) ? *s++ : 0;         break;
          case 'e': optarg    = &mname;                  break;
          case 'z': params[0] =        strtod(s, &s);    break;
          case 'p': params[1] =        strtod(s, &s);    break;
          case 'i': minval    =        strtod(s, &s);    break;
          case 'w': wgtd      = 0;                       break;
          case 'm': mincnt    = (float)strtod(s, &s);    break;
          case 's': dtflags  |= DT_SUBSET;               break;
          case 't': maxht     =   (int)strtol(s, &s, 0); break;
          case 'l': maxlen    =   (int)strtol(s, &s, 0); break;
          case 'a': dmode    |= DT_ALIGN;                break;
          case 'v': dmode    |= DT_REL;                  break;
          case 'b': optarg    = &blanks;                 break;
          case 'f': optarg    = &fldseps;                break;
          case 'r': optarg    = &recseps;                break;
          case 'u': optarg    = &uvchars;                break;
          case 'n': inflags  |= AS_WEIGHT;               break;
          case 'd': inflags  |= AS_DFLT;                 break;
          case 'h': optarg    = &fn_hdr;                 break;
          default : error(E_OPTION, *--s);               break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_dom = s;      break;
        case  1: fn_tab = s;      break;
        case  2: fn_dt  = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check option argument */
  if (k != 3) error(E_ARGCNT);  /* check number of arguments */
  if (mincnt < 0) error(E_MINCNT, mincnt);
  if (fn_hdr && (strcmp(fn_hdr, "-") == 0))
    fn_hdr = "";                /* convert "-" to "" */
  i = (!fn_dom || !*fn_dom) ? 1 : 0;
  if  (!fn_tab || !*fn_tab) i++;
  if  ( fn_hdr && !*fn_hdr) i++;/* check assignments of stdin: */
  if (i > 1) error(E_STDIN);    /* stdin must not be used twice */
  if ((balance !=  0)  && (balance != 'l')
  &&  (balance != 'b') && (balance != 's'))
    error(E_BALANCE, balance);  /* check balancing mode */
  if (fn_hdr)                   /* set the header file flag */
    inflags = AS_ATT | (inflags & ~AS_DFLT);

  /* --- read attribute set --- */
  scan = sc_create(fn_dom);     /* create a scanner */
  if (!scan) error((!fn_dom || !*fn_dom) ? E_NOMEM : E_FOPEN, fn_dom);
  attset = as_create("domains", att_delete);
  if (!attset) error(E_NOMEM);  /* create an attribute set */
  fprintf(stderr, "\nreading %s ... ", sc_fname(scan));
  if ((sc_nexter(scan)   <  0)  /* start scanning (get first token) */
  ||  (as_parse(attset, scan, AT_ALL) != 0)
  ||  (as_attcnt(attset) <= 0)  /* parse attribute set and */
  ||  !sc_eof(scan))            /* check for end of file */
    error(E_PARSE, sc_fname(scan));
  fprintf(stderr, "[%d attribute(s)] done.", as_attcnt(attset));

  /* --- determine id of target column --- */
  if (!trgname)                 /* if no target att. name is given, */
    trgid = as_attcnt(attset)-1;/* use the last att. as the target */
  else {                        /* if a target att. name is given */
    trgid = as_attid(attset, trgname);
    if (trgid < 0) error(E_TARGET, trgname, sc_fname(scan));
  }                             /* check whether the target exists */
  att = as_att(attset, trgid);  /* get the target attribute */
  sc_delete(scan); scan = NULL; /* and delete the scanner */

  /* --- translate measure --- */
  if (att_type(att) == AT_SYM){ /* if symbolic target attribute */
    if (!mname) mname = "infgr";/* set default measure name and */
    measure = code(symtab, mname); }    /* get the measure code */
  else {                        /* if numeric target attribute */
    if (!mname) mname = "rmse"; /* set default measure name and */
    measure = code(numtab, mname);      /* get the measure code */
  }                             /* then check the result */
  if (measure < 0) error(E_MEASURE, mname);
  measure |= wgtd;              /* add flag for weighted measure */
  fprintf(stderr, "\n");        /* terminate the startup message */

  /* --- read table --- */
  as_chars(attset, blanks, fldseps, recseps, uvchars);
  table = io_tabin(attset, fn_hdr, fn_tab, inflags, "table", 1);
  if (!table) error(1);         /* read the table file */
  fprintf(stderr, "reducing%s table ... ",
                  (balance) ? " and balancing" : "");
  tab_reduce(table);            /* reduce table for speed up */
  if (balance                   /* if the balance flag is set */
  && (att_type(as_att(attset, trgid)) == AT_SYM)) {
    tab_balance(table, trgid, (balance == 'l') ? -2.0F
                            : (balance == 'b') ? -1.0F : 0.0F, NULL);
  }                             /* balance the class frequencies */
  fprintf(stderr, "[%d/%g tuple(s)] done.\n",
          tab_tplcnt(table), tab_getwgt(table, 0, INT_MAX));

  /* --- grow decision/regression tree --- */
  fprintf(stderr, "growing %s tree ... ",
          (att_type(att) == AT_SYM) ? "decision" : "regression");
  dtree = dt_grow(table, trgid, measure, params, minval,
                  maxht, mincnt, dtflags);
  if (!dtree) error(E_NOMEM);   /* grow decision/regression tree */
  attcnt  = dt_attchk(dtree);   /* mark occuring attributes */
  nodecnt = dt_size(dtree);     /* get the number of nodes, */
  maxht   = dt_height(dtree);   /* the height of the tree, */
  tplwgt  = dt_total(dtree);    /* and the tuple weight sum */
  fprintf(stderr, "[%d attribute(s)/%d node(s)/%d level(s)] done.\n",
          attcnt, nodecnt, maxht);

  /* --- write decision/regression tree --- */
  if (fn_dt && *fn_dt)          /* if an output file name is given, */
    out = fopen(fn_dt, "w");    /* open the output file */
  else {                        /* if no output file name is given, */
    out = stdout; fn_dt = "<stdout>"; }     /* write to std. output */
  fprintf(stderr, "writing %s ... ", fn_dt);
  if (!out) error(E_FOPEN, fn_dt);
  if (as_desc(attset, out, AS_TITLE|AS_MARKED|AS_IVALS, maxlen) != 0)
    error(E_FWRITE, fn_dt);     /* describe attribute domains */
  fputc('\n', out);             /* leave one line empty */
  if (dt_desc(dtree, out, DT_TITLE|DT_INFO|dmode, maxlen) != 0)
    error(E_FWRITE, fn_dt);     /* describe decision/regression tree */
  if (out != stdout) {          /* if not written to standard output, */
    i = fclose(out); out = NULL;/* close the output file */
    if (i != 0) error(E_FWRITE, fn_dt);
  }                             /* check for success and */
  fprintf(stderr, "done.\n");   /* print a success message */

  /* --- clean up --- */
  #ifndef NDEBUG
  dt_delete(dtree, 0);          /* delete decision/regression tree, */
  tab_delete(table, 1);         /* table, and attribute set */
  #endif
  #ifdef STORAGE
  showmem("at end of program"); /* check memory usage */
  #endif
  return 0;                     /* return 'ok' */
}  /* main() */
