/*----------------------------------------------------------------------
  File    : dtp.c
  Contents: decision and regression tree pruning
  Authors : Christian Borgelt
  History : 04.08.1997 file created
            28.08.1997 pessimistic pruning added
            17.09.1997 option '-a' (aligned output) added
            11.01.1998 unknown value characters (option -u) added
            08.02.1998 adapted to changed parse functions
            09.02.1998 adapted to changed order of pruning methods
            23.06.1998 adapted to modified attset functions
            29.09.1998 table reading simplified
            20.10.1998 output of relative class frequencies added
            09.02.1999 input from stdin, output to stdout added
            17.04.1999 simplified using the new module 'io'
            30.04.1999 log messages improved
            17.12.2000 option -q (balance class frequencies) added
            18.12.2000 extended to regression trees
            23.07.2001 adapted to modified module scan
            19.10.2001 conf. level pruning for numeric targets added
            01.03.2002 option -k (check largest branch) added
            16.08.2003 slight changes in error message output
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <assert.h>
#ifndef SC_SCAN
#define SC_SCAN
#endif
#include "scan.h"
#ifndef AS_RDWR
#define AS_RDWR
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#ifndef TAB_RDWR
#define TAB_RDWR
#endif
#ifndef DT_PRUNE
#define DT_PRUNE
#endif
#ifndef DT_PARSE
#define DT_PARSE
#endif
#include "io.h"
#include "dtree.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "dtp"
#define DESCRIPTION "decision and regression tree pruning"
#define VERSION     "version 3.3 (2003.08.16)         " \
                    "(c) 1997-2003   Christian Borgelt"

/* --- error codes --- */
#define OK            0         /* no error */
#define E_NONE        0         /* no error */
#define E_NOMEM     (-1)        /* not enough memory */
#define E_FOPEN     (-2)        /* cannot open file */
#define E_FREAD     (-3)        /* read error on file */
#define E_FWRITE    (-4)        /* write error on file */
#define E_OPTION    (-5)        /* unknown option */
#define E_OPTARG    (-6)        /* missing option argument */
#define E_ARGCNT    (-7)        /* wrong number of arguments */
#define E_STDIN     (-8)        /* double assignment of stdin */
#define E_PARSE     (-9)        /* parse error */
#define E_BALANCE  (-10)        /* double assignment of stdin */
#define E_METHOD   (-11)        /* unknown pruning method */
#define E_UNKNOWN  (-12)        /* unknown error */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- method information --- */
  int  code;                    /* pruning method code */
  char *name;                   /* name of the method */
} MINFO;                        /* (method information) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
/* --- pruning methods --- */
static const MINFO pmtab[] = {
  { PM_NONE, "none" },          /* no pruning */
  { PM_PESS, "pess" },          /* pessimistic pruning */
  { PM_CLVL, "clvl" },          /* confidence level pruning */
  { -1,      NULL   }           /* sentinel */
};

/* --- error messages --- */
const char *errmsgs[] = {       /* error messages */
  /* E_NONE      0 */  "no error\n",
  /* E_NOMEM    -1 */  "not enough memory\n",
  /* E_FOPEN    -2 */  "cannot open file %s\n",
  /* E_FREAD    -3 */  "read error on file %s\n",
  /* E_FWRITE   -4 */  "write error on file %s\n",
  /* E_OPTION   -5 */  "unknown option -%c\n",
  /* E_OPTARG   -6 */  "missing option argument\n",
  /* E_ARGCNT   -7 */  "wrong number of arguments\n",
  /* E_STDIN    -8 */  "double assignment of standard input\n",
  /* E_PARSE    -9 */  "parse error(s) on file %s\n",
  /* E_BALANCE -10 */  "unknown balancing mode %c\n",
  /* E_METHOD  -11 */  "unknown pruning method %s\n",
  /* E_UNKNOWN -12 */  "unknown error\n"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
const  char   *prgname = NULL;  /* program name for error messages */
static SCAN   *scan    = NULL;  /* scanner */
static DTREE  *dtree   = NULL;  /* decision/regression tree */
static ATTSET *attset  = NULL;  /* attribute set */
static TABLE  *table   = NULL;  /* table */
static FILE   *out     = NULL;  /* output file */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- print error message */
  va_list    args;              /* list of variable arguments */
  const char *msg;              /* error message */

  assert(prgname);              /* check the program name */
  if (code < E_UNKNOWN) code = E_UNKNOWN;
  if (code < 0) {               /* if to report an error, */
    msg = errmsgs[-code];       /* get the error message */
    if (!msg) msg = errmsgs[-E_UNKNOWN];
    fprintf(stderr, "\n%s: ", prgname);
    va_start(args, code);       /* get variable arguments */
    vfprintf(stderr, msg, args);/* print error message */
    va_end(args);               /* end argument evaluation */
  }
  #ifndef NDEBUG
  if (table)  tab_delete(table, 0);
  if (dtree)  dt_delete(dtree, 0);
  if (attset) as_delete(attset);   /* clean up memory */
  if (scan)   sc_delete(scan);     /* and close files */
  if (out && (out != stdout)) fclose(out);
  #endif
  #ifdef STORAGE
  showmem("at end of program"); /* check memory usage */
  #endif
  exit(code);                   /* abort the program */
}  /* error() */

/*--------------------------------------------------------------------*/

static int code (const MINFO *tab, const char *name)
{                               /* --- get pruning method code */
  for ( ; tab->name; tab++)     /* look up name in table */
    if (strcmp(tab->name, name) == 0)
      return tab->code;         /* return the method code */
  return -1;                    /* or an error indicator */
}  /* code() */

/*--------------------------------------------------------------------*/

int main (int argc, char* argv[])
{                               /* --- main function */
  int    i, k = 0;              /* loop variables, counter */
  char   *s;                    /* to traverse options */
  char   **optarg = NULL;       /* option argument */
  char   *fn_dt   = NULL;       /* name of dec./reg. tree file */
  char   *fn_pdt  = NULL;       /* name of output file */
  char   *fn_hdr  = NULL;       /* name of table header file */
  char   *fn_tab  = NULL;       /* name of table file */
  char   *blanks  = NULL;       /* blanks */
  char   *fldseps = NULL;       /* field  separators */
  char   *recseps = NULL;       /* record separators */
  char   *uvchars = NULL;       /* unknown value characters */
  char   *mname   = NULL;       /* name of pruning method */
  int    method   = 2;          /* pruning method */
  float  param    = 0.5F;       /* pruning parameter */
  int    maxht    = INT_MAX;    /* maximal height of tree */
  int    chklb    = 0;          /* whether to check largest branch */
  int    flags    = AS_NOXATT;  /* table file read flags */
  int    balance  = 0;          /* flag for balancing class freqs. */
  int    maxlen   = 0;          /* maximal output line length */
  int    dmode    = 0;          /* description mode */
  double tplwgt;                /* weight of tuples */
  int    attcnt;                /* number of attributes in dec. tree */
  int    nodecnt;               /* number of nodes in dec./reg. tree */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument is given */
    printf("usage: %s [options] dtfile pdtfile "
                     "[[-d|-h hdrfile] tabfile]\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-m#      pruning method (default: clvl)\n");
    printf("         none: no pruning\n");
    printf("         pess: pessimistic pruning\n");
    printf("         clvl: confidence level pruning\n");
    printf("-p#      pruning parameter (default: %g)\n", param);
    printf("-q#      balance class frequencies (weight tuples)\n");
    printf("         l: lower, b: boost, s: shift tuple weights\n");
    printf("-t#      maximal height of the tree (default: no limit)\n");
    printf("-k       check largest branch "
                    "(can be very time consuming)\n");
    printf("-l#      output line length (default: no limit)\n");
    printf("-a       align values of test attributes "
                    "(default: do not align)\n");
    printf("-v       print relative frequencies (in percent)\n");
    printf("-b/f/r#  blank characters, field and record separators\n"
           "         (default: \" \\t\\r\", \" \\t\", \"\\n\")\n");
    printf("-u#      unknown value characters (default: \"?\")\n");
    printf("-n       number of tuple occurrences in last field\n");
    printf("dtfile   file containing decision/regression "
                    "tree description\n");
    printf("pdtfile  file to write pruned "
                    "decision/regression tree to\n");
    printf("-d       use default header "
                    "(field names = field numbers)\n");
    printf("-h       read table header (field names) from hdrfile\n");
    printf("hdrfile  file containing table header (field names)\n");
    printf("tabfile  table file to read "
                    "(field names in first record)\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case 'm': optarg  = &mname;                  break;
          case 'p': param   = (float)strtod(s, &s);    break;
          case 'q': balance = (*s) ? *s++ : 0;         break;
          case 't': maxht   =   (int)strtol(s, &s, 0); break;
          case 'k': chklb   = 1;                       break;
          case 'a': dmode  |= DT_ALIGN;                break;
          case 'v': dmode  |= DT_REL;                  break;
          case 'l': maxlen  =   (int)strtol(s, &s, 0); break;
          case 'b': optarg  = &blanks;                 break;
          case 'f': optarg  = &fldseps;                break;
          case 'r': optarg  = &recseps;                break;
          case 'u': optarg  = &uvchars;                break;
          case 'n': flags  |= AS_WEIGHT;               break;
          case 'd': flags  |= AS_DFLT;                 break;
          case 'h': optarg  = &fn_hdr;                 break;
          default : error(E_OPTION, *--s);             break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_dt  = s;      break;
        case  1: fn_pdt = s;      break;
        case  2: fn_tab = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check option argument */
  if ((k < 2) || (k > 3)) error(E_ARGCNT);
  if (fn_hdr && (strcmp(fn_hdr, "-") == 0))
    fn_hdr = "";                /* convert "-" to "" */
  i = (!fn_dt || !*fn_dt) ? 1 : 0;
  if  (fn_tab && !*fn_tab) i++;
  if  (fn_hdr && !*fn_hdr) i++; /* check assignments of stdin: */
  if (i > 1) error(E_STDIN);    /* stdin must not be used twice */
  if ((balance !=  0)  && (balance != 'l')
  &&  (balance != 'b') && (balance != 's'))
    error(E_BALANCE, balance);  /* check balancing mode */
  if (fn_hdr)                   /* set the header file flag */
    flags = AS_ATT | (flags & ~AS_DFLT);

  /* --- translate pruning method --- */
  if (!mname) mname = "clvl";   /* set default method name */
  method = code(pmtab, mname);  /* and get the method code */
  if (method < 0) error(E_METHOD, method);

  /* --- read decision tree --- */
  scan = sc_create(fn_dt);      /* create a scanner */
  if (!scan) error((!fn_dt || !*fn_dt) ? E_NOMEM : E_FOPEN, fn_dt);
  attset = as_create("domains", att_delete);
  if (!attset) error(E_NOMEM);  /* create an attribute set */
  fprintf(stderr, "\nreading %s ... ", sc_fname(scan));
  if ((sc_nexter(scan)   <  0)  /* start scanning (get first token) */
  ||  (as_parse(attset, scan, AT_ALL) != 0)
  ||  (as_attcnt(attset) <= 0)) /* parse attribute set */
    error(E_PARSE, sc_fname(scan));
  dtree = dt_parse(attset,scan);/* parse the decision tree */
  if (!dtree || !sc_eof(scan)) error(E_PARSE, fn_dt);
  sc_delete(scan); scan = NULL; /* delete the scanner */
  fprintf(stderr, "[%d attribute(s)/%d node(s)/%d level(s)] done.\n",
          as_attcnt(attset), dt_size(dtree), dt_height(dtree));

  /* --- read table --- */
  as_chars(attset, blanks, fldseps, recseps, uvchars);
  if (k > 2) {                  /* if a table file name is given */
    table  = io_tabin(attset, fn_hdr, fn_tab, flags, "table", 1);
    if (!table) error(1);       /* read the table file */
    fprintf(stderr, "reducing table ... ");
    tab_reduce(table);          /* reduce the table for speed up */
    if (balance                 /* if the balance flag is set */
    &&  (att_type(as_att(attset, dt_trgid(dtree))) == AT_SYM)) {
      tab_balance(table, dt_trgid(dtree),
                  (balance == 'l') ? -2.0F
                : (balance == 'b') ? -1.0F : 0.0F, NULL);
    }                           /* balance the class frequencies */
    fprintf(stderr,"[%d/%g tuple(s)] done.\n",
            tab_tplcnt(table), tab_getwgt(table, 0, INT_MAX));
  }                             /* print a log message */

  /* --- prune decision/regression tree --- */
  fprintf(stderr, "pruning tree ... ");     /* print a log message */
  if (dt_prune(dtree, method, param, maxht, chklb, table) != 0)
    error(E_NOMEM);             /* prune the decision/regression tree */
  attcnt  = dt_attchk(dtree);   /* mark the occurring attributes */
  nodecnt = dt_size(dtree);     /* get the number of nodes, */
  maxht   = dt_height(dtree);   /* the height of the tree, */
  tplwgt  = dt_total(dtree);    /* and the tuple weight sum */
  fprintf(stderr, "[%d attribute(s)/%d node(s)/%d level(s)] done.\n",
          attcnt, nodecnt, maxht);

  /* --- write decision/regression tree --- */
  if (fn_pdt && *fn_pdt)        /* if an output file name is given, */
    out = fopen(fn_pdt, "w");   /* open output file for writing */
  else {                        /* if no output file name is given, */
    out = stdout; fn_pdt = "<stdout>"; }    /* write to std. output */
  fprintf(stderr, "writing %s ... ", fn_pdt);
  if (!out) error(E_FOPEN, fn_pdt);
  if (as_desc(attset, out, AS_TITLE|AS_MARKED|AS_IVALS, maxlen) != 0)
    error(E_FWRITE, fn_pdt);    /* describe the attribute domains */
  fputc('\n', out);             /* leave one line empty */
  if (dt_desc(dtree, out, DT_TITLE|dmode, maxlen) != 0)
    error(E_FWRITE, fn_pdt);    /* describe the dec./reg. tree */
  if (maxlen <= 0) maxlen = 72; /* determine maximal line length */
  fputs("\n/*", out);           /* append additional information */
  for (k = maxlen -2; --k >= 0; ) fputc('-', out);
  fprintf(out, "\n  number of attributes: %d",   attcnt);
  fprintf(out, "\n  tree height         : %d",   maxht);
  fprintf(out, "\n  number of nodes     : %d",   nodecnt);
  fprintf(out, "\n  number of tuples    : %g\n", tplwgt);
  for (k = maxlen -2; --k >= 0; ) fputc('-', out);
  fputs("*/\n", out);           /* terminate additional information */
  if (out != stdout) {          /* if not written to standard output, */
    i = fclose(out); out = NULL;/* close the output file */
    if (i != 0) error(E_FWRITE, fn_pdt);
  }                             /* check for success and */
  fprintf(stderr, "done.\n");   /* print a success message */

  /* --- clean up --- */
  #ifndef NDEBUG
  if (table) tab_delete(table, 0);
  dt_delete(dtree, 1);          /* delete table, dec./reg. tree, */
  #endif                        /* and attribute set */
  #ifdef STORAGE
  showmem("at end of program"); /* check memory usage */
  #endif
  return 0;                     /* return 'ok' */
}  /* main() */
