/*----------------------------------------------------------------------
  File    : dtree1.c
  Contents: decision and regression tree management, base functions
  Author  : Christian Borgelt
  History : 12.06.1997 file created (as dtree.c)
            08.08.1997 first version completed
            25.08.1997 multiple child references made possible
            30.08.1997 treatment of new values during pruning enabled
            25.08.1997 multiple child references replaced by links
            02.02.1998 links between children simplified
            03.02.1998 adapted to changed interface of module 'table'
            27.05.1998 treatment of val == cut in _exec changed
            31.05.1998 rule extraction adapted to interface changes
            15.10.1998 major revision completed
            20.10.1998 output of relative class frequencies added
            15.12.1998 minor improvements, assertions added
            13.11.1999 bug in parse error reporting removed
            02.12.2000 growing and pruning functions moved to dtree2.c
            16.12.2000 decision tree parsing simplified
            17.12.2000 regression tree functionality added
            22.07.2001 all global variables removed
            27.08.2001 bug in _aggr concerning node->cls fixed
            12.10.2001 parse error recovery corrected/improved
            11.01.2002 number of attributes stored in DTREE structure
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "vecops.h"
#ifndef DT_PARSE
#include "scan.h"
#endif
#include "dtree.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define EPSILON     1e-6        /* to handle roundoff errors */

/* --- sizes --- */
#define BS_PATH     32          /* block size for the path vector */

#ifdef DT_PARSE
/* --- error codes --- */
#define E_CHREXP    (-16)       /* character expected */
#define E_STREXP    (-17)       /* string expected */
#define E_ATTEXP    (-18)       /* attribute expected */
#define E_VALEXP    (-19)       /* attribute value expected */
#define E_NUMEXP    (-20)       /* number expected */
#define E_ILLNUM    (-21)       /* illegal number */
#define E_UNKATT    (-22)       /* unknown attribute */
#define E_UNKVAL    (-23)       /* unknown attribute value */
#define E_DUPVAL    (-24)       /* duplicate attribute value */
#endif  /* #ifdef DT_PARSE */

/* --- functions --- */
#define islink(d,n) (((d)->link >= (n)->vec) && \
                     ((d)->link <  (n)->vec +(n)->size))
#ifdef DT_PARSE
#define ERROR(c)    _paerr(scan, c,        -1, NULL)
#define ERR_CHR(c)  _paerr(scan, E_CHREXP,  c, NULL)
#define ERR_STR(s)  _paerr(scan, E_STREXP, -1, s)
#define GET_TOK()   if (sc_next(scan) < 0) \
                      return sc_error(scan, sc_token(scan))
#define GET_CHR(c)  if (sc_token(scan) != (c)) return ERR_CHR(c); \
                    else GET_TOK();
#define RECOVER()   if (sc_recover(scan, ';', '{', '}', 0) == T_EOF) \
                      return 1;
#endif  /* #ifdef DT_PARSE */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- decision tree output info --- */
  FILE    *file;                /* output file */
  ATTSET  *attset;              /* attribute set of decision tree */
  int     type;                 /* target attribute type */
  int     ind;                  /* indentation (in characters) */
  int     pos;                  /* position in output line */
  int     max;                  /* maximal line length */
  int     mode;                 /* description mode, e.g. DT_ALIGN */
  char    buf[4*AS_MAXLEN+68];  /* output buffer */
} DTOUT;                        /* (decision tree output info) */

#ifdef DT_RULES
typedef struct {                /* --- rule extraction info --- */
  RULESET *ruleset;             /* rule set to add to */
  ATTSET  *attset;              /* attribute set */
  int     type;                 /* target type */
  int     subset[1];            /* subset of attribute values */
} DTRULE;                       /* (rule extraction info) */
#endif

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#ifdef DT_PARSE
#ifdef GERMAN                     /* deutsche Texte */
static const char *errmsgs[] = {  /* Fehlermeldungen */
  /* E_CHREXP  -16 */  "\"%c\" erwartet statt %s",
  /* E_STREXP  -17 */  "\"%s\" erwartet statt %s",
  /* E_ATTEXP  -18 */  "Attribut erwartet statt %s",
  /* E_VALEXP  -19 */  "Attributwert erwartet statt %s",
  /* E_NUMEXP  -20 */  "Zahl erwartet statt %s",
  /* E_ILLNUM  -21 */  "ungltige Zahl %s",
  /* E_UNKATT  -22 */  "unbekanntes Attribut %s",
  /* E_UNKVAL  -23 */  "unbekannter Attributwert %s",
  /* E_DUPVAL  -24 */  "doppelter Attributwert %s",
};
#else                             /* English texts */
static const char *errmsgs[] = {  /* error messages */
  /* E_CHREXP  -16 */  "\"%c\" expected instead of %s",
  /* E_STREXP  -17 */  "\"%s\" expected instead of %s",
  /* E_ATTEXP  -18 */  "attribute expected instead of %s",
  /* E_VALEXP  -19 */  "attribute value expected instead of %s",
  /* E_NUMEXP  -20 */  "number expected instead of %s",
  /* E_ILLNUM  -21 */  "illegal number %s",
  /* E_UNKATT  -22 */  "unknown attribute %s",
  /* E_UNKVAL  -23 */  "unknown attribute value %s",
  /* E_DUPVAL  -24 */  "duplicate attribute value %s",
};
#endif
#define MSGCNT  (int)(sizeof(errmsgs)/sizeof(const char*))

#endif  /* #ifdef DT_PARSE */
/*----------------------------------------------------------------------
  Recursive Parts of Main Functions
----------------------------------------------------------------------*/

static void _delete (DTNODE *node)
{                               /* --- recursively delete a subtree */
  int    i;                     /* loop variable */
  DTDATA *data;                 /* to traverse the data vector */

  assert(node);                 /* check the function argument */
  if (!(node->flags & DT_LEAF)){/* if the node is no leaf */
    for (data = node->vec +(i = node->size); --i >= 0; )
      if ((--data)->child && !islink(data, node))
        _delete(data->child);   /* recursively delete all subtrees, */
  }                             /* avoiding empty subtrees and links */
  free(node);                   /* delete the node itself */
}  /* _delete() */

/*--------------------------------------------------------------------*/

static void _aggr (DTNODE *node, int n, double *buf)
{                               /* --- aggregate node data */
  int    i;                     /* loop variable */
  DTDATA *data, *dmax;          /* to traverse data vector */
  double *frq,  *max;           /* to traverse the class frequencies */
  double sum;                   /* sum of the class frequencies */

  assert(node && buf);          /* check the function arguments */

  /* --- leaf node --- */
  if (node->flags & DT_LEAF) {  /* if the node is a leaf */
    if (n < 0) {                /* if the target att. is numeric, */
      buf[0] += node->frq;      /* compute sums over f, f*v, f*v*v */
      buf[1] += sum = node->frq *node->trg.f;
      buf[2] += node->err + sum *node->trg.f; }
    else {                      /* if the target att. is symbolic, */
      assert(node->size == n);  /* check the data vector size */
      dmax = data = node->vec;  /* get pointers to the data vector */
      *buf += sum = data->frq;  /* and start summing the frequencies */
      while (--n > 0) {         /* traverse the data vector */
        sum += (++data)->frq;   /* sum the class frequencies and */
        *++buf += data->frq;    /* aggregate them for classes */
        if (data->frq > dmax->frq) dmax = data;
      }                         /* update the maximal frequency */
      node->frq = (float)sum;   /* note the node frequency and */
      node->err = sum -dmax->frq;      /* the number of errors */
      node->trg.i = (int)(dmax -node->vec);
    }                           /* determine the most freq. class */
    return;                     /* leaves terminate the recursion, */
  }                             /* so abort the function */

  /* --- test node --- */
  for (frq = buf +abs(n) +(i = abs(n)); --i >= 0; )
    *--frq = 0;                 /* clear the next buffer section */
  for (data = node->vec +(i = node->size); --i >= 0; ) {
    if ((--data)->child && !islink(data, node))
      _aggr(data->child, n, frq);
  }                             /* aggregate data recursively */
  if (n < 0) {                  /* if the target att. is numeric, */
    buf[0] += frq[0];           /* aggregate sums over f, f*v, f*v*v */
    buf[1] += frq[1];           /* for the next higher level and */
    buf[2] += frq[2];           /* compute the node information */
    sum = (frq[0] > 0) ? frq[1] /frq[0] : 0;
    node->frq = (float)frq[0];  /* note the node frequency and the */
    node->err = frq[2] -sum *frq[1];      /* sum of squared errors */
    node->trg.f = (float)sum; } /* note the mean value for the pred. */
  else {                        /* if the target att. is symbolic */
    max = frq; *buf += sum = *frq;
    while (--n > 0) {           /* traverse the class frequencies */
      sum    += *++frq;         /* sum the class frequencies and */
      *++buf += *frq;           /* aggregate them for classes */
      if (*frq > *max) max = frq;
    }                           /* update the maximal frequency */
    node->frq = (float)sum;     /* note the node frequency */
    node->err = sum -*max;      /* and the number of errors */
    node->trg.i = (int)(max-buf-1);
  }                             /* determine the most freq. class */
}  /* _aggr() */

/*--------------------------------------------------------------------*/

static void _mark (ATTSET *attset, DTNODE *node)
{                               /* --- mark occurring attributes */
  int    i;                     /* loop variable */
  DTDATA *data;                 /* to traverse data vector */

  assert(attset && node);       /* check the function arguments */
  if (node->flags & DT_LEAF)    /* if the node is a leaf, abort, */
    return;                     /* otherwise mark the test attribute */
  att_setmark(as_att(attset, node->attid), 1);
  for (data = node->vec +(i = node->size); --i >= 0; )
    if ((--data)->child && !islink(data, node))
      _mark(attset,data->child);/* recursively mark the attributes, */
}  /* _mark() */                /* avoiding empty subtrees and links */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

DTREE* dt_create (ATTSET *attset, int trgid)
{                               /* --- create a decision tree */
  int   k;                      /* size of the buffer */
  int   colcnt;                 /* number of table columns */
  int   clscnt;                 /* number of classes (sym. target) */
  ATT   *att;                   /* target attribute */
  int   type;                   /* type of the target attribute */
  DTREE *dt;                    /* created decision tree */

  assert(attset);               /* check the function argument */
  colcnt = as_attcnt(attset);   /* get the number of columns */
  if ((trgid < 0) || (trgid >= colcnt))
    trgid = colcnt -1;          /* check and adapt the target id, */
  att  = as_att(attset, trgid); /* then get the target attribute */
  type = att_type(att);         /* and its type */
  if (type != AT_SYM) {         /* if the target att. is numeric, */
    clscnt = 0; k = 3; }        /* there are no classes */
  else {                        /* if the target att. is symbolic */
    clscnt = k = att_valcnt(att);
    assert(clscnt >= 1);        /* get and check */
  }                             /* the number of classes */
  dt = (DTREE*)malloc(sizeof(DTREE));
  if (!dt) return NULL;         /* create a dec./reg. tree body */
  dt->buf = (double*)malloc(k *sizeof(double));
  if (!dt->buf) { free(dt); return NULL; }
                                /* create a buffer vector */
  dt->attset = attset;          /* initialize the fields: */
  dt->trgid  = trgid;           /* note the attribute set, */
  dt->type   = type;            /* the target's id and type, */
  dt->clscnt = clscnt;          /* and number of classes */
  dt->attcnt = 1;               /* clear the number of attributes, */
  dt->size   = dt->height = 0;  /* the number of nodes, the height, */
  dt->total  = 0.0F;            /* the total of frequencies, */
  dt->root   = dt->curr = NULL; /* and the node pointers */
  dt->newp   = &(dt->root);     /* (no root node is created) */
  dt->path   = NULL;            /* do not allocate a path */
  dt->plen   = dt->pvsz = 0;    /* until it is needed */
  return dt;                    /* return created decision tree */
}  /* dt_create() */

/*--------------------------------------------------------------------*/

void dt_delete (DTREE *dt, int delas)
{                               /* --- delete a decision tree */
  assert(dt);                   /* check argument */
  if (dt->root) _delete(dt->root);     /* recursively delete nodes */
  if (dt->path) free(dt->path);        /* delete the path vector */
  if (dt->buf)  free(dt->buf);         /* and the buffer vector */
  if (delas)    as_delete(dt->attset); /* delete attribute set */
  free(dt);                     /* delete decision tree body */
}  /* dt_delete() */

/*--------------------------------------------------------------------*/

void dt_up (DTREE *dt, int root)
{                               /* --- go up in decision tree */
  assert(dt);                   /* check argument */
  if (root || (dt->plen <= 0)) {
    dt->curr = dt->root;        /* if to reset to root, */
    dt->plen = 0; }             /* make the root the current node */
  else                          /* if to go up one level */
    dt->curr = dt->path[--dt->plen];
}  /* dt_up() */                /* get preceding node in path */

/*--------------------------------------------------------------------*/

int dt_down (DTREE *dt, int index, int check)
{                               /* --- go down in decision tree */
  DTNODE **vec;                 /* new path vector */
  int    vsz;                   /* new size of path vector */

  assert(dt);                   /* check arguments */
  if (!dt->curr) return -2;     /* and current node */
  assert((index >= 0) && (index < dt->curr->size));
  if (check)                    /* if only to check for child node */
    return (dt->curr->vec[index].child) ? 0 : 1;
  if (dt->plen >= dt->pvsz) {   /* if the path vector is full, */
    vsz = dt->pvsz +BS_PATH;    /* enlarge the path vector */
    vec = (DTNODE**)realloc(dt->path, vsz *sizeof(DTNODE*));
    if (!vec) return -1;        /* allocate a new path vector */
    dt->path = vec; dt->pvsz = vsz;
  }                             /* set the new path vector */
  dt->path[dt->plen++] = dt->curr;  /* append node to path */
  dt->newp = &(dt->curr->vec[index].child);
  dt->curr = *(dt->newp);       /* get the index-th child */
  return (dt->curr) ? 0 : 1;    /* return 'ok' or 'node missing' */
}  /* dt_down() */

/*--------------------------------------------------------------------*/

int dt_node (DTREE *dt, int attid, float cut)
{                               /* --- create a new tree node */
  ATT    *att;                  /* test attribute */
  DTNODE *node;                 /* new tree node */
  DTDATA *data, tmp;            /* to traverse data vector */
  int    size;                  /* size of node data vector */

  assert(dt && (attid < as_attcnt(dt->attset)));
  if (dt->curr) return -2;      /* check the current node and */
  if (attid < 0) attid = dt->trgid;   /* get the attribute id */
  att  = as_att(dt->attset, attid);
  size = (att_type(att) == AT_SYM) ? att_valcnt(att) : 2;
  node = (DTNODE*)malloc(sizeof(DTNODE) +(size-1) *sizeof(DTDATA));
  if (!node) return -1;         /* create a new node and */
  node->size  = size;           /* initialize its fields */
  node->flags = (attid == dt->trgid) ? DT_LEAF : 0;
  node->attid = attid;          /* set attribute id, cut value, */
  node->cut   = cut;            /* node freq., and number of errors */
  node->frq   = 0.0F;
  node->err   = 0.0;
  if (dt->type == AT_SYM) node->trg.i = 0;
  else                    node->trg.f = 0.0F;
  if (attid != dt->trgid) tmp.child = NULL;
  else { dt->total = -1;  tmp.frq   = 0.0F; }
  data = node->vec +size;       /* clear the data vector */
  while (--size >= 0) *--data = tmp;
  *(dt->newp) = dt->curr = node;/* set pointers to the new node */
  dt->size++;                   /* increase the node counter */
  if (dt->plen >= dt->height)   /* if path is longer than height, */
    dt->height = dt->plen+1;    /* update the tree height */
  dt->attcnt = -1;              /* invalidate the number of atts. */
  return 0;                     /* return 'ok' */
}  /* dt_node() */

/*--------------------------------------------------------------------*/

int dt_link (DTREE *dt, int src, int dst)
{                               /* --- link a child to another child */
  DTDATA *data;                 /* data vector element */

  assert(dt);                   /* check the function arguments */
  if (!dt->curr) return -2;     /* and the current node */
  assert((src >= 0) && (src < dt->curr->size)
      && (dst >= 0) && (dst < dt->curr->size));
  data = dt->curr->vec +src;    /* get the data vector element */
  if (data->child) return -1;   /* if a child/link exists, abort */
  data->link = dt->curr->vec +dst;  /* link the child nodes */
  dt->curr->flags |= DT_LINK;       /* and set the link flag */
  return 0;                     /* return 'ok' */
}  /* dt_link() */

/*--------------------------------------------------------------------*/

int dt_dest (DTREE *dt, int index)
{                               /* --- get destination of a link */
  DTDATA *data;                 /* data vector element */

  assert(dt);                   /* check the function argument */
  if (!dt->curr) return -2;     /* and the current node */
  assert((index >= 0) && (index < dt->curr->size));
  data = dt->curr->vec +index;  /* get the data vector element */
  while (islink(data, dt->curr))/* while the field is a link, */
    data = data->link;          /* follow it to the next field */
  return (int)(data -dt->curr->vec);
}  /* dt_dest() */              /* return the destination index */

/*--------------------------------------------------------------------*/

float dt_total (DTREE *dt)
{                               /* --- get total of frequencies */
  int    i, k;                  /* loop variable, buffer section size */
  double *buf;                  /* aggr. buffer (for all levels) */

  assert(dt);                   /* check the function argument */
  if (dt->total >= 0)           /* if the frequency total is valid, */
    return dt->total;           /* return it directly */
  if (!dt->root)                /* if there is no tree, */
    return dt->total = 0.0F;    /* return 0 (no tuples) */
  k   = (dt->type == AT_SYM) ? dt->clscnt : -3;
  buf = (double*)malloc(dt->height *abs(k) *sizeof(double));
  if (!buf) return -1;          /* allocate an aggregation buffer */
  for (buf += (i = abs(k)); --i >= 0; )
    *--buf = 0;                 /* clear the first buffer section, */
  _aggr(dt->root, k, buf);      /* aggregate the class frequencies, */
  free(buf);                    /* and delete the aggregation buffer */
  return dt->total = dt->root->frq;
}  /* dt_total() */             /* return the total frequency */

/*--------------------------------------------------------------------*/

int dt_attchk (DTREE *dt)
{                               /* --- check occurring attributes */
  int i, n = 0;                 /* loop variable, counter */

  assert(dt);                   /* check the function argument */
  for (i = as_attcnt(dt->attset); --i >= 0; )    /* unmark all */
    att_setmark(as_att(dt->attset, i), -1);      /* attributes */
  att_setmark(as_att(dt->attset, dt->trgid), 0); /* mark target */
  if (dt->root) _mark(dt->attset, dt->root);  /* and test atts. */
  for (i = as_attcnt(dt->attset); --i >= 0; )
    if (att_getmark(as_att(dt->attset, i)) >= 0)
      n++;                      /* count occurring attributes */
  return dt->attcnt = n;        /* and return their number */
}  /* dt_attchk() */

/*--------------------------------------------------------------------*/

int dt_attcnt (DTREE *dt)
{                               /* --- get the number of occ. atts. */
  if (dt->attcnt > 0)           /* if the number of atts. is valid, */
    return dt->attcnt;          /* return it directly */
  return dt->attcnt = dt_attchk(dt);
}  /* dt_attcnt() */            /* otherwise determine it */

/*----------------------------------------------------------------------
  Execution Functions
----------------------------------------------------------------------*/

static void _exec (DTREE *dt, DTNODE *node, double weight)
{                               /* --- recursively classify a tuple */
  int    i;                     /* loop variable, index */
  int    type;                  /* type of test attribute */
  DTDATA *data;                 /* to traverse data vector */
  double *frq;                  /* to traverse the class frequencies */
  float  f;                     /* value of continuous attribute */

  assert(dt && node && (weight >= 0));   /* check the function args. */

  /* --- leaf node --- */
  if (node->flags & DT_LEAF) {  /* if the node is a leaf */
    frq = dt->buf;              /* get the buffer vector of the tree */
    if (dt->type != AT_SYM) {   /* if the target att. is numeric, */
      frq[0] += weight *node->frq;   /* aggregate the data */
      frq[1] += weight *node->err;   /* for the prediction */
      frq[2] += weight *node->trg.f; }
    else {                      /* if the target att. is symbolic, */
      assert(node->size == dt->clscnt);
      frq += node->size;        /* traverse the data vector */
      for (data = node->vec +(i = node->size); --i >= 0; )
        *--frq += (--data)->frq *weight;
    }                           /* add the class frequencies */
    return;                     /* to the result distribution, */
  }                             /* then terminate the function */

  /* --- test node --- */
  type = att_type(as_att(dt->attset, node->attid));
  if (type != AT_FLT) {         /* if symbolic or integer attribute */
    if (dt->tuple) i = tpl_colval(dt->tuple, node->attid)->i;
    else           i = att_inst(as_att(dt->attset, node->attid))->i;
    if (type == AT_SYM) {       /* if symbolic attribute */
      if (i >= node->size) i = -1; }
    else {                      /* if integer attribute */
      i = ((i <= UV_INT) || (i == node->cut))
        ? -1 : ((i <= node->cut) ? 0 : 1);
    } }                         /* determine the vector index */
  else {                        /* if real/float attribute */
    if (dt->tuple) f = tpl_colval(dt->tuple, node->attid)->f;
    else           f = att_inst(as_att(dt->attset, node->attid))->f;
    i = ((f <= UV_FLT) || (f == node->cut))
      ? -1 : ((f <= node->cut) ? 0 : 1);
  }                             /* determine the vector index */

  /* --- attribute value is known --- */
  if (i >= 0) {                 /* if an index has been determined, */
    data = node->vec +i;        /* get corresponding vector element */
    while (islink(data, node))  /* follow subset links */
      data = data->link;        /* to the true child node */
    if (data->child)            /* if the child node exists, */
      _exec(dt, data->child, weight); /* execute the subtree */
    else if (dt->type == AT_SYM)/* if the target att. is symbolic */
      dt->buf[node->trg.i] += weight *EPSILON;
    else {                      /* if the target att. is numeric */
      dt->buf[0] += weight *EPSILON;
      dt->buf[2] += weight *EPSILON *node->trg.f;
    }                           /* predict with the node's data */
    return;                     /* (use the node as a fallback), */
  }                             /* then terminate the function */

  /* -- attribute value is unknown --- */
  weight /= node->frq;          /* remove the node weight and */
  data    = node->vec;          /* get the node's data vector */
  for (i = node->size; --i >= 0; data++) {
    if (islink(data, node))     /* traverse all child nodes, */
      continue;                 /* but skip links to other children */
    if      (data->child)       /* if the child node exists, execute */
      _exec(dt, data->child, weight *data->child->frq);   /* subtree */
    else if (dt->type == AT_SYM)/* if the target att. is symbolic */
      dt->buf[node->trg.i] += weight *EPSILON;
    else {                      /* if the target att. is numeric */
      dt->buf[0] += weight *EPSILON;
      dt->buf[2] += weight *EPSILON *node->trg.f;
    }                           /* predict with the node's data */
  }                             /* (use the node as a fallback) */
}  /* _exec() */

/*--------------------------------------------------------------------*/

int dt_exec (DTREE *dt, TUPLE *tuple, INST *res,
             float *supp, float *conf)
{                               /* --- classify tuple with dec. tree */
  int    i, k;                  /* loop variable, buffer size */
  double *frq, *max;            /* to traverse the class frequencies */
  double sum;                   /* sum of the class frequencies */

  assert(dt);                   /* check the function argument */
  dt->tuple = tuple;            /* store the tuple to classify */
  if (dt->total < 0) {          /* if total is not valid, compute it */
    if (dt_total(dt) < 0) return -1; }
  k = (dt->type == AT_SYM) ? dt->clscnt : 3;
  for (frq = dt->buf +(i = k); --i >= 0; )
    *--frq = 0;                 /* clear the buffer vector */
  if (dt->root)                 /* if a decision tree exists, */
    _exec(dt, dt->root, 1);     /* execute the decision tree */
  if (dt->type != AT_SYM) {     /* if the target att. is numeric, */
    sum    = dt->buf[0];        /* compute the pedicted value and */
    res->f = (float)dt->buf[2]; /* store prediction support and error */
    if (supp) *supp = (float)sum;
    if (conf) *conf = (sum > 0) ? (float)sqrt(dt->buf[1] /sum) : 0.0F; }
  else {                        /* if the target att. is symbolic */
    frq = max = dt->buf;        /* traverse the */
    sum = *frq;                 /* class frequency table */
    for (i = dt->clscnt; --i > 0; ) {
      if (*++frq > *max) max = frq;
      sum += *frq;              /* find the most frequent class */
    }                           /* and sum the class frequencies */
    res->i = (int)(max-dt->buf);/* set the classification result */
    if (supp) *supp = (float)sum;
    if (conf) *conf = (float)(*max /sum);
  }                             /* store class. support and error */
  return 0;                     /* return 'ok' */
}  /* dt_exec() */

/*----------------------------------------------------------------------
  Description Functions
----------------------------------------------------------------------*/

static void _indent (DTOUT *out, int newline)
{                               /* --- indent output line */
  int i;                        /* loop variable */

  assert(out);                  /* check the function arguments */
  if (newline) {                /* terminate line */
    fputc('\n', out->file); out->pos = 0; }
  for (i = out->pos; i < out->ind; i++)
    fputc(' ',  out->file);     /* indent next line */
  if (out->ind > out->pos) out->pos = out->ind;
}  /* _indent() */              /* note new position */

/*--------------------------------------------------------------------*/

static void _clsout (DTOUT *out, DTNODE *node)
{                               /* --- output of class frequencies */
  int    i, k;                  /* loop variables, buffer */
  ATT    *att;                  /* class attribute */
  DTDATA *data;                 /* to traverse node's data vector */
  int    len;                   /* length of the class output */
  double sum = 0;               /* sum of class frequencies */

  att = as_att(out->attset, node->attid);
  if (out->mode & DT_REL) {     /* if to print relative numbers */
    data = node->vec +node->size;
    for (i = node->size; --i >= 0; ) sum += (--data)->frq;
  }                             /* compute the frequency sum */
  data = node->vec;             /* traverse the classes */
  for (i = k = 0; i < node->size; data++, i++) {
    if (data->frq <= 0)         /* if a class does not occur, */
      continue;                 /* skip it, otherwise get its name */
    if (k > 0) {                /* if not the first class value, */
      fputc(',', out->file); out->pos++; }  /* print a separator */
    len  = sc_format(out->buf, att_valname(att, i), 0);
    len += sprintf(out->buf +len, ": %g", data->frq);
    if (out->mode & DT_REL)     /* if to print relative numbers */
      len += sprintf(out->buf +len, " (%.1f%%)", (data->frq /sum)*100);
    if ((out->pos +len > out->max -4)   /* if the current line */
    &&  (out->pos      > out->ind))     /* would get too long, */
      _indent(out, 1);          /* start a new line and indent */
    else if (k > 0) {           /* if there is sufficient space, */
      fputc(' ', out->file); out->pos += 1; }   /* print a blank */
    fputs(out->buf, out->file); /* print the class information */
    out->pos += len; k++;       /* compute the new position */
  }                             /* and count the printed class */
}  /* _clsout() */

/*--------------------------------------------------------------------*/

static void _dtout (DTOUT *out, DTNODE *node)
{                               /* --- output of decision tree */
  int    i, k, n;               /* loop variables */
  ATT    *att;                  /* target/test attribute */
  DTDATA *data, *d2;            /* to traverse node's data vector */
  int    len, swd;              /* length of the class output */
  int    type;                  /* attribute type */

  assert(out && node);          /* check the function arguments */
  fputs("{ ", out->file);       /* start a branch and */
  out->pos = out->ind += 2;     /* compute the new position */

  /* --- print class frequencies/predicted value --- */
  if (node->flags & DT_LEAF) {  /* if the node is a leaf */
    if (out->type == AT_SYM)    /* if the target is symbolic, */
      _clsout(out, node);       /* print the class frequencies */
    else {                      /* if the target is numeric */
      out->pos += fprintf(out->file, "%g ", node->trg.f);
      out->pos += fprintf(out->file, "~%g ", (node->frq > 0)
                          ? sqrt(node->err/node->frq) : 0);
      out->pos += fprintf(out->file, "[%g]", node->frq);
    }                           /* print the predicted value */
    fputs(" }", out->file);     /* terminate the leaf, */
    out->pos += 2;              /* compute the new position, */
    out->ind -= 2; return;      /* reset the indentation, */
  }                             /* and terminate the function */

  /* --- print attribute test --- */
  att = as_att(out->attset, node->attid);
  sc_format(out->buf, att_name(att), 0);
  fputc('(', out->file);        /* format and print */
  fputs(out->buf, out->file);   /* the attribute name */
  type = att_type(att);         /* get the attribute type */
  if (type != AT_SYM) fprintf(out->file, "|%g", node->cut);
  fputc(')', out->file);        /* print cut value for numeric atts. */
  _indent(out, 1);              /* start a new line for the children */
  data = node->vec;             /* traverse child nodes (subtrees) */
  for (i = k = 0; i < node->size; data++, i++) {
    if (!data->child || islink(data, node))
      continue;                 /* skip empty subtrees and links */
    if (k++ > 0) {              /* if this is not the first subtree, */
      fputc(',', out->file); _indent(out, 1); } /* print a separator */
    if (type != AT_SYM) {       /* if numeric attribute */
      out->buf[0] = (i > 0) ? '>' : '<';
      out->buf[1] = '\0';       /* get one a character name */
      len = swd = 1; }          /* representing the comparison */
    else {                      /* if symbolic attribute */
      len = sc_format(out->buf, att_valname(att, i), 0);
      swd = att_valwd(att, 1);  /* get value name and column width */
      for (n = 0; n < node->size; n++) {
        if (n == i) continue;   /* traverse children except current */
        d2 = node->vec +n;      /* follow link chain */
        while (islink(d2, node)) d2 = d2->link;
        if (d2 != data)         /* skip children that are */
          continue;             /* not linked to current child */
        fputs(out->buf, out->file);
        fputc(',', out->file);  /* print value name and separator */
        if (out->mode & DT_ALIGN)  /* if to align values, */
          _indent(out, 1);         /* start a new line */
        len = sc_format(out->buf, att_valname(att, n), 0);
      }                         /* get name of next value */
    }                           /* (in the value subset) */
    fputs(out->buf, out->file); /* print name of attribute value */
    out->pos += len;            /* calculate new position */
    out->ind += (out->mode & DT_ALIGN) ? swd : 1;
    _indent(out, 0);            /* set indentation and indent */
    fputc(':', out->file); out->pos++; out->ind++;
    _dtout(out, data->child);   /* print next level (recursively) */
    out->ind -= (out->mode & DT_ALIGN) ? swd : 1;
    out->ind--;                 /* reset indentation */
  }
  out->ind -= 2;                /* reset indentation */
  if ((out->pos >= out->max)    /* if line would get too long */
  || ((out->ind <= 0) && (out->pos >= out->max-1)))
    _indent(out, 1);            /* start a new line */
  fputc('}', out->file); out->pos++;
}  /* _dtout() */               /* terminate branch */

/*--------------------------------------------------------------------*/

int dt_desc (DTREE *dt, FILE *file, int mode, int maxlen)
{                               /* --- describe a decision tree */
  int   i, k;                   /* loop variables */
  DTOUT out;                    /* output information */

  assert(dt && file);           /* check the function arguments */

  /* --- print header (as a comment) --- */
  if (mode & DT_TITLE) {        /* if the title flag is set */
    i = k = (maxlen > 0) ? maxlen -2 : 70;
    fputs("/*", file); while (--i >= 0) putc('-', file);
    fprintf(file, (dt->type == AT_SYM)
            ? "\n  decision tree\n" : "\n  regression tree\n");
    while (--k >= 0) putc('-', file); fputs("*/\n", file);
  }                             /* print a title header */
  out.max    = (maxlen <= 0) ? INT_MAX : maxlen;
  out.file   = file;            /* note the function arguments */
  out.attset = dt->attset;      /* in the output structure */
  out.mode   = mode;            /* that will be passed down */
  out.type   = dt->type;

  /* --- print decision tree --- */
  fputs("dtree(", file);        /* decision tree indicator */
  sc_format(out.buf, att_name(as_att(dt->attset, dt->trgid)), 0);
  fputs(out.buf, file);         /* format and print */
  fputs(") =\n", file);         /* the target attribute's name */
  out.ind = 0;                  /* initialize the indentation */
  if (dt->root) _dtout(&out, dt->root);
  else fputs("{ }", file);      /* recursively print the tree */
  fputs(";\n", file);           /* terminate the last output line */

  /* --- print additional information (as a comment) --- */
  if (mode & DT_INFO) {         /* if the add. info. flag is set */
    i = k = (maxlen > 0) ? maxlen -2 : 70;
    fputs("\n/*", file); while (--i >= 0) putc('-', file);
    fprintf(file, "\n  number of attributes: %d",   dt_attcnt(dt));
    fprintf(file, "\n  tree height         : %d",   dt->height);
    fprintf(file, "\n  number of nodes     : %d",   dt->size);
    fprintf(file, "\n  number of tuples    : %g\n", dt_total(dt));
    while (--k >= 0) putc('-', file); fputs("*/\n", file);
  }                             /* print add. tree information */
  return (ferror(file)) ? -1 : 0;
}  /* dt_desc() */              /* return the writing result */

/*----------------------------------------------------------------------
  Parse Functions
----------------------------------------------------------------------*/
#ifdef DT_PARSE

static int _paerr (SCAN *scan, int code, int c, const char *s)
{                               /* --- report a parse error */
  char src[  AS_MAXLEN+1];      /* buffer for string to format */
  char dst[4*AS_MAXLEN+4];      /* buffer for fomatted string */

  assert(scan);                 /* check the function arguments */
  strncpy(src, sc_value(scan), AS_MAXLEN); src[AS_MAXLEN] = '\0';
  sc_format(dst, src, 1);       /* copy and format the token value */
  if      (code == E_CHREXP) return sc_error(scan, code, c, dst);
  else if (code == E_STREXP) return sc_error(scan, code, s, dst);
  else                       return sc_error(scan, code,    dst);
}  /* _paerr() */               /* print an error message */

/*--------------------------------------------------------------------*/

static int _clsin (DTREE *dt, SCAN *scan, DTNODE *node)
{                               /* --- read class frequencies */
  ATT    *att;                  /* class attribute */
  DTDATA *data;                 /* to traverse the data vector */
  int    cls;                   /* class value identifier */
  int    t;                     /* buffer for token/result */
  double f;                     /* buffer for class frequency */

  assert(dt && scan && node);   /* check the function arguments */
  for (data = node->vec +(cls = dt->clscnt); --cls >= 0; )
    (--data)->frq = -1.0F;      /* clear the class frequencies */
  att = as_att(dt->attset, node->attid);
  while (1) {                   /* class read loop */
    t = sc_token(scan);         /* check for a name */
    if ((t != T_ID) && (t != T_NUM)) return ERROR(E_VALEXP);
    cls = att_valid(att, sc_value(scan));
    if (cls < 0)                     return ERROR(E_UNKVAL);
    if (data[cls].frq >= 0)          return ERROR(E_DUPVAL);
    GET_TOK();                  /* get and check the class value */
    GET_CHR(':');               /* consume ':' */
    if (sc_token(scan) != T_NUM)     return ERROR(E_NUMEXP);
    f = atof(sc_value(scan));   /* get and check the value frequency */
    if ((f < 0) || (f > FLT_MAX))    return ERROR(E_ILLNUM);
    data[cls].frq = (float)f;   /* set the class frequency */
    GET_TOK();                  /* and consume it */
    if (sc_token(scan) == '('){ /* if a relative number follows */
      GET_TOK();                /* consume '(' */
      if (sc_token(scan) != T_NUM)   return ERROR(E_NUMEXP);
      f = atof(sc_value(scan)); /* get and check the rel. frequency */
      if ((f < 0) || (f > 100))      return ERROR(E_ILLNUM);
      GET_TOK();                /* consume the rel. frequency */
      GET_CHR('%');             /* consume '%' */
      GET_CHR(')');             /* consume ')' */
    }
    if (sc_token(scan) != ',') break; 
    GET_TOK();                  /* if at the end of the list, abort, */
  }                             /* otherwise consume ',' */
  for (data = node->vec +(cls = dt->clscnt); --cls >= 0; )
    if ((--data)->frq < 0) data->frq = 0.0F;
  return 0;                     /* clear the unset frequencies */
}  /* _clsin() */               /* and return 'ok' */

/*--------------------------------------------------------------------*/

static int _getval (DTREE *dt, SCAN *scan, ATT *att)
{                               /* --- read (list of) values */
  int v1, v2;                   /* attribute values */
  int t;                        /* buffer for token */

  assert(dt && scan && att);    /* check the function arguments */
  t = sc_token(scan);           /* check for a name */
  if ((t != T_ID) && (t != T_NUM))   return ERROR(E_VALEXP);
  v1 = att_valid(att, sc_value(scan));
  if (v1 < 0)                        return ERROR(E_UNKVAL);
  if (dt->curr->vec[v1].child)       return ERROR(E_DUPVAL);
  GET_TOK();                    /* get and consume an attribute value */
  while (sc_token(scan) == ','){/* read a value list (subset) */
    GET_TOK();                  /* consume ',' */
    t = sc_token(scan);         /* check for a name */
    if ((t != T_ID) && (t != T_NUM)) return ERROR(E_VALEXP);
    v2 = att_valid(att, sc_value(scan));
    if (v2 < 0)                      return ERROR(E_UNKVAL);
    if ((v2 == v1)              /* get and check the attribute value */
    ||  (dt_link(dt, v2, v1) < 0))   return ERROR(E_DUPVAL);
    GET_TOK();                  /* link subtrees of decision tree */
  }                             /* and consume the attribute value */
  GET_CHR(':');                 /* consume ':' */
  return v1;                    /* return the value identifier */
}  /* _getval() */

/*--------------------------------------------------------------------*/

static int _dtin (DTREE *dt, SCAN *scan)
{                               /* --- recursively read dec. tree */
  int    type;                  /* type of test attribute */
  ATT    *att;                  /* test attribute */
  int    i, k, t;               /* att./value id, flag, buffer */
  double f;                     /* buffer for a floating point number */
  int    err = 0;               /* parse error status */

  assert(dt && scan);           /* check the function arguments */
  GET_CHR('{');                 /* consume '{' */
  if (sc_token(scan) == '}') {  /* if the (sub)tree is empty, */
    GET_TOK(); return 0; }      /* consume '}' and return */

  /* --- read a leaf node --- */
  if (sc_token(scan) != '(') {  /* if at a leaf (no further test) */
    if (dt_node(dt, -1, UV_FLT) != 0)
      return ERROR(E_NOMEM);    /* create a leaf node */
    if (dt->type == AT_SYM) {   /* if the target attrib. is symbolic, */
      t = _clsin(dt, scan, dt->curr);
      if (t) return t; }        /* read the class frequencies */
    else {                      /* if the target att. is numeric */
      if (sc_token(scan) != T_NUM)       return ERROR(E_NUMEXP);
      f = atof(sc_value(scan)); /* get and check the target value */
      if ((f < UV_FLT) || (f > FLT_MAX)) return ERROR(E_ILLNUM);
      GET_TOK();                /* consume the target value */
      dt->curr->trg.f = (float)f;           /* and store it */
      GET_CHR('~');             /* consume '~' */
      if (sc_token(scan) != T_NUM)       return ERROR(E_NUMEXP);
      f = atof(sc_value(scan)); /* get and check the prediction error */
      if ((f < 0) || (f > DBL_MAX))      return ERROR(E_ILLNUM);
      GET_TOK();                /* consume the prediction error */
      dt->curr->err = f*f;      /* and store its square (mse) */
      GET_CHR('[');             /* consume '[' */
      if (sc_token(scan) != T_NUM)       return ERROR(E_NUMEXP);
      f = atof(sc_value(scan)); /* get and check the node frequency */
      if ((f < 0) || (f > FLT_MAX))      return ERROR(E_ILLNUM);
      GET_TOK();                /* consume the node frequency */
      dt->curr->frq = (float)f; /* and store it */
      dt->curr->err *= f;       /* compute the sse from the mse */
      GET_CHR(']');             /* consume ']' */
    }  
    GET_CHR('}'); return 0;     /* consume '}' and */
  }                             /* terminate the function */

  /* --- read a test node --- */
  GET_TOK();                    /* consume '(' */
  t = sc_token(scan);           /* check for a name */
  if ((t != T_ID) && (t != T_NUM))     return ERROR(E_ATTEXP);
  i = as_attid(dt->attset, sc_value(scan));
  if (i < 0)                           return ERROR(E_UNKATT);
  GET_TOK();                    /* get and consume the attribute */
  if (dt_node(dt, i, 0) != 0)          return ERROR(E_NOMEM);
  att  = as_att(dt->attset, i); /* create a test node and */
  type = att_type(att);         /* get the test attribute */
  if (type != AT_SYM) {         /* if the attribute is numeric, */
    GET_CHR('|');               /* consume '|' */
    if (sc_token(scan) != T_NUM)       return ERROR(E_NUMEXP);
    f = atof(sc_value(scan));   /* get and check the cut value */
    if ((f < UV_FLT) || (f > FLT_MAX)) return ERROR(E_ILLNUM);
    dt->curr->cut = (float)f;   /* store the cut value */
    GET_TOK();                  /* and consume it */
  }
  GET_CHR(')');                 /* consume ')' */
  k = 0;                        /* initialize the read counter */
  while (sc_token(scan) != '}'){/* attribute value read loop */
    if (++k > 1) {              /* if this is not the first value, */
      GET_CHR(','); }           /* consume ',' (separator) */
    if (type != AT_SYM) {       /* --- if the attribute is numeric */
      t = sc_token(scan);       /* get the next token */
      if      (t == '<') i = 0; /* if lower   than cut: index 0 */
      else if (t == '>') i = 1; /* if greater than cut: index 1 */
      else { ERROR(E_UNKVAL); RECOVER(); err = 1; continue; }
      if (dt->curr->vec[i].child) {
             ERROR(E_DUPVAL); RECOVER(); err = 1; continue; }
      GET_TOK();                /* consume '<' or '>' */
      GET_CHR(':'); }           /* consume ':' */
    else {                      /* --- if the attribute is symbolic */
      i = _getval(dt, scan, att);
      if (i < 0) {            RECOVER(); err = 1; continue; }
    }                           /* read a (list of) value(s) */
    if (dt_down(dt, i, 0) < 0)  /* go down in the tree */
      return ERROR(E_NOMEM);
    t = _dtin(dt, scan);        /* get subtree recursively */
    dt_up(dt, 0);               /* go up again in the tree */
    if (t) {                    /* if an error occurred */
      if      (t < E_FWRITE) { RECOVER(); }
      else if (t < 0)        return t;
      err = 1;                  /* try to recover, but */
    }                           /* abort on fatal errors */
  }
  GET_TOK();                    /* consume '}' */
  return err;                   /* return parse error status */
}  /* _dtin() */

/*--------------------------------------------------------------------*/

static int _parse (ATTSET *attset, SCAN *scan, DTREE **pdt)
{                               /* --- read decision tree */
  int trgid;                    /* class attribute identifier */
  int t;                        /* buffer for token/type/result */

  assert(attset && scan && pdt);/* check the function arguments */
  if ((sc_token(scan) != T_ID)
  ||  (strcmp(sc_value(scan), "dtree") != 0))
    return ERR_STR("dtree");    /* check for 'dtree' */
  GET_TOK();                    /* consume 'dtree' */
  GET_CHR('(');                 /* consume '(' */
  t = sc_token(scan);           /* check for a name */
  if ((t != T_ID) && (t != T_NUM)) return ERROR(E_ATTEXP);
  trgid = as_attid(attset, sc_value(scan));
  if (trgid < 0)                   return ERROR(E_UNKATT);
  *pdt = dt_create(attset, trgid);  /* get the target attribute */
  if (!*pdt) return ERROR(E_NOMEM); /* and create a dec./reg. tree */
  GET_TOK();                    /* consume the target attribute name */
  GET_CHR(')');                 /* consume ')' */
  GET_CHR('=');                 /* consume '=' */
  if (_dtin(*pdt, scan) != 0)   /* recursively read dec./reg. tree */
    return -1;                  /* and check for an error */
  GET_CHR(';');                 /* consume ';' */
  return 0;                     /* return 'ok' */
}  /* _parse() */

/*--------------------------------------------------------------------*/

DTREE* dt_parse (ATTSET *attset, SCAN *scan)
{                               /* --- parse a decision tree */
  DTREE *dt = NULL;             /* created decision tree */

  assert(attset && scan);       /* check the function arguments */
  sc_errmsgs(scan, errmsgs, MSGCNT);
  if (_parse(attset, scan, &dt) != 0) {
    if (dt) dt_delete(dt, 0);   /* parse a decision tree */
    return NULL;                /* if an error occurred, */
  }                             /* delete the tree and abort */
  dt_total(dt);                 /* compute the node frequencies and */
  return dt;                    /* return the created decision tree */
}  /* dt_parse() */

#endif  /* #ifdef DT_PARSE */
/*----------------------------------------------------------------------
  Rule Extraction Functions
----------------------------------------------------------------------*/
#ifdef DT_RULES

static int _rules (DTRULE *rx, DTNODE *node, RULE *path)
{                               /* --- recursively collect rules */
  RULE   *rule;                 /* created rule */
  int    i, k, r;               /* loop variables, buffers */
  DTDATA *data, *d2;            /* to traverse data vector */
  INST   inst;                  /* temporary buffer for instance */

  assert(rx && node);           /* check the function arguments */

  /* --- leaf node --- */
  if (node->flags & DT_LEAF) {  /* if the node is a leaf */
    rule = r_create(rx->attset, (path) ? r_condcnt(path) : 0);
    if (!rule) return -1;       /* create a new rule and */
    if (path) r_copy(rule,path);/* copy the path into it */
    inst = node->trg;           /* get the target value */
    r_headset(rule, node->attid, R_EQ, &inst);
    r_setsupp(rule, node->frq); /* set head and rule information */
    if (rx->type == AT_SYM)     /* if target attribute is symbolic */
      r_setconf(rule, (node->frq > 0) ? 1 -node->err/node->frq : 1);
    else                        /* if target attribute is numeric */
      r_setconf(rule, (node->frq > 0) ? sqrt(node->err/node->frq) : 0);
    if (rs_ruleadd(rx->ruleset, rule) < 0) {
      r_delete(rule); return -1; }
    return 0;                   /* add the rule to the rule set */
  }                             /* and return 'ok' */

  /* --- test node --- */
  rule = r_create(rx->attset, (path) ? (r_condcnt(path)+1) : 1);
  if (!rule) return -1;         /* create a new rule */
  if (att_type(as_att(rx->attset, node->attid)) != AT_SYM) {
    data   = node->vec;         /* if the attribute is numeric */
    inst.f = node->cut;         /* get the cut value */
    if (data->child) {          /* if there is a left subtree */
      if (path) r_copy(rule, path);   /* copy path into rule */
      r = r_condadd(rule, node->attid, R_LE, &inst);
      if ((r == -1)             /* add condition to the rule */
      || ((r >=  0) && (_rules(rx, data->child, rule) != 0))) {
        r_delete(rule); return -1; }
    }                           /* collect rules from left subtree */
    if ((++data)->child) {      /* if there is a right subtree */
      if (path) r_copy(rule, path);   /* copy path into rule or */
      else      r_condrem(rule, -1);  /* remove all conditions */
      r = r_condadd(rule, node->attid, R_GE, &inst);
      if ((r == -1)             /* add condition to the rule */
      || ((r >=  0) && (_rules(rx, data->child, rule) != 0))) {
        r_delete(rule); return -1; }
    } }                         /* collect rules from right subtree */
  else {                        /* if attribute is symbolic */
    inst.p = rx->subset;        /* set the subset vector */
    for (i = 0; i < node->size; i++) {
      data = node->vec +i;      /* traverse attribute values */
      if (!data->child || islink(data, node))
        continue;               /* skip empty children and links */
      for (k = r = 0; k < node->size; k++) {
        if (k == i) continue;   /* traverse children except current */
        d2 = node->vec +k;      /* follow link chain */
        while (islink(d2, node)) d2 = d2->link;
        if (d2 != data)         /* skip children that are */
          continue;             /* not linked to current child */
        rx->subset[++r] = k;    /* note the value identifier */
      }                         /* in the value subset vector */
      rx->subset[++r] = i;      /* store the current value */
      rx->subset[0]   = r;      /* and the number of values */
      if (path) r_copy(rule, path);   /* copy path into rule */
      else      r_condrem(rule, -1);  /* remove all conditions */
      r = r_condadd(rule, node->attid, R_IN, &inst);
      if ((r == -1)             /* add condition to the rule */
      || ((r >=  0) && (_rules(rx, data->child, rule) != 0))) {
        r_delete(rule); return -1; }
    }                           /* collect the rules */
  }                             /* from the subtree */
  r_delete(rule);               /* delete the rule buffer */
  return 0;                     /* and return 'ok' */
}  /* _rules() */

/*--------------------------------------------------------------------*/

RULESET* dt_rules (DTREE *dt)
{                               /* --- create a rule set */
  int     i, n;                 /* loop variable, buffer */
  int     max = 0;              /* maximal number of values */
  RULESET *ruleset;             /* created rule set */
  DTRULE  *rx;                  /* rule extraction info */

  assert(dt);                   /* check the function argument */
  ruleset = rs_create(dt->attset, r_delete);
  if (!ruleset || !dt->root)    /* create an empty rule set */
    return ruleset;             /* and check for an empty tree */
  for (i = as_attcnt(dt->attset); --i >= 0; ) {
    n = att_valcnt(as_att(dt->attset, i));
    if (n > max) max = n;       /* determine the maximal number */
  }                             /* of values per attribute */
  rx = (DTRULE*)malloc(sizeof(DTRULE) +max *sizeof(int));
  if (!rx) { rs_delete(ruleset, 0); return NULL; }
  rx->ruleset = ruleset;        /* create a rule extraction info. */
  rx->attset  = dt->attset;     /* and note the relevant objects */
  rx->type    = dt->type;
  n = _rules(rx,dt->root,NULL); /* collect rules from the tree */
  free(rx);                     /* delete the rule extraction info. */
  if (n != 0) { rs_delete(ruleset, 0); return NULL; }
  return ruleset;               /* return the created rule set */
}  /* dt_rules() */

#endif  /* #ifdef DT_RULES */
