/*----------------------------------------------------------------------
  File    : dtree2.c
  Contents: decision and regression tree management,
            functions for growing and pruning
  Author  : Christian Borgelt
  History : 12.06.1997 file created (as dtree.c)
            08.08.1997 first version completed
            15.08.1997 pruning functions completed
            25.08.1997 multiple child references made possible
            26.08.1997 node counting added (function _count)
            28.08.1997 pessimistic pruning added
            30.08.1997 treatment of new values during pruning enabled
            25.08.1997 multiple child references replaced by links
            18.09.1997 bug in measure evaluation removed
            25.09.1997 subsets for symbolic attributes added
            01.10.1997 bugs in function _pr_tab removed
            02.02.1998 links between children simplified
            03.02.1998 adapted to changed interface of module 'table'
            30.03.1998 bug in _eval_flt conc. unknown values removed
            27.08.1998 single class value coll. in _eval_set corrected
            21.09.1998 bug in function _adapt (concerning links) removed
            15.10.1998 major revision completed
            30.10.1998 bug in function _pr_tab removed
            15.12.1998 minor improvements, assertions added
            24.02.1999 parameters 'params' and 'minval' added to dt_grow
            20.03.1999 roundoff error handling improved
            29.02.2000 bug in function _eval_num removed
            02.12.2000 growing and pruning moved to this file
            03.12.2000 decision tree growing partly redesigned
            12.12.2000 leaf creation in pruning with a table redesigned
            17.12.2000 regression tree functionality added
            23.07.2001 global variables removed
            26.09.2001 normal distribution table replaced by a function
            19.10.2001 conf. level pruning for numeric targets added
            01.03.2002 check of largest branch in _pr_tab made optional
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "vecops.h"
#include "dtree.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define WORTHLESS   -DBL_MAX    /* to indicate a worthless attribute */
#define EPSILON     1e-6        /* to handle roundoff errors */
#define MINERROR    0.1         /* minimum error for division */
#define MAXFACT     1e6         /* maximum factor for error estim. */

/* --- normal distribution parameters --- */
#define NQ_A0       (-0.322232431088)      /* coefficients of */
#define NQ_A1       (-1)                   /* the polynomials */
#define NQ_A2       (-0.342242088547)      /* used to compute */
#define NQ_A3       (-0.0204231210245)     /* an approximation of */
#define NQ_A4       (-0.453642210148e-4)   /* the quantiles of the */
#define NQ_B0         0.0993484626060      /* normal distribution */
#define NQ_B1         0.588581570495
#define NQ_B2         0.531103462366
#define NQ_B3         0.103537752850
#define NQ_B4         0.0038560700634
#define NQ_PMIN     1e-9        /* minimum value for the probability */
#define NQ_MAX      6           /* maximum value for quantile */

/* --- functions --- */
#define islink(d,n) (((d)->link >= (n)->vec) && \
                     ((d)->link <  (n)->vec +(n)->size))

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- tuple selection info. --- */
  int    col;                   /* column index */
  INST   val;                   /* column value */
  DTDATA *vec;                  /* data vector of the current node */
} TSEL;                         /* (tuple selection information) */

typedef int    GRPFN  (TUPLE **tpls, int n, const TSEL *tsel);
typedef double ESTFN  (double n, double e);

typedef struct _grow * PGROW;   /* to avoid warning in def. of EVALFN */
typedef double EVALFN (PGROW gi, TUPLE **tpls, int n,
                       int attid, float *cut);

typedef struct _grow {          /* --- tree grow information --- */
  DTREE  *dtree;                /* current decision/regression tree */
  ATTSET *attset;               /* attribute set of decision tree */
  TUPLE  **tpls;                /* vector of tuples with known class */
  EVALFN *eval_sym;             /* eval. function for symbolic atts. */
  EVALFN *eval_num;             /* eval. function for numeric  atts. */
  void   *curr;                 /* freq./var. table for current att. */
  void   *best;                 /* freq./var. table for best    att. */
  void   *numt;                 /* freq./var. table for numeric att. */
  double err;                   /* number of errors of subtree */
  int    type;                  /* type of the target attribute */
  int    measure;               /* evaluation measure */
  double *params;               /* optional parameters */
  double minval;                /* minimal value of measure for split */
  float  mincnt;                /* minimal number of tuples */
  int    maxht;                 /* maximal height of decision tree */
  int    flags;                 /* grow flags, e.g. DT_SUBSET */
  int    sets[1];               /* index vector for value subsets */
} GROW;                         /* (tree grow information) */

typedef struct {                /* --- tree pruning information --- */
  DTREE  *dtree;                /* current decision/regression tree */
  ATTSET *attset;               /* attribute set of decision tree */
  TUPLE  **tpls;                /* vector of tuples with known class */
  void   *tab;                  /* freq./var. table for leaf creation */
  ESTFN  *est;                  /* error estimation function */
  double err;                   /* number of errors of subtree */
  int    type;                  /* type of the target attribute */
  int    maxht;                 /* maximal height of decision tree */
  int    chklb;                 /* whether to check largest branch */
} PRUNE;                        /* (tree pruning information) */

/*----------------------------------------------------------------------
  Normal Distribution Function
----------------------------------------------------------------------*/
#ifdef DT_PRUNE

static double _ndqtl (double prob)
{                               /* --- comp. quantile of normal dist. */
  double p, x;                  /*     with mean 0 and variance 1 */

  assert((prob >= 0)            /* check the function argument */
      && (prob <= 1));          /* (a probability must be in [0,1]) */
  if (prob == 0.5) return 0;    /* treat prob = 0.5 as a special case */
  p = (prob > 0.5) ? 1-prob : prob; /* transform to left tail if nec. */
  if (p < NQ_PMIN) return (prob < 0.5) ? -NQ_MAX : NQ_MAX;
  x  = sqrt(log(1/(p*p)));      /* compute quotient of polynomials */
  x += ((((NQ_A4 *x +NQ_A3) *x +NQ_A2) *x +NQ_A1) *x +NQ_A0)
     / ((((NQ_B4 *x +NQ_B3) *x +NQ_B2) *x +NQ_B1) *x +NQ_B0);
  return (prob < 0.5) ? -x : x; /* retransform to right tail if nec. */
}  /* _ndqtl() */

/*----------------------------------------------------------------------
References: - R.E. Odeh and J.O. Evans.
              The percentage points of the normal distribution.
              Applied Statistics 22:96--97, 1974
            - J.D. Beasley and S.G. Springer.
              Algorithm AS 111:
              The percentage points of the normal distribution.
              Applied Statistics 26:118--121, 1977
            - M.J. Wichura
              Algorithm AS 241:
              The percentage points of the normal distribution.
              Applied Statistics 37:477--484, 1988
----------------------------------------------------------------------*/
#endif  /* #ifdef DT_PRUNE */
/*----------------------------------------------------------------------
  Grouping Functions
----------------------------------------------------------------------*/

static int _grp_sym (TUPLE **tpls, int n, const TSEL *tsel)
{                               /* --- group symbolic values */
  TUPLE **src, **dst, *t;       /* source and destination, buffer */

  assert(tpls && tsel && (n >= 0));   /* check the function arguments */
  for (src = dst = tpls; --n >= 0; src++)
    if (tpl_colval(*src, tsel->col)->i == tsel->val.i) {
      t = *src; *src = *dst; *dst++ = t; }
  return (int)(dst -tpls);      /* group qualifying tuples */
}  /* _grp_sym() */             /* and return their number */

/*--------------------------------------------------------------------*/

static int _grp_set (TUPLE **tpls, int n, const TSEL *tsel)
{                               /* --- group a set of values */
  TUPLE **src, **dst, *t;       /* source and destination, buffer */
  register int i;               /* check value and link */

  assert(tpls && tsel && (n >= 0));   /* check the function arguments */
  for (src = dst = tpls; --n >= 0; src++)
    if (((i = tpl_colval(*src, tsel->col)->i) == tsel->val.i)
    ||  (tsel->vec[i].link -tsel->vec         == tsel->val.i)) {
      t = *src; *src = *dst; *dst++ = t; }
  return (int)(dst -tpls);      /* group qualifying tuples */
}  /* _grp_set() */             /* and return their number */

/*--------------------------------------------------------------------*/

static int _grp_uvint (TUPLE **tpls, int n, const TSEL *tsel)
{                               /* --- group unknown values (integer) */
  TUPLE **src, **dst, *t;       /* source and destination, buffer */

  assert(tpls && tsel && (n >= 0));   /* check the function arguments */
  for (src = dst = tpls; --n >= 0; src++)
    if (tpl_colval(*src, tsel->col)->i <= UV_INT) {
      t = *src; *src = *dst; *dst++ = t; }
  return (int)(dst -tpls);      /* group qualifying tuples */
}  /* _grp_uvint() */           /* and return their number */

/*--------------------------------------------------------------------*/

static int _grp_int (TUPLE **tpls, int n, const TSEL *tsel)
{                               /* --- group greater than (integer) */
  TUPLE **src, **dst, *t;       /* source and destination, buffer */

  assert(tpls && tsel && (n >= 0));   /* check the function arguments */
  for (src = dst = tpls; --n >= 0; src++)
   if (tpl_colval(*src, tsel->col)->i > tsel->val.i) {
      t = *src; *src = *dst; *dst++ = t; }
  return (int)(dst -tpls);      /* group qualifying tuples */
}  /* _grp_int() */             /* and return their number */

/*--------------------------------------------------------------------*/

static int _grp_uvflt (TUPLE **tpls, int n, const TSEL *tsel)
{                               /* --- group unknown values (float) */
  TUPLE **src, **dst, *t;       /* source and destination, buffer */

  assert(tpls && tsel && (n >= 0));   /* check the function arguments */
  for (src = dst = tpls; --n >= 0; src++)
    if (tpl_colval(*src, tsel->col)->f <= UV_FLT) {
      t = *src; *src = *dst; *dst++ = t; }
  return (int)(dst -tpls);      /* group qualifying tuples */
}  /* _grp_uvflt() */           /* and return their number */

/*--------------------------------------------------------------------*/

static int _grp_flt (TUPLE **tpls, int n, const TSEL *tsel)
{                               /* --- group greater than (float) */
  TUPLE **src, **dst, *t;       /* source and destination, buffer */

  assert(tpls && tsel && (n >= 0));   /* check the function arguments */
  for (src = dst = tpls; --n >= 0; src++)
    if (tpl_colval(*src, tsel->col)->f > tsel->val.f) {
      t = *src; *src = *dst; *dst++ = t; }
  return (int)(dst -tpls);      /* group qualifying tuples */
}  /* _grp_flt() */             /* and return their number */

/*--------------------------------------------------------------------*/

static void _mul_wgt (TUPLE **tpls, int n, double f)
{                               /* --- multiply tuple weights */
  INST *info;                   /* temporary buffer */

  assert(tpls && (n >= 0));     /* check the function arguments */
  for (tpls += n; --n >= 0; ) { /* traverse the tuples */
    info = tpl_info(*--tpls);   /* get the additional information */
    info->f = (float)(info->f *f);
  }                             /* multiply with weighting factor */
}  /* _mul_wgt() */

/*--------------------------------------------------------------------*/
#ifdef DT_PRUNE

static double _sum_wgt (TUPLE **tpls, int n)
{                               /* --- sum tuple weights */
  double wgt = 0;               /* traverse the tuples */

  assert(tpls && (n >= 0));     /* check the function arguments */
  for (tpls += n; --n >= 0; ) wgt += tpl_info(*--tpls)->f;
  return wgt;                   /* sum weights and return this sum */
}  /* _sum_wgt() */

#endif
/*----------------------------------------------------------------------
  Comparison Functions
----------------------------------------------------------------------*/
#ifdef DT_GROW

static int _cmp_int (const void *p1, const void *p2, void *data)
{                               /* --- compare a column of two tuples */
  register int i1 = tpl_colval((TUPLE*)p1, (int)data)->i;
  register int i2 = tpl_colval((TUPLE*)p2, (int)data)->i;
  if     (i1 < i2) return -1;   /* get column values and */
  return (i1 > i2) ? 1 : 0;     /* return sign of their difference */
}  /* _cmp_int() */

/*--------------------------------------------------------------------*/

static int _cmp_flt (const void *p1, const void *p2, void *data)
{                               /* --- compare a column of two tuples */
  register float f1 = tpl_colval((TUPLE*)p1, (int)data)->f;
  register float f2 = tpl_colval((TUPLE*)p2, (int)data)->f;
  if     (f1 < f2) return -1;   /* get column values and */
  return (f1 > f2) ? 1 : 0;     /* return sign of their difference */
}  /* _cmp_flt() */

#endif
/*----------------------------------------------------------------------
  Auxiliary Functions for Growing and Pruning
----------------------------------------------------------------------*/

static TUPLE** _tuples (TABLE *table, int trgid, int *n)
{                               /* --- collect tuples w. known target */
  int   i;                      /* loop variable */
  int   type;                   /* type of target attribute */
  TUPLE **vec, **p, *tpl;       /* to traverse the tuple vector */
  INST  *inst;                  /* to access the target value */

  assert(table && n             /* check the function arguments */
      && (trgid >= 0) && (trgid < tab_colcnt(table)));
  i = tab_tplcnt(table);        /* get the number of tuples */
  p = vec = (TUPLE**)malloc(i *sizeof(TUPLE*));
  if (!p) return NULL;          /* allocate a tuple vector */
  type = att_type(tab_col(table, trgid));
  while (--i >= 0) {            /* traverse the tuples in the table */
    tpl  = tab_tpl(table, i);   /* and check the target value */
    inst = tpl_colval(tpl, trgid);
    if (((type == AT_SYM) && (inst->i <= UV_SYM))
    ||  ((type == AT_INT) && (inst->i <= UV_INT))
    ||  ((type == AT_FLT) && (inst->f <= UV_FLT)))
      continue;                 /* skip tuples with unknown target, */
    *p++ = tpl;                 /* but collect all other tuples */
    tpl_info(tpl)->f = tpl_getwgt(tpl);
  }                             /* copy the tuple weight */
  *n = (int)(p -vec);           /* compute the number of tuples */
  return vec;                   /* and return the tuple vector */
}  /* _tuples() */

/*--------------------------------------------------------------------*/

static DTNODE* _leaf_sym (DTREE *dt, FRQTAB *frqtab, int x)
{                               /* --- create a leaf (sym. target) */
  int    i, k;                  /* loop variable, class id */
  DTNODE *node;                 /* created leaf node */
  DTDATA *data;                 /* to traverse the data vector */
  double sum, frq;              /* (sum of) the class frequencies */
  double max = 0;               /* maximum of the class frequencies */
  double wgt;                   /* weight of cases with unknown value */

  assert(dt && frqtab           /* check the function arguments */
      && (ft_known(frqtab)    > 0)
      && (ft_frq_x(frqtab, x) > 0));
  node = (DTNODE*)malloc(sizeof(DTNODE) +sizeof(DTDATA)*(dt->clscnt-1));
  if (!node) return NULL;       /* create a leaf node */
  node->flags = DT_LEAF;        /* set the leaf flag */
  node->attid = dt->trgid;      /* store the class attribute id */
  node->size  = dt->clscnt;     /* and the number of classes */
  /* It is convenient to abuse the cut value for storing the   */
  /* frequency of tuples with a known value. This is possible, */
  /* because the cut value is not used in leaf nodes. Hence:   */
  sum = ft_frq_x(frqtab, x);    /* get and store the number of */
  node->cut = (float)sum;       /* tuples with a known value */
  frq = ft_frq_x(frqtab, -1);   /* check for unknown values */
  if (frq <= 0) {               /* if there are no unknown values */
    data = node->vec;           /* traverse the classes */
    for (i = k = 0; i < node->size; i++) {
      frq = ft_frq_xy(frqtab, x, i);
      if (frq > max) { max = frq; k = i; }
      (data++)->frq = (float)frq;  /* note the class frequencies and */
    } }                         /* determine the most frequent class */
  else {                        /* if there are unknown values */
    wgt = ft_known(frqtab);     /* compute the weighting factor */
    wgt = (wgt > 0) ? sum/wgt : (1.0/ft_xcnt(frqtab));
    sum += wgt *frq;            /* and add the weight of unknowns, */
    data = node->vec;           /* then traverse the classes */
    for (i = k = 0; i < node->size; i++) {
      frq = ft_frq_xy(frqtab, x, i) +wgt *ft_frq_xy(frqtab, -1, i);
      if (frq > max) { max = frq; k = i; }
      (data++)->frq = (float)frq;
    }                           /* compute the class frequencies and */
  }                             /* determine the most frequent class */
  node->frq   = (float)sum;     /* store the total frequency, */
  node->err   = sum -max;       /* the number of leaf errors, */
  node->trg.i = k;              /* and the most frequent class */
  return node;                  /* return the created leaf node */
}  /* _leaf_sym() */

/*--------------------------------------------------------------------*/

static DTNODE* _leaf_num (DTREE *dt, VARTAB *vartab, int x)
{                               /* --- create a leaf (num. target) */
  DTNODE *node;                 /* created leaf node */
  double frq, unk;              /* (sum of) the column frequencies */
  double wgt;                   /* weight of cases with unknown value */
  double m;                     /* temporary buffer */

  assert(dt && vartab           /* check the function arguments */
      && (vt_known (vartab)    > 0)
      && (vt_colfrq(vartab, x) > 0));
  node = (DTNODE*)malloc(sizeof(DTNODE) -sizeof(DTDATA));
  if (!node) return NULL;       /* create a leaf node */
  node->flags = DT_LEAF;        /* set the leaf flag */
  node->attid = dt->trgid;      /* store the target attribute id */
  node->size  = 0;              /* clear the vector size */
  /* Again the cut value is abused for storing the frequency */
  /* of tuples with a known value:                           */
  frq = vt_colfrq(vartab, x);   /* get and store the number of */
  node->cut = (float)frq;       /* tuples with a known value */
  unk = vt_colfrq(vartab, -1);  /* check for unknown values */
  if (unk <= 0) {               /* if there are no unknown values, */
    node->frq   = (float)frq;   /* copy the aggregates directly */
    node->err   = vt_colsse(vartab, x);
    node->trg.f = (float)vt_colmean(vartab, x); }
  else {                        /* if there are unknown values, */
    wgt = vt_known(vartab);     /* compute the weighting factor */
    unk *= wgt = (wgt > 0) ? frq/wgt : (1.0/vt_colcnt(vartab));
    node->err   = vt_colssv(vartab, x) +wgt *vt_colssv(vartab, -1);
    node->frq   = (float)(wgt = frq +unk);
    node->trg.f = (float)(m   = (frq *vt_colmean(vartab, x)
                                +unk *vt_colmean(vartab,-1)) /wgt);
    node->err  -= wgt *m*m;     /* combine the column aggregates by */
  }                             /* recomputing mean and sum of sq.e. */
  return node;                  /* return the created leaf node */
}  /* _leaf_num() */

/*--------------------------------------------------------------------*/

void _delete (DTNODE *node)
{                               /* --- recursively delete a subtree */
  int    i;                     /* loop variable */
  DTDATA *data;                 /* to traverse the data vector */

  assert(node);                 /* check the function argument */
  if (!(node->flags & DT_LEAF)){/* if the node is no leaf */
    for (data = node->vec +(i = node->size); --i >= 0; )
      if ((--data)->child && !islink(data, node))
        _delete(data->child);   /* recursively delete all subtrees, */
  }                             /* avoiding empty subtrees and links */
  free(node);                   /* delete the node itself */
}  /* _delete() */

/*--------------------------------------------------------------------*/
#ifdef DT_PRUNE

static DTNODE* _adapt (ATTSET *attset, DTNODE *node)
{                               /* --- adapt test node to new values */
  int    i, n;                  /* loop variable, number of values */
  DTNODE *new;                  /* adapted test node */
  DTDATA *data, *tmp;           /* to traverse data vector, buffer */

  assert(attset && node);       /* check the function arguments */
  n = att_valcnt(as_att(attset, node->attid));
  if (node->size >= n)          /* get number of values and check the */
    return node;                /* node's data vector size against it */
  tmp = node->vec;              /* note the old data vector */
  new = (DTNODE*)realloc(node, sizeof(DTNODE) +(n-1) *sizeof(DTDATA));
  if (!new) return NULL;        /* create a new test node */
  if ((new != node)             /* if the node has been changed */
  &&  (new->flags & DT_LINK)) { /* and it contains links */
    for (data = new->vec +(i = new->size); --i >= 0; ) {
      --data;                   /* traverse the data vector */
      if ((data->link >= tmp)   /* if a pointer is a link */
      &&  (data->link <  tmp +new->size))
        data->link = new->vec +(data->link -tmp);
    }                           /* adapt the link, i.e., compute the */
  }                             /* address of the new vector field */
  data = new->vec +n;           /* enlarge node and clear pointers */
  do { (--data)->child = NULL; } while (++new->size < n);
  return new;                   /* return the enlarged node */
}  /* _adapt() */

/*--------------------------------------------------------------------*/

static DTNODE* _coll_rec (DTNODE *node, double *frq)
{                               /* --- recursive part of _collapse */
  int    i;                     /* loop variable */
  DTNODE *res, *tmp;            /* resulting leaf node */
  DTDATA *data;                 /* to traverse the data vector */

  assert(node && frq);          /* check the function arguments */
  if (node->flags & DT_LEAF) {  /* if the node is a leaf */
    frq += node->size;          /* traverse the frequency buffer */
    for (data = node->vec +(i = node->size); --i >= 0; )
      *--frq += (--data)->frq;  /* aggregate the class frequencies */
    return node;                /* and return the leaf node */
  }                             /* as a possible result node */
  res = NULL;                   /* traverse the child nodes */
  for (data = node->vec +(i = node->size); --i >= 0; ) {
    if (!(--data)->child || islink(data, node))
      continue;                 /* skip empty nodes and links */
    tmp = _coll_rec(data->child, frq);
    if (!res) res = tmp;        /* recursively aggregate */
    else      free(tmp);        /* the class frequencies */
  }                             /* and keep one leaf node */
  free(node);                   /* delete the current node */
  return res;                   /* return the leaf node kept */
}  /* _coll_rec() */

/*--------------------------------------------------------------------*/

static DTNODE* _collapse (DTREE *dt, DTNODE *node)
{                               /* --- collapse subtree into a leaf */
  int    i;                     /* loop variable */
  DTDATA *data;                 /* to traverse the data vector */
  double *frq, *max;            /* to traverse the class frequencies */
  double sum;                   /* sum of the class frequencies */

  assert(dt && node);           /* check the function arguments */
  if (!node || (node->flags & DT_LEAF))
    return node;                /* if the node is a leaf, abort */

  /* --- numeric target attribute --- */
  if (dt->type != AT_SYM) {     /* if the target att. is numeric */
    for (data = node->vec +(i = node->size); --i >= 0; )
      if ((--data)->child && !islink(data, node))
        _delete(data->child);   /* delete all subtrees of the node */
    node->flags = DT_LEAF;      /* turn the node into a leaf, */
    node->attid = dt->trgid;    /* store the target identifier, */
    node->size  = 0;            /* and clear the vector size */
    return (DTNODE*)realloc(node, sizeof(DTNODE) -sizeof(DTDATA));
  }                             /* shrink the node to leaf size */

  /* --- symbolic target attribute --- */
  for (frq = dt->buf +(i = dt->clscnt); --i >= 0; )
    *--frq = 0;                 /* clear the class frequencies */
  node = _coll_rec(node, frq);  /* collapse the subtree into a leaf */
  if (!node) return NULL;       /* if no result leaf found, abort */
  max = frq = dt->buf; sum = *frq;
  for (i = dt->clscnt; --i > 0; ) {
    sum += *++frq;              /* traverse the class frequencies, */
    if (*frq > *max) max = frq; /* compute the frequency sum, and */
  }                             /* determine the most frequent class */
  frq = dt->buf +node->size;    /* traverse the counter vector */
  for (data = node->vec +(i = node->size); --i >= 0; )
    (--data)->frq = (float)*--frq; /* copy the class frequencies */
  node->frq   = (float)sum;        /* set the node frequency, */
  node->err   = sum -*max;         /* the number of errors, and */
  node->trg.i = (int)(max-frq);    /* the most frequent class */
  return node;                  /* return the modified leaf node */
}  /* _collapse() */

#endif  /* #ifdef DT_PRUNE */
/*--------------------------------------------------------------------*/

static int _count (DTREE *dt, DTNODE *node, int height)
{                               /* --- count the number of nodes */
  int    i, n = 1;              /* loop variable, node counter */
  DTDATA *data;                 /* to traverse the data vector */

  assert(dt && node && (height >= 0));  /* check the function args. */
  if (++height > dt->height)    /* if beyond the curr. max. height, */
    dt->height = height;        /* update the maximal height */
  if (!(node->flags & DT_LEAF)){/* if the current node is no leaf */
    for (data = node->vec +(i = node->size); --i >= 0; )
      if ((--data)->child && !islink(data, node))
        n += _count(dt, data->child, height);
  }                             /* recursively count the nodes */
  return n;                     /* return the number of nodes */
}  /* _count() */

/*----------------------------------------------------------------------
  Evaluation Functions for Symbolic Target Attributes
----------------------------------------------------------------------*/
#ifdef DT_GROW

static double _sym_sym (GROW *gi, TUPLE **tpls, int n,
                        int attid, float *cut)
{                               /* --- evaluate a symbolic attribute */
  int i, k;                     /* class attribute id, loop variable */
  int cls, val;                 /* class and attribute value */

  assert(gi && tpls && (n > 0));/* check the function arguments */
  k = att_valcnt(as_att(gi->attset, attid));
  ft_init((FRQTAB*)gi->curr, k, gi->dtree->clscnt);
  i = gi->dtree->trgid;         /* initialize the frequency table */
  for (tpls += n; --n >= 0; ) { /* and traverse the tuples */
    cls = tpl_colval(*--tpls, i)->i;
    val = tpl_colval(*tpls, attid)->i;
    ft_add((FRQTAB*)gi->curr, val, cls, tpl_info(*tpls)->f);
  }                             /* fill the frequency table */
  ft_marg(gi->curr);            /* and marginalize it */
  if (gi->flags & DT_SUBSET)    /* if only called to compute the */
    return 0;                   /* initial table, abort the function */
  for (i = 0; --k >= 0; )       /* count reasonable tuple sets */
    if (ft_frq_x((FRQTAB*)gi->curr, k) >= gi->mincnt)
      i++;                      /* there must be at least two values */
  if (i < 2) return WORTHLESS;  /* with at least _mincnt cases each */
  return ft_eval((FRQTAB*)gi->curr, gi->measure, gi->params);
}  /* _sym_sym() */             /* evaluate the frequency table */

/*--------------------------------------------------------------------*/

static double _sym_set (GROW *gi, TUPLE **tpls, int n,
                        int attid, float *cut)
{                               /* -- evaluate subsets of sym. att. */
  int    i, k, cls;             /* loop variables */
  int    valcnt;                /* number of values/subsets */
  int    rtscnt;                /* number of reasonable tuple sets */
  int    *sets;                 /* to traverse subset index vector */
  int    isrc, idst, imax;      /* indices     of sets to combine */
  double fsrc, fdst, fmax;      /* frequencies of sets to combine */
  double curr, prev, best;      /* current/previous/best worth */

  assert(gi && tpls && (n > 0));     /* check the function arguments */
  _sym_sym(gi, tpls, n, attid, cut); /* compute initial freq. table */

  /* --- remove unsupported and merge single class values --- */
  valcnt = att_valcnt(as_att(gi->attset, attid));
  imax   = rtscnt = 0;          /* initialize the variables and */
  fmax   = 0; sets = gi->sets;  /* traverse the set index vector */
  for (i = 0; i < valcnt; i++){ /* and the attribute values */
    if (ft_dest((FRQTAB*)gi->curr, i) != i)
      continue;                 /* skip linked/combined values */
    fdst = ft_frq_x((FRQTAB*)gi->curr, i);
    if (fdst <= 0) continue;    /* skip unsupported attribute values */
    *sets++ = i;                /* note value index in index vector */
    for (cls = gi->dtree->clscnt; --cls >= 0; ) {
      fsrc = ft_frq_xy((FRQTAB*)gi->curr, i, cls);
      if (fsrc < fdst *(1-EPSILON)) { /* if no single class set, */
        if (fsrc > EPSILON) break;    /* if class is supp., abort */
        else continue;                /* otherwise continue */
      }                               /* with the next class */
      for (k = i; ++k < valcnt; ) {   /* if single class tuple set, */
        fsrc = ft_frq_x((FRQTAB*)gi->curr, k); 
        if (fsrc <= 0) continue;      /* traverse remaining values */
        if (ft_frq_xy((FRQTAB*)gi->curr, k, cls) >= fsrc *(1-EPSILON)) {
          ft_comb((FRQTAB*)gi->curr, k, i); fdst += fsrc; }
      } break;                  /* if second single class set with */
    }                           /* same class found, combine values */
    if (fdst > fmax) { fmax = fdst; imax = i; }
    if (fdst >= gi->mincnt) rtscnt++;
  }                             /* count reasonable tuple sets */
  valcnt = (int)(sets -gi->sets);
  sets   = gi->sets;            /* compute new number of values */

  /* --- find best pair mergers --- */
  prev = ft_eval((FRQTAB*)gi->curr, gi->measure, gi->params);
  if (valcnt <= 2)              /* check whether a merge is possible */
    return (rtscnt >= 2) ? prev : WORTHLESS;
  while (1) {                   /* select and merge loop */
    best = WORTHLESS;           /* initialize worth of partition */
    isrc = idst = 0;            /* and value set indices */
    for (i = 0; i < valcnt; i++) {  /* traverse the value sets */
      if ((rtscnt <= 2)         /* if less than two reasonable sets, */
      &&  (i == imax)) continue;    /* skip mergers with largest set */
      for (k = i+1; k < valcnt; k++) {    /* traverse remaining sets */
        if ((rtscnt <= 2)       /* if less than two reasonable sets, */
        &&  (k == imax)) continue;  /* skip mergers with largest set */
        ft_comb((FRQTAB*)gi->curr, sets[k], sets[i]);/* combine sets */
        curr = ft_eval((FRQTAB*)gi->curr, gi->measure, gi->params);
        ft_uncomb((FRQTAB*)gi->curr, sets[k]); /* evaluate partition, */
        if (curr <= best) continue;            /* then uncombine sets */
        best = curr;            /* if the current partition is better */
        isrc = k;               /* than the best one found up to now, */
        idst = i;               /* note worth and value set indices */
      }                         /* of the current pair merger */
    }
    if ((best <= WORTHLESS)     /* if no useful partition found */
    ||  (best <  prev))         /* or best partition found is worse */
      break;                    /* than the previous one, abort loop */
    prev = best;                /* note the best worth */
    fsrc = ft_frq_x((FRQTAB*)gi->curr, sets[isrc]);
    if (fsrc >= gi->mincnt)     /* get the source frequency */
      rtscnt--;                 /* and update reasonable tuple sets */
    fdst = ft_frq_x((FRQTAB*)gi->curr, sets[idst]);
    if (fdst >= gi->mincnt)     /* get the destination frequency */
      rtscnt--;                 /* and update reasonable tuple sets */
    ft_comb(gi->curr, sets[isrc], sets[idst]);
    fdst = ft_frq_x((FRQTAB*)gi->curr, sets[idst]);
    if (fdst >= gi->mincnt)     /* combine sets, get new frequency, */
      rtscnt++;                 /* and update reasonable tuple sets */
    if (fdst >= fmax) { fmax = fdst; imax = idst; }
    if (imax > isrc) imax--;    /* adapt vector index of largest set */
    for (i = isrc; i < valcnt-1; i++)
      sets[i] = sets[i+1];      /* remove source set from vector */
    valcnt--;                   /* (shift to preserve value order) */
  }                             /* and reduce number of values */
  if (rtscnt < 2) return WORTHLESS;
  return prev;                  /* return the evaluation result */
}  /* _sym_set() */

/*--------------------------------------------------------------------*/

static double _sym_num (GROW *gi, TUPLE **tpls, int n,
                        int attid, float *cut)
{                               /* --- evaluate a numeric attribute */
  int    i, k;                  /* loop variable, class att. index */
  int    type;                  /* type of the test attribute */
  int    cls;                   /* class of current tuple */
  INST   *val, *pval;           /* value of current and prec. tuple */
  double wgt,  sum = 0;         /* tuple weight, sum of tuple weights */
  double curr, best;            /* current and best cut worth */

  assert(gi && tpls && (n > 0) && cut);   /* check the function args. */

  /* --- build initial frequency table --- */
  type = att_type(as_att(gi->attset, attid));
  v_sort(tpls, n, (type == AT_FLT) ? _cmp_flt : _cmp_int, (void*)attid);
  ft_init((FRQTAB*)gi->numt, 2, gi->dtree->clscnt);
  k = gi->dtree->trgid;         /* sort tuples and initialize table */
  while (1) {                   /* traverse tuples with unknown value */
    if (n <= 1)                 /* if there is at most one tuple */
      return WORTHLESS;         /* with a known att. value, abort */
    val = tpl_colval(*tpls, attid);  /* get the next value */
    if ((type == AT_INT) ? (val->i > UV_INT)
    :                      (val->f > UV_FLT))
      break;                    /* if the value is known, abort loop */
    n--;                        /* go to the next tuple */
    cls = tpl_colval(*tpls, k)->i;   /* get the class */
    wgt = tpl_info  (*tpls++)->f;    /* and the tuple weight */
    ft_add((FRQTAB*)gi->numt, -1, cls, wgt);
  }                             /* store the tuple weight */
  do {                          /* traverse tuples with known value */
    if (--n <= 0)               /* if there is no tuple left for */
      return WORTHLESS;         /* the other side of the cut, abort */
    cls = tpl_colval(*tpls, k)->i;   /* get the class */
    wgt = tpl_info  (*tpls++)->f;    /* and the tuple weight */
    ft_add((FRQTAB*)gi->numt, 0, cls, wgt);
    sum += wgt;                 /* store the tuple weight */
  } while (sum < gi->mincnt);   /* while minimal size not reached */
  for (i = n; --i >= 0; ) {     /* traverse the remaining tuples */
    cls = tpl_colval(tpls[i], k)->i; /* get the class */
    wgt = tpl_info  (tpls[i])->f;    /* and the tuple weight */
    ft_add((FRQTAB*)gi->numt, 1, cls, wgt);
  }                             /* store the tuple weight */
  ft_marg(gi->numt);            /* marginalize frequency table and */
  sum = ft_frq_x((FRQTAB*)gi->numt,1); /* get weight of upper part */
  if (sum < gi->mincnt) return WORTHLESS;

  /* --- find the best cut value --- */
  pval = tpl_colval(*(tpls-1), attid);
  best = WORTHLESS;             /* note the class and the value and */
  *cut = UV_FLT;                /* init. the worth and the cut value */
  while (1) {                   /* traverse the tuples above the cut */
    val = tpl_colval(*tpls, attid);  /* if next value differs */
    if ((type == AT_INT) ? (val->i > pval->i)
    :     (0.5F *(pval->f +val->f) > pval->f)) {
      curr = ft_eval((FRQTAB*)gi->numt, gi->measure, gi->params);
      if (curr >= best) {       /* evaluate the frequency table, */
        best = curr;            /* update the best cut worth and */
        ft_copy((FRQTAB*)gi->curr, gi->numt);  /* copy the table */
        *cut = (type == AT_INT) /* depending on the attribute type */
             ? (0.5F *((float)val->i +pval->i))
             : (0.5F *(       val->f +pval->f));
      }                         /* compute the cut value as the */
    }                           /* average of the adjacent values */
    if (--n <= 0) break;        /* if on the last tuple, abort loop */
    sum -= wgt = tpl_info(*tpls)->f;  /* get and sum the tuple weight */
    if (sum < gi->mincnt)       /* if the weight of the remaining */
      break;                    /* tuples is too small, abort loop */
    cls = tpl_colval(*tpls, k)->i;
    ft_move((FRQTAB*)gi->numt, 1, 0, cls, wgt);
    pval = val; tpls++;         /* move weight in frequency table, */
  }                             /* note the value, and get next tuple */
  return best;                  /* return worth of best cut */
}  /* _sym_num() */

/*----------------------------------------------------------------------
  Evaluation Functions for Numeric Target Attributes
----------------------------------------------------------------------*/

static double _num_sym (GROW *gi, TUPLE **tpls, int n,
                        int attid, float *cut)
{                               /* --- evaluate a symbolic attribute */
  int    i, k;                  /* class attribute id, loop variable */
  INST   *inst;                 /* instance of the target attribute */
  double val;                   /* value of candidate test attribute */
  int    vid;                   /* identifier of attribute value */

  assert(gi && tpls && (n > 0));/* check the function arguments */
  k = att_valcnt(as_att(gi->attset, attid));
  vt_init((VARTAB*)gi->curr, k);
  i = gi->dtree->trgid;         /* initialize the variation table */
  for (tpls += n; --n >= 0; ) { /* and traverse the tuples */
    inst = tpl_colval(*--tpls, i);
    val  = (double)((gi->type == AT_INT) ? inst->i : inst->f);
    vid  = tpl_colval(*tpls, attid)->i;
    vt_add((VARTAB*)gi->curr, vid, val, tpl_info(*tpls)->f);
  }                             /* fill the variation table */
  vt_calc(gi->curr);            /* and calculate aggregates */
  if (gi->flags & DT_SUBSET)    /* if only called to compute the */
    return 0;                   /* initial table, abort the function */
  for (i = 0; --k >= 0; )       /* count reasonable tuple sets */
    if (vt_colfrq((VARTAB*)gi->curr, k) >= gi->mincnt)
      i++;                      /* there must be at least two values */
  if (i < 2) return WORTHLESS;  /* with at least _mincnt cases each */
  return vt_eval((VARTAB*)gi->curr, gi->measure, gi->params);
}  /* _num_sym() */             /* evaluate the variation table */

/*--------------------------------------------------------------------*/

static double _num_set (GROW *gi, TUPLE **tpls, int n,
                        int attid, float *cut)
{                               /* -- evaluate subsets of sym. att. */
  int    i, k;                  /* loop variables */
  int    valcnt;                /* number of values/subsets */
  int    rtscnt;                /* number of reasonable tuple sets */
  int    *sets;                 /* to traverse subset index vector */
  int    isrc, idst, imax;      /* indices     of sets to combine */
  double fsrc, fdst, fmax;      /* frequencies of sets to combine */
  double curr, prev, best;      /* current/previous/best worth */

  assert(gi && tpls && (n > 0));     /* check the function arguments */
  _num_sym(gi, tpls, n, attid, cut); /* compute initial freq. table */

  /* --- collect supported values --- */
  valcnt = att_valcnt(as_att(gi->attset, attid));
  imax   = rtscnt = 0;          /* initialize the variables and */
  fmax   = 0; sets = gi->sets;  /* traverse the set index vector */
  for (i = 0; i < valcnt; i++){ /* and the attribute values */
    if (vt_dest((VARTAB*)gi->curr, i) != i)
      continue;                 /* skip linked/combined values */
    fdst = vt_colfrq((VARTAB*)gi->curr, i);
    if (fdst <= 0) continue;    /* skip unsupported attribute values */
    *sets++ = i;                /* note value index in index vector */
    if (fdst > fmax) { fmax = fdst; imax = i; }
    if (fdst >= gi->mincnt) rtscnt++;
  }                             /* count reasonable tuple sets */
  valcnt = (int)(sets -gi->sets);
  sets   = gi->sets;            /* compute new number of values */

  /* --- find best pair mergers --- */
  prev = vt_eval((VARTAB*)gi->curr, gi->measure, gi->params);
  if (valcnt <= 2)              /* check whether a merge is possible */
    return (rtscnt >= 2) ? prev : WORTHLESS;
  while (1) {                   /* select and merge loop */
    best = WORTHLESS;           /* initialize worth of partition */
    isrc = idst = 0;            /* and value set indices */
    for (i = 0; i < valcnt; i++) {  /* traverse the value sets */
      if ((rtscnt <= 2)         /* if less than two reasonable sets, */
      &&  (i == imax)) continue;    /* skip mergers with largest set */
      for (k = i+1; k < valcnt; k++) {    /* traverse remaining sets */
        if ((rtscnt <= 2)       /* if less than two reasonable sets, */
        &&  (k == imax)) continue;  /* skip mergers with largest set */
        vt_comb((VARTAB*)gi->curr, sets[k], sets[i]);/* combine sets */
        curr = vt_eval(gi->curr, gi->measure, gi->params);
        vt_uncomb((VARTAB*)gi->curr, sets[k]); /* evaluate partition, */
        if (curr <= best) continue;            /* then uncombine sets */
        best = curr;            /* if the current partition is better */
        isrc = k;               /* then the best one found up to now, */
        idst = i;               /* note worth and value set indices */
      }                         /* of the current pair merger */
    }
    if ((best <= WORTHLESS)     /* if no useful partition found */
    ||  (best <  prev))         /* or best partition found is worse */
      break;                    /* than the previous one, abort loop */
    prev = best;                /* note the best worth */
    fsrc = vt_colfrq((VARTAB*)gi->curr, sets[isrc]);
    if (fsrc >= gi->mincnt)     /* get the source frequencies */
      rtscnt--;                 /* and update reasonable tuple sets */
    fdst = vt_colfrq((VARTAB*)gi->curr, sets[idst]);
    if (fdst >= gi->mincnt)     /* get the destination frequency */
      rtscnt--;                 /* and update reasonable tuple sets */
    vt_comb(gi->curr, sets[isrc], sets[idst]);
    fdst = vt_colfrq((VARTAB*)gi->curr, sets[idst]);
    if (fdst >= gi->mincnt)     /* combine sets, get new frequency, */
      rtscnt++;                 /* and update reasonable tuple sets */
    if (fdst > fmax) { fmax = fdst; imax = idst; }
    if (imax > isrc) imax--;    /* adapt vector index of largest set */
    for (i = isrc; i < valcnt-1; i++)
      sets[i] = sets[i+1];      /* remove source set from vector */
    valcnt--;                   /* (shift to preserve value order) */
  }                             /* and reduce number of values */
  if (rtscnt < 2) return WORTHLESS;
  return prev;                  /* return the evaluation result */
}  /* _num_set() */

/*--------------------------------------------------------------------*/

static double _num_num (GROW *gi, TUPLE **tpls, int n,
                        int attid, float *cut)
{                               /* --- evaluate a numeric attribute */
  int    i, k;                  /* loop variable, target att. index */
  int    type, t;               /* type of the test/target attribute */
  INST   *inst;                 /* target instance of current tuple */
  double trg;                   /* target value    of current tuple */
  INST   *val, *pval;           /* value of current and prec. tuple */
  double wgt,  sum = 0;         /* tuple weight, sum of tuple weights */
  double curr, best;            /* current and best cut worth */

  assert(gi && tpls && (n > 0) && cut);   /* check the function args. */

  /* --- build initial variation table --- */
  type = att_type(as_att(gi->attset, attid));
  v_sort(tpls, n, (type == AT_FLT) ? _cmp_flt : _cmp_int, (void*)attid);
  vt_init((VARTAB*)gi->numt,2); /* sort tuples and initialize table */
  k = gi->dtree->trgid;         /* note the target attribute index */
  t = gi->dtree->type;          /* and the target type */
  while (1) {                   /* traverse tuples with unknown value */
    if (n <= 1)                 /* if there is at most one tuple */
      return WORTHLESS;         /* with a known att. value, abort */
    val = tpl_colval(*tpls, attid);   /* get the next value */
    if ((type == AT_INT) ? (val->i > UV_INT)
    :                      (val->f > UV_FLT))
      break;                    /* if the value is known, abort loop */
    n--;                        /* go to the next tuple */
    inst = tpl_colval(*tpls, k);
    trg  = (double)((t == AT_INT) ? inst->i : inst->f);
    wgt  = tpl_info(*tpls++)->f;/* get the target value */
    vt_add((VARTAB*)gi->numt, -1, trg, wgt);
  }                             /* store the tuple weight */
  do {                          /* traverse tuples with known value */
    if (--n <= 0)               /* if there is no tuple left for */
      return WORTHLESS;         /* the other side of the cut, abort */
    inst = tpl_colval(*tpls, k);
    trg  = (double)((t == AT_INT) ? inst->i : inst->f);
    wgt  = tpl_info(*tpls++)->f;      /* get the target value and */
    vt_add(gi->numt, 0, trg, wgt);    /* store the tuple weight */
    sum += wgt;                       /* in the variation table */
  } while (sum < gi->mincnt);   /* while minimal size not reached */
  for (i = n; --i >= 0; ) {     /* traverse the remaining tuples */
    inst = tpl_colval(tpls[i], k);
    trg  = (double)((t == AT_INT) ? inst->i : inst->f);
    wgt  = tpl_info(tpls[i])->f;/* get the target value */
    vt_add((VARTAB*)gi->numt, 1, trg, wgt);    
  }                             /* store the tuple weight */
  vt_calc((VARTAB*)gi->numt);   /* calculate aggregates and get */
  sum = vt_colfrq((VARTAB*)gi->numt,1); /* weight of upper part */
  if (sum < gi->mincnt) return WORTHLESS;

  /* --- find the best cut value --- */
  pval = tpl_colval(*(tpls-1), attid);
  best = WORTHLESS;             /* note the class and the value and */
  *cut = UV_FLT;                /* init. the worth and the cut value */
  while (1) {                   /* traverse the tuples above the cut */
    val = tpl_colval(*tpls, attid);  /* if next value differs */
    if ((type == AT_INT) ? (val->i > pval->i)
    :     (0.5F *(pval->f +val->f) > pval->f)) {
      curr = vt_eval((VARTAB*)gi->numt, gi->measure, gi->params);
      if (curr >= best) {       /* evaluate the variation table, */
        best = curr;            /* update the best cut worth, and */
        vt_copy((VARTAB*)gi->curr, gi->numt);   /* copy the table */
        *cut = (type == AT_INT) /* depending on the attribute type */
             ? (0.5F *((float)val->i +pval->i))
             : (0.5F *(       val->f +pval->f));
      }                         /* compute the cut value as the */
    }                           /* average of the adjacent values */
    if (--n <= 0) break;        /* if on the last tuple, abort loop */
    sum -= wgt = tpl_info(*tpls)->f;  /* get and sum the tuple weight */
    if (sum < gi->mincnt)       /* if the weight of the remaining */
      break;                    /* tuples is too small, abort loop */
    inst = tpl_colval(*tpls, k);
    trg  = (double)((t == AT_INT) ? inst->i : inst->f);
    vt_move((VARTAB*)gi->numt, 1, 0, trg, wgt);
    pval = val; tpls++;         /* move weight in variation table, */
  }                             /* note the value, and get next tuple */
  return best;                  /* return worth of best cut */
}  /* _num_num() */

/*----------------------------------------------------------------------
  Growing Functions
----------------------------------------------------------------------*/

static DTNODE* _grow (GROW *gi, DTNODE *leaf, TUPLE **tpls, int n)
{                               /* --- recursively grow tree */
  int    i, k;                  /* loop variables, buffers */
  int    mark;                  /* buffer for attribute mark */
  ATT    *att;                  /* current/test attribute */
  double curr, best;            /* current and best worth */
  float  cut;                   /* cut value for best attribute */
  void   *tab;                  /* exchange buffer for eval. tables */
  DTNODE *node;                 /* created test node (subtree) */
  DTDATA *data;                 /* to traverse the data vector */
  TSEL   tsel;                  /* tuple selection information */
  GRPFN  *group;                /* tuple grouping function */
  double frq, known;            /* tuple frequencies for weighting */
  double e_tree;                /* sum of subtree errors */

  assert(leaf && tpls && (n > 0)); /* check the function arguments */

  /* --- check the leaf node --- */
  gi->err = leaf->err;          /* set the number of leaf errors */
  if ((leaf->frq < 2*gi->mincnt)/* if too few tuples to divide */
  ||  (leaf->err < MINERROR)    /* or all cases have the same value */
  ||  (gi->maxht <= 0))         /* or maximal tree height reached, */
    return leaf;                /* simply return the leaf node */

  /* --- search for a test attribute --- */
  best       = WORTHLESS;       /* clear worth of best attribute, */
  tsel.col   = -1;              /* identifier of best attribute, */
  tsel.val.f = cut = UV_FLT;    /* and cut value (numeric atts.) */
  k = as_attcnt(gi->attset);    /* get the number of attributes */
  for (i = 0; i < k; i++) {     /* traverse the attributes, */
    att = as_att(gi->attset,i); /* but skip marked attributes */
    if (att_getmark(att) < 0) continue;
    if (att_type(att) == AT_SYM) curr = gi->eval_sym(gi, tpls,n,i,NULL);
    else                         curr = gi->eval_num(gi, tpls,n,i,&cut);
    if (curr <= best) continue; /* evaluate attribute (compute worth) */
    best       = curr;          /* if the current worth is better */
    tsel.col   = i;             /* than that of the best attribute, */
    tsel.val.f = cut;           /* note worth, attribute id, and cut */
    tab        = gi->best;      /* exchange the freq./var. tables, */
    gi->best   = gi->curr;      /* so that the current table */
    gi->curr   = tab;           /* becomes the best table */
  }                             /* (much more efficient than copying) */
  if ((tsel.col < 0)            /* if no test attribute found */
  ||  (best     < gi->minval))  /* or the best attribute is not */
    return leaf;                /* good enough, abort the function */

  /* --- create a test node --- */
  att  = as_att(gi->attset, tsel.col);
  i    = (att_type(att) == AT_SYM) ? att_valcnt(att) : 2;
  node = (DTNODE*)malloc(sizeof(DTNODE) +(i-1) *sizeof(DTDATA));
  if (!node) { gi->err = -1; return leaf; }
  node->flags = 0;              /* create a test node */
  node->attid = tsel.col;       /* set attribute id */
  node->cut   = tsel.val.f;     /* and cut value */
  node->frq   = leaf->frq;      /* copy the node frequency, */
  node->err   = leaf->err;      /* the number of errors, */
  node->trg   = leaf->trg;      /* and the target value */
  node->size  = i;              /* set the data vector size */

  /* --- create child nodes --- */
  for (data = node->vec +(k = node->size); --k >= 0; ) {
    if (gi->type == AT_SYM) i = ft_dest(gi->best, k);
    else                    i = vt_dest(gi->best, k);
    if (i == k)                 /* if the value is not linked, */
      (--data)->link = NULL;    /* clear the link pointer */
    else {                      /* if the value is linked */
      (--data)->link = node->vec +i;
      node->flags |= DT_LINK;   /* set a link to the dest. value */
    }                           /* and set the flag for node links */
  }                             /* (init. pointers for cleanup) */
  if (gi->type == AT_SYM) {     /* if the target is symbolic */
    for (data += k = node->size; --k >= 0; ) {
      if (!(--data)->link       /* traverse all uncombined values */
      &&  (ft_frq_x((FRQTAB*)gi->best, k) >= EPSILON)) {
        data->child = _leaf_sym(gi->dtree, (FRQTAB*)gi->best, k);
        if (!data->child) { _delete(node); gi->err = -1; return leaf; }
      }                         /* create a leaf node */
    }                           /* for all supported values */
    frq = ft_known((FRQTAB*)gi->best); }
  else {                        /* if the target is numeric */
    for (data += k = node->size; --k >= 0; ) {
      if (!(--data)->link       /* traverse all uncombined values */
      &&  (vt_colfrq((VARTAB*)gi->best, k) >= EPSILON)) {
        data->child = _leaf_num(gi->dtree, (VARTAB*)gi->best, k);
        if (!data->child) { _delete(node); gi->err = -1; return leaf; }
      }                         /* create a leaf node */
    }                           /* for all supported values */
    frq = vt_known((VARTAB*)gi->best);
  }                             /* note the sum of the weights of */
  known  = frq;                 /* tuples with a known att. value */
  e_tree = 0;                   /* clear the number of errors */
  gi->maxht--;                  /* reduce the maximal tree height */

  /* --- branch on symbolic attribute --- */
  if (att_type(att) == AT_SYM){ /* if the test attribute is symbolic */
    mark = att_getmark(att);    /* mark attribute (to prevent tests) */
    if (!(gi->flags & DT_SUBSET)) att_setmark(att, -1);
    tsel.val.i = UV_SYM;        /* group tuples with */
    i = _grp_sym(tpls,n,&tsel); /* an unknown attribute value */
    group    = (node->flags & DT_LINK) ? _grp_set : _grp_sym;
    frq      = known;           /* get the denominator of the weight */
    tsel.vec = data;            /* traverse the attribute values */
    for (data += tsel.val.i = node->size; --tsel.val.i >= 0; ) {
      if (!(--data)->child      /* if an att. value is not supported */
      ||  islink(data, node))   /* or combined with another value, */
        continue;               /* skip the attribute value */
      k = group(tpls+i, n-i, &tsel); /* group tuples with next value */
      if (i > 0) {              /* if there are unknown values */
        _mul_wgt(tpls, i, data->child->cut/frq);
        frq = data->child->cut; /* weight tuples with unknown value, */
      }                         /* note the denom. for reweighting */
      data->child = _grow(gi, data->child, tpls, k+i);
      if (gi->err < 0) { _delete(node); return leaf; }
      e_tree += gi->err;        /* grow a leaf/subtree for the value */
      if (i > 0)                /* if there are unknown values, */
        group(tpls,k+i, &tsel); /* regroup tuples with current value */
      tpls += k; n -= k;        /* and skip processed tuples */
    }                           
    att_setmark(att, mark);     /* restore mark of test attribute */
    if (i > 0) _mul_wgt(tpls, i, known/frq);
  }                             /* reweight tuples with unknowns */

  /* --- branch on numeric attribute --- */
  else {                        /* if the test attribute is numeric */
    if (att_type(att) == AT_FLT) {              group = _grp_flt; }
    else { tsel.val.i = (int)floor(tsel.val.f); group = _grp_int; }
    k = group(tpls, n, &tsel);  /* group tuples greater than the cut */
    group = (att_type(att) == AT_FLT) ? _grp_uvflt : _grp_uvint;
    i = group(tpls+k,n-k,&tsel);/* group tuples with unknown value */
    if (i > 0) {                /* if there are unknown values */
      _mul_wgt(tpls+k, i, data[0].child->cut/known);
      frq = data[0].child->cut; /* weight tuples with unknown value */
    }                           /* and note freq. for reweighting */
    data[0].child = _grow(gi, data[0].child, tpls+k, n-k);
    if (gi->err < 0) { _delete(node); return leaf; }
    e_tree += gi->err;          /* grow a leaf/subtree for <= cut */
    if (i > 0) {                /* if there are unknown values, */
      group(tpls+k,n-k, &tsel); /* regroup tuples with unknown value */
      _mul_wgt(tpls+k, i, data[1].child->cut/frq);
      frq = data[1].child->cut; /* weight tuples with unknown value */
    }                           /* and note freq. for reweighting */
    data[1].child = _grow(gi, data[1].child, tpls, k+i);
    if (gi->err < 0) { _delete(node); return leaf; }
    e_tree += gi->err;          /* grow a leaf/subtree for > cut */
    if (i > 0) {                /* if there are unknown values, */
      group(tpls, k+i, &tsel);  /* regroup tuples with unknowns */
      _mul_wgt(tpls, i, known/frq);
    }                           /* reweight tuples with unknown value */
  }
  gi->maxht++;                  /* restore maximal (sub)tree height */

  /* --- test subtree against leaf --- */
  if (leaf->err < e_tree *(1+EPSILON)) {
    _delete(node);              /* if the leaf is not worse */
    gi->err = leaf->err;        /* than the subtree, */
    return leaf;                /* discard the subtree */
  }                             /* and return the leaf */
  free(leaf);                   /* if the subtree is better, */
  gi->err = e_tree;             /* discard the leaf and */
  return node;                  /* return the subtree */
}  /* _grow() */

/*--------------------------------------------------------------------*/

static DTREE* _cleanup (GROW *gi, int err)
{                               /* --- clean up after an error */
  if (gi->type == AT_SYM) {     /* if the target is symbolic */
    if (gi->curr) ft_delete(gi->curr);
    if (gi->best) ft_delete(gi->best);
    if (gi->numt) ft_delete(gi->numt); }
  else {                        /* if the target is numeric */
    if (gi->curr) vt_delete(gi->curr);
    if (gi->best) vt_delete(gi->best);
    if (gi->numt) vt_delete(gi->numt);
  }                             /* delete frequency/variation tables */
  if (gi->tpls) free(gi->tpls); /* and the tuple vector */
  if (err) {                    /* if to clean up after an error */
    if (gi->dtree) dt_delete(gi->dtree, 0);
    if (gi->attset && (gi->flags & DT_DUPAS))
      as_delete(gi->attset);    /* delete grown decision tree and */
  }                             /* corresponding attribute set */
  free(gi); return NULL;        /* delete tree grow information */
}  /* _cleanup() */

/*--------------------------------------------------------------------*/

DTREE* dt_grow (TABLE *table, int trgid, int measure, double *params,
                double minval, int maxht, float mincnt, int flags)
{                               /* --- grow dec./reg. tree from table */
  int    i, m;                  /* loop variable, flagless measure */
  TUPLE  **tp;                  /* to traverse the tuple vector */
  int    tplcnt;                /* number of tuples */
  int    valcnt;                /* number of possible values */
  int    maxcnt;                /* maximal number of values */
  ATTSET *attset;               /* attribute set */
  ATT    *att;                  /* target attribute */
  int    mark;                  /* mark     of the target attribute */
  INST   *inst;                 /* instance of the target attribute */
  double val;                   /* value    of the target attribute */
  GROW   *gi;                   /* tree grow information */
  DTREE  *dt;                   /* created decision tree */

  assert(table);                /* check the function argument */

  /* --- create an empty decision/regression tree --- */
  attset = tab_attset(table);   /* get the attribute set */
  for (maxcnt = 2, i = as_attcnt(attset); --i >= 0; ) {
    if (i == trgid) continue;   /* traverse attributes except target */
    valcnt = att_valcnt(as_att(attset, i));
    if (valcnt > maxcnt) maxcnt = valcnt;
  }                             /* determine maximum number of values */
  i  = (flags & DT_SUBSET) ? maxcnt : 1;
  gi = (GROW*)calloc(1, sizeof(GROW) +(i-1) *sizeof(int));
  if (!gi) return NULL;         /* create the tree grow information */
  gi->flags  = flags;           /* and store the induction flags */
  if (flags & DT_DUPAS) attset = as_dup(attset);
  if (!attset) return _cleanup(gi, 1);
  gi->attset = attset;          /* duplicate att. set if requested */
  gi->type   = att_type(as_att(attset, trgid));
  if (gi->type == AT_SYM) {     /* if the target att. is symbolic */
    m = measure & ~FEF_WGTD;    /* remove the flags from measure code */
    if ((m <= FEM_NONE) || (m >= FEM_UNKNOWN))
      measure = m = FEM_NONE; } /* check the selection measure code */
  else {                        /* if the target att. is numeric */
    m = measure & ~VEF_WGTD;    /* remove the flags from measure code */
    if ((m <= VEM_NONE) || (m >= VEM_UNKNOWN))
      measure = m = VEM_NONE;   /* check the selection measure code */
  }
  gi->measure = measure;        /* note measure, parameters, */
  gi->params  = params;         /* minimal value for a split, */
  gi->minval  = minval;         /* and minimal number of cases */
  gi->mincnt  = (mincnt <= 0) ? (float)EPSILON : mincnt;
  gi->maxht   = (maxht  <= 0) ? INT_MAX : (maxht -1);
  gi->flags   = flags;          /* note maximal tree height and flags */

  /* --- create decision/regression tree --- */
  dt = dt_create(attset, trgid);
  if (!dt) return _cleanup(gi, 1);
  gi->dtree = dt;               /* create an empty decision tree */
  tp = _tuples(table, trgid, &tplcnt);
  if (!tp) return _cleanup(gi, 1);
  gi->tpls = tp;                /* collect tuples with known class */

  /* --- create a root node --- */
  if (gi->type == AT_SYM) {     /* if the target att. is symbolic */
    gi->numt = ft_create(2, dt->clscnt);
    if (!gi->numt) return _cleanup(gi, 1);
    ft_init((FRQTAB*)gi->numt, 1, dt->clscnt);
                                /* create and init. a freq. table */
    for (tp += i = tplcnt; --i >= 0;) {
      inst = tpl_colval(*--tp, trgid);
      ft_add((FRQTAB*)gi->numt, 0, inst->i, tpl_info(*tp)->f);
    }                           /* aggregate the tuple weight and */
    ft_marg(gi->numt);          /* marginalize the frequency table */
    if (ft_known((FRQTAB*)gi->numt) > 0) {  /* if there are tuples */
      dt->root = _leaf_sym(dt, gi->numt, 0);
      if (!dt->root) return _cleanup(gi, 1);
    } }                         /* create a leaf node as the root */ 
  else {                        /* if the target att. is numeric, */
    gi->numt = vt_create(2);    /* create a variation table */
    if (!gi->numt) return _cleanup(gi, 1);
    vt_init((VARTAB*)gi->numt, 1); /* init. the variation table */
    for (tp += i = tplcnt; --i >= 0;) {
      inst = tpl_colval(*--tp, trgid);
      val  = (double)((dt->type == AT_INT) ? inst->i : inst->f);
      vt_add((VARTAB*)gi->numt, 0, val, tpl_info(*tp)->f);
    }                           /* aggregate the tuple weight and */
    vt_calc(gi->numt);          /* calculate the var. aggregates */
    if (vt_known((VARTAB*)gi->numt) > 0) { /* if there are tuples */
      dt->root = _leaf_num(dt, (VARTAB*)gi->numt, 0);
      if (!dt->root) return _cleanup(gi, 1);
    }                           /* create a root node */ 
  }                             /* from the variation table */

  /* --- recursively grow a tree --- */
  if (dt->root                  /* if a root node has been created */
  &&  (gi->maxht > 0)           /* and the maximum height is not zero */
  &&  (m > ((dt->type == AT_SYM)/* and a selection measure is given */
         ? FEM_NONE : VEM_NONE))) {
    for (i = as_attcnt(attset); --i >= 0; ) {
      if (i == trgid) continue; /* traverse attributes except class */
      valcnt = att_valcnt(as_att(attset, i));
      if (valcnt > maxcnt) maxcnt = valcnt;
    }                           /* determine maximal number of values */
    if (gi->type != AT_SYM) {   /* if the target att. is numeric */
      gi->best = vt_create(maxcnt);  /* create */
      gi->curr = vt_create(maxcnt);  /* variation tables */
      if (!gi->best || !gi->curr) return _cleanup(gi, 1);
      gi->eval_sym = (flags & DT_SUBSET) ? _num_set : _num_sym;
      gi->eval_num = _num_num;} /* set the evaluation functions */
    else {                      /* if the target att. is symbolic */
      gi->best = ft_create(maxcnt, dt->clscnt);
      gi->curr = ft_create(maxcnt, dt->clscnt);
      if (!gi->best || !gi->curr) return _cleanup(gi, 1);
      gi->eval_sym = (flags & DT_SUBSET) ? _sym_set : _sym_sym;
      gi->eval_num = _sym_num;  /* create frequency tables and */
    }                           /* set the evaluation functions */
    att  = as_att(attset, trgid);
    mark = att_getmark(att);    /* note mark of target attribute */
    att_setmark(att, -1);       /* and mark it (to prevent tests) */
    dt->root = _grow(gi, dt->root, tp, tplcnt);
    att_setmark(att, mark);     /* recursively grow a decision tree */
  }                             /* restore mark of target attribute */
  i = (gi->err < 0);            /* check whether an error occurred */
  _cleanup(gi, i);              /* clean up the temporary objects */
  if (i) return NULL;           /* if an error occured, abort */

  /* --- compute statistics --- */
  dt->attcnt = -1;              /* invalidate the attribute counter */
  dt->height =  0;              /* determine the tree information */
  dt->total  = (!dt->root) ? 0.0F : dt->root->frq;
  dt->size   = (!dt->root) ? 0    : _count(dt, dt->root, 0);
  dt->curr   = dt->root;        /* set the node cursor to the root */
  return dt;                    /* return the grown dec./reg. tree */
}  /* dt_grow() */

#endif  /* #ifdef DT_GROW */
/*----------------------------------------------------------------------
  Error Estimation Functions
----------------------------------------------------------------------*/
#ifdef DT_PRUNE

static double _sym_pess (double n, double e)
{                               /* --- pessimistic error estimation */
  static double adderr = 0.5;   /* additional errors */

  if (n < 0) {                  /* initialize add. number of errors */
    adderr = (e < 0) ? 0 : e; return 0; }
  e += adderr;                  /* estimate number of errors */
  return (e > n) ? n : e;       /* and return it */
}  /* _sym_pess() */

/*--------------------------------------------------------------------*/

static double _sym_clvl (double n, double e)
{                               /* --- conf. level error estimation */
  static double beta = 0.25;    /* beta = alpha/2 = (1-clvl)/2 */
  static double z2   = 0.454936;/* z^2_beta = z^2_{alpha/2} */
  double t0, t1;                /* temporary buffers */

  /* --- initialize --- */
  if (n < 0) {                  /* if to set the confidence level */
    if (e < 0) e = 0;           /* check the confidence level and */
    if (e > 1) e = 1;           /* correct it, if necessary */
    beta = 0.5 *(1-e);          /* note beta (percentage of errors) */
    z2   = _ndqtl(beta);        /* compute quantile of normal dist. */
    z2  *= z2; return 0;        /* for beta = alpha/2 = (1-clvl)/2 */
  }                             /* and note its square */

  /* --- estimate errors --- */
  assert(e <= n);               /* check the function arguments */
  if (z2 <= 0) return e;        /* check for a confidence level of 0 */
  if (e < EPSILON) {            /* if no errors were observed */
    return (n <= 0) ? 0 : (n *(1 -pow(beta, 1/n)));
  }                             /* compute a precise upper limit */
  if (e < 1) {                  /* if less than one error observed */
    t0 = _sym_clvl(n, 0);       /* compute estimate for zero errors */
    t1 = _sym_clvl(n, 1);       /* and estimate for one error and */
    return t0 +e *(t1 -t0);     /* interpolate between these, */
  }                             /* otherwise use an approx. formula */
  t0 = n *(e +z2/2 +sqrt(z2 *(e *(1 -e/n) +z2/4))) /(n +z2);
  return (e > t0) ? e : t0;     /* return the error estimate */
}  /* _sym_clvl() */

/*----------------------------------------------------------------------
Confidence level error estimation idea:
  J.R. Quinlan.
  C4.5: Programs for Machine Learning.
  Morgan Kaufman, San Mateo, CA, USA 1992, p.40--41
Mathematical formulae used in the above function:
  R.L. Larsen and M.L. Marx.
  An Introduction to Mathematical Statistics and its Applications.
  Prentice Hall, Englewood Cliffs, NJ, USA 1986, p.278--279
----------------------------------------------------------------------*/

static double _num_pess (double n, double e)
{                               /* --- pessimistic error estimation */
  static double adderr = 0.5;   /* additional (squared) errors */

  if (n < 0) {                  /* initialize add. number of errors */
    adderr = (e < 0) ? 0 : e; return 0; }
  return e +adderr;             /* return increased sum */
}  /* _num_pess() */            /* of squared errors */

/*--------------------------------------------------------------------*/

static double _num_clvl (double n, double e)
{                               /* --- conf. level error estimation */
  static double z = -0.67449;   /* z_{alpha/2}, alpha = (1-clvl)/2 */
  double t0, t1;                /* temporary buffer */

  /* --- initialize --- */
  if (n < 0) {                  /* if to set the confidence level */
    if (e < 0) e = 0;           /* check the confidence level and */
    if (e > 1) e = 1;           /* correct it, if necessary */
    z = _ndqtl(0.5 *(1-e));     /* compute quantile of normal dist. */
    return 0;                   /* for beta = alpha/2 = (1-clvl)/2 */
  }                             /* and abort the function */

  /* --- estimate errors --- */
  t0 = 2*n; t1 = sqrt(t0) +z;   /* compute denominator of fraction */
  if (t1 <= 0) t1 = (1.0/MAXFACT)*t0;
  else         t1 *= t1;        /* check and adapt the denominator */
  t0 /= t1; if (t0 > MAXFACT) t0 = MAXFACT;    /* bound the factor */
  return t0*e;                  /* return the increased sum */
}  /* _num_clvl() */            /* of squared errors */

/*----------------------------------------------------------------------
The above function computes a very coarse approximation of the upper
bound of a confidence interval for the sum of squared errors based on
the upper bound of the confidence interval for the variance:
  u(Var,clvl) = \frac{(n-1) Var^2}{\chi^2_{(1-clvl)/2}
              = \frac{sum of squared errors}{\chi^2_{(1-clvl)/2},
where clvl is the chosen confidence level and \chi^2_p the p quantile
of the \chi^2 distribution.
Source: K. Bosch.
        Elementare Einf\"uhrung in die angewandte Statistik.
        Vieweg, Braunschweig/Wiesbaden, 1995
For the quantile of the \chi^2 distribution the approximation
  \chi^2_p \approx 0.5*(\sqrt(2n-1) +z_p)^2        (same source)
is used, with 2n-1 replaced by 2n to avoid problems with very small n
(z_p is the p quantile of the normal distribution). Of course, this
approximation is justifiable only for large n (n >= 30 is a common rule
of thumb), but since the whole approach is based on crude heuristics
anyway ... (normal distribution assumption is usually not satisfied,
set of cases is not a random sample etc.)
----------------------------------------------------------------------*/

/*----------------------------------------------------------------------
  Pruning Functions
----------------------------------------------------------------------*/

static DTNODE* _prune (PRUNE *pi, DTNODE *node)
{                               /* --- recursively prune dec. tree */
  int    i;                     /* loop variable */
  DTDATA *data;                 /* to traverse data vector */
  double e_leaf, e_tree;        /* errors of the leaf/subtree */

  assert(pi && node);           /* check the function arguments */
  e_leaf  = (pi->est) ? pi->est(node->frq, node->err) : 0;
  pi->err = e_leaf;             /* estimate errors for a leaf */
  if (node->flags & DT_LEAF)    /* if the node is a leaf, */
    return node;                /* there is nothing else to do */
  if (pi->maxht <= 0)           /* if maximal tree height is reached, */
    return _collapse(pi->dtree, node);   /* collapse tree into a leaf */
  pi->maxht--;                  /* reduce maximal (sub)tree height */
  e_tree = 0;                   /* traverse the data vector */
  for (data = node->vec +(i = node->size); --i >= 0; ) {
    if (!(--data)->child || islink(data, node))
      continue;                 /* skip empty and linked subtrees */
    data->child = _prune(pi, data->child);
    e_tree += pi->err;          /* prune the subtree (branch) */
  }                             /* and sum the number of errors */
  pi->maxht++;                  /* restore maximal (sub)tree height */
  if (pi->est                   /* if to prune the tree */
  &&  (e_leaf < e_tree *(1+EPSILON))) {
    pi->err = e_leaf;           /* and the leaf is not worse */
    return _collapse(pi->dtree, node);
  }                             /* collapse the subtree into */ 
  pi->err = e_tree;             /* a leaf (prune the tree), */
  return node;                  /* otherwise keep the subtree */
}  /* _prune() */

/*--------------------------------------------------------------------*/

static DTNODE* _pr_tab (PRUNE *pi, DTNODE *node,
                        TUPLE **tpls, int n, int exec)
{                               /* --- recursively prune with table */
  int    i, k, trgid;           /* loop variables, buffers */
  int    type;                  /* type of test attribute */
  TUPLE  **tp;                  /* to traverse the tuple vector */
  DTDATA *data;                 /* to traverse the data vector */
  INST   *inst;                 /* instance of the target attribute */
  DTNODE *leaf;                 /* buffer for leaf */
  DTNODE *lbra;                 /* buffer for largest branch */
  double lbcnt;                 /* number of tuples in largest branch */
  TSEL   tsel;                  /* tuple selection information */
  GRPFN  *group;                /* tuple grouping function */
  double e_leaf,e_bran,e_tree;  /* errors of the leaf/branch/subtree */
  double frq, t0, t1;           /* temporary buffers */

  assert(pi && node && tpls && (n >= 0)); /* check the function args. */

  /* --- create a leaf node --- */
  if (n <= 0) {                 /* if there are no tuples, there */
    leaf = NULL; e_leaf = 0; }  /* can be no leaf and no errors */
  else {                        /* if there are tuples */
    trgid = pi->dtree->trgid;   /* get the target attribute index */
    if (pi->type == AT_SYM) {   /* if the target att. is symbolic */
      ft_init(pi->tab, 1, pi->dtree->clscnt); /* init. the  table */
      for (tp = tpls +(i = n); --i >= 0; ) {
        inst = tpl_colval(*--tp, trgid);   /* traverse the tuples */
        ft_add((FRQTAB*)pi->tab, 0, inst->i, tpl_info(*tp)->f);
      }                         /* aggregate the tuple weights and */
      ft_marg((FRQTAB*)pi->tab);/* marginalize the frequency table */
      leaf = _leaf_sym(pi->dtree, (FRQTAB*)pi->tab, 0);
      }                         /* create a new leaf node */
    else {                      /* if the target att. is numeric, */
      vt_init(pi->tab, 1);      /* initialize the variation table */
      for (tp = tpls +(i = n); --i >= 0; ) {
        inst = tpl_colval(*--tp, trgid);   /* traverse the tuples */
        t0 = (double)((pi->type == AT_INT) ? inst->i : inst->f);
        vt_add((VARTAB*)pi->tab, 0, t0, tpl_info(*tp)->f);
      }                         /* aggregate the tuple weights and */
      vt_calc((VARTAB*)pi->tab);/* marginalize the variation table */
      leaf = _leaf_num(pi->dtree, (VARTAB*)pi->tab, 0);
    }                           /* create a new leaf node */
    if (!leaf) { pi->err = -1; return node; }
    e_leaf = (pi->est) ? pi->est(leaf->frq, leaf->err) : 0;
  }                             /* estimate errors for a leaf */
  if (!leaf || !node            /* if no leaf created/no node exists */
  ||  (node->flags & DT_LEAF)   /* or the node is a leaf */
  ||  (pi->maxht <= 0)) {       /* or maximal tree height is reached */
    pi->err = e_leaf;           /* set the number of leaf errors */
    if (!exec) { if (leaf) free(leaf); return node; }
    if (node) _delete(node);    /* if not to execute pruning, abort */
    return leaf;                /* if a node exists, delete subtree */
  }                             /* and return the created leaf node */

  /* --- reset test node --- */
  lbra = _adapt(pi->attset, node);
  if (!lbra) { pi->err = -1; return node; }
  node = lbra;                  /* adapt the test node (more values) */
  if (exec) {                   /* if to execute pruning, */
    node->frq = leaf->frq;      /* copy the node frequency, */
    node->trg = leaf->trg;      /* the target value, */
    node->err = leaf->err;      /* and the number of leaf errors */
  }
  frq      = leaf->frq;         /* note the node frequency */
  tsel.col = node->attid;       /* get the column index and */
  e_tree   =  0;                /* clear the number of errors */
  lbcnt    = -1; lbra = NULL;   /* init. largest branch variables */
  pi->maxht--;                  /* reduce maximal (sub)tree height */

  /* --- branch on symbolic attribute --- */
  type = att_type(as_att(pi->attset, node->attid));
  if (type == AT_SYM){          /* if the test att. is symbolic */
    tp = tpls;                  /* traverse the tuple vector */
    tsel.val.i = UV_SYM;        /* group tuples with */
    i  = _grp_sym(tp, n,&tsel); /* an unknown attribute value */
    t0 = _sum_wgt(tp, i);       /* and sum their weights */
    t0 = frq -= t0;             /* compute weight of other tuples */
    group = (node->flags & DT_LINK) ? _grp_set : _grp_sym;
    data = tsel.vec = node->vec;/* traverse attribute values */
    for (tsel.val.i = 0; tsel.val.i < node->size; tsel.val.i++, data++){
      if (islink(data, node))   /* skip all links */
        continue;               /* to other values */
      k  = group(tp+i, n-i, &tsel); /* group tuples with next value */
      t1 = _sum_wgt(tp+i, k);       /* and sum their weights */
      if (t1 < EPSILON) {       /* if there are no tuples */
        if (data->child && exec) {     /* with this value */
          _delete(data->child); data->child = NULL; }
        continue;               /* delete subtree and */
      }                         /* continue with next value */
      if (i > 0) {              /* weight tuples with unknowns */
        _mul_wgt(tp, i, t1 /t0); t0 = t1; }
      data->child = _pr_tab(pi, data->child, tp, i+k, exec);
      if (pi->err < 0) { free(leaf); return node; }
      e_tree += pi->err;        /* prune the subtree */
      if (t1 > lbcnt) {         /* update largest branch */
        lbcnt = t1; lbra = data->child; }
      group(tp, i+k, &tsel);    /* regroup tuples with current value */
      tp += k; n -= k;          /* and skip processed tuples */
    }
    if (i > 0) _mul_wgt(tp, i, frq /t0);
    n = (int)(tp -tpls) +i;     /* reweight tuples with unknowns */
  }                             /* and restore the tuple counter */

  /* --- branch on numeric attribute --- */
  else {                        /* if the test att. is numeric */
    if (type == AT_FLT) { tsel.val.f = node->cut; group = _grp_flt; }
    else {    tsel.val.i = (int)floor(node->cut); group = _grp_int; }
    k  = group(tpls, n, &tsel); /* group tuples greater than the cut */
    t1 = _sum_wgt(tpls, k);     /* and sum their weights */
    group = (type == AT_FLT) ? _grp_uvflt : _grp_uvint;
    i  = group(tpls+k, n-k, &tsel); /* group tuples with unknowns, */
    t0 = _sum_wgt(tpls+k, i);   /* sum their weights, and */
    t0 = frq -= t0;             /* compute weight of other tuples */
    data  = node->vec;          /* get pointer to first subtree */
    lbcnt = t0 -t1;             /* compute number of tuples <= cut */
    if (lbcnt < EPSILON) {      /* if there are no such tuples, */
      if (data->child && exec){ /* delete the subtree */
        _delete(data->child); data->child = NULL; } }
    else {                      /* if tuples with value <= cut exist */
      if (i > 0) {              /* weight tuples with unknowns */
        _mul_wgt(tpls+k, i, lbcnt /t0); t0 = lbcnt; }
      data->child = _pr_tab(pi, data->child, tpls+k, n-k, exec);
      if (pi->err < 0) { free(leaf); return node; }
      e_tree += pi->err;        /* prune the 1st subtree (< cut) */
      lbra = data->child;       /* note child as largest branch */
      if (i > 0) group(tpls+k, n-k, &tsel);
    }                           /* regroup tuples with unknowns */
    data++;                     /* get pointer to second subtree */
    if (t1 < EPSILON) {         /* if no tuples with value > cut */
      if (data->child && exec){ /* exist, delete subtree */
        _delete(data->child); data->child = NULL; } }
    else {                      /* if tuples with value > cut exist */
      if (i > 0) {              /* weight tuples with unknowns */
        _mul_wgt(tpls+k, i, t1 /t0); t0 = t1; }
      data->child = _pr_tab(pi, data->child, tpls, i+k, exec);
      if (pi->err < 0) { free(leaf); return node; }
      e_tree += pi->err;        /* prune the 2nd subtree (> cut) */
      if (t1 > lbcnt) lbra = data->child;  /* update largest branch */
      if (i > 0) { group(tpls, i+k, &tsel); k = 0; }
    }                           /* regroup tuples with unknowns */
    if (i > 0) _mul_wgt(tpls+k, i, frq/t0);
  }                             /* reweight tuples with unknowns */
  pi->maxht++;                  /* restore max. (sub)tree height */
  if (!pi->est) {               /* if no pruning requested, abort */
    pi->err = e_tree; free(leaf); return node; }

  /* --- check largest branch --- */
  if (!pi->chklb                /* if not to test largest branch */
  ||  !lbra                     /* or no largest branch exists */
  ||  (lbra->flags & DT_LEAF))  /* or it is a leaf node, */
    e_bran = e_leaf;            /* simply set the leaf errors, */
  else {                        /* otherwise (branch is a subtree) */
    lbra   = _pr_tab(pi, lbra, tpls, n, 0);
    e_bran = pi->err;           /* compute the number of errors */
  }                             /* of the largest branch */

  /* --- prune according to results --- */
  if (!exec) {                  /* if not to execute pruning, */
    free(leaf);                 /* always discard the leaf node */
    pi->err = e_leaf;           /* determine the minimal error */
    if (e_tree *(1+EPSILON) < pi->err) pi->err = e_tree;
    if (e_bran *(1+EPSILON) < pi->err) pi->err = e_bran;
    return node;                /* return the original subtree */
  }
  if ((e_bran *(1+EPSILON) < e_leaf)    /* if the largest branch */
  &&  (e_bran *(1-EPSILON) < e_tree)) { /* is the best alternative */
    for (data = node->vec +(i = node->size); --i >= 0; ) {
      if ((--data)->child && !islink(data, node)
      &&  (data->child != lbra))/* traverse the child nodes */
        _delete(data->child);   /* and delete all subtrees */
    }                           /* except the largest branch */
    free(node);                 /* delete the subtree root */
    node = _pr_tab(pi, lbra, tpls, n, 1);
    if (pi->err < 0) { free(leaf); return node; }
    e_tree = pi->err;           /* prune the largest branch */
  }                             /* and get the number of errors */
  if (e_leaf < e_tree *(1+EPSILON)) {
    _delete(node);              /* if the leaf is better */
    pi->err = e_leaf;           /* than the subtree/branch, */
    return leaf;                /* discard the subtree/branch */
  }                             /* and return the leaf */
  free(leaf);                   /* if the subtree is better, */
  pi->err = e_tree;             /* discard the leaf and */
  return node;                  /* and return the subtree */
}  /* _pr_tab() */

/*--------------------------------------------------------------------*/

int dt_prune (DTREE *dt, int method, double param, int maxht,
              int chklb, TABLE *table)
{                               /* --- prune a dec./reg. tree */
  int    tplcnt;                /* number of tuples */
  int    clscnt;                /* number of classes */
  double *vec;                  /* buffer for class frequency vector */
  PRUNE  pi;                    /* prune information */

  assert(dt);                   /* check the function arguments */

  /* --- initialize variables --- */
  pi.dtree  = dt;               /* note the decision tree, its */
  pi.attset = dt->attset;       /* attribute set and the tree height */
  pi.maxht  = (maxht <= 0) ? INT_MAX : (maxht-1);
  pi.type   = dt->type;         /* note the target attribute type and */
  pi.chklb  = chklb;            /* whether to check largest branch */
  switch (method) {             /* evaluate the pruning method */
    case PM_PESS: pi.est = (dt->type == AT_SYM)
                         ? _sym_pess : _num_pess; break;
    case PM_CLVL: pi.est = (dt->type == AT_SYM)
                         ? _sym_clvl : _num_clvl; break;
    default     : pi.est = (ESTFN*)0;             break;
  }                             /* set the error estimation function */
  if (pi.est) pi.est(-1,param); /* and initialize it */
  if (dt_total(dt) < 0)         /* get (or compute) */
    return -1;                  /* the total frequency */

  /* --- adapt the number of classes --- */
  if (dt->type == AT_SYM) {     /* if the target att. is symbolic */
    clscnt = att_valcnt(as_att(dt->attset, dt->trgid));
    if (clscnt > dt->clscnt) {  /* get and check number of classes */
      vec = (double*)realloc(dt->buf, clscnt *sizeof(double));
      if (!vec) return -1;      /* enlarge the buffer vector */
      dt->buf = vec; dt->clscnt = clscnt;
    }                           /* set the new buffer vector */
  }                             /* and the new number of classes */

  /* --- prune the decision/regression tree --- */
  if (!table) {                 /* if no table given, prune directly */
    if (dt->root) dt->root = _prune(&pi, dt->root); }
  else {                        /* if a table is given */
    pi.tpls = _tuples(table, dt->trgid, &tplcnt);
    if (!pi.tpls) return -1;    /* collect tuples with known class */
    if (dt->type != AT_SYM)     /* if the target att. is numeric, */
      pi.tab = vt_create(1);    /* create a variation table, */
    else                        /* otherwise create a freq. table */
      pi.tab = ft_create(1, dt->clscnt);
    if (!pi.tab) { free(pi.tpls); return -1; }
    dt->root = _pr_tab(&pi, dt->root, pi.tpls, tplcnt, 1);
                                /* prune the tree with the table */
    if (dt->type == AT_SYM) ft_delete((FRQTAB*)pi.tab);
    else                    vt_delete((VARTAB*)pi.tab);
    free(pi.tpls);              /* delete the freq./var. table */
    if (pi.err < 0) return -1;  /* and the tuple vector and */
  }                             /* check for a pruning error */

  /* --- compute statistics --- */
  dt->attcnt = -1;              /* invalidate the attribute counter */
  dt->height =  0;              /* determine the tree information */
  dt->total  = (!dt->root) ? 0.0F : dt->root->frq;
  dt->size   = (!dt->root) ? 0    : _count(dt, dt->root, 0);
  dt->curr   = dt->root;        /* set the node cursor to the root */
  return 0;                     /* return 'ok' */
}  /* dt_prune() */

#endif  /* #ifdef DT_PRUNE */
