/*----------------------------------------------------------------------
  File    : frqtab.c
  Contents: frequency table management
  Author  : Christian Borgelt
  History : 26.06.1997 file created
            29.07.1997 first version completed
            11.08.1997 some functions changed to #define
            25.08.1997 functions ft_comb, ft_uncomb, and ft_dest added
            18.09.1997 bug in measure evaluation removed
            24.09.1997 function ft_alldst added
            29.09.1997 bug in function ft_comb removed
            30.09.1997 bug in evaluation with combined columns removed
            09.02.1998 order of evaluation measures changed
            24.02.1998 bug in function _wevid fixed
            23.03.1998 parameters added to evaluation functions
            20.03.1999 all 'float' fields/variables changed to 'double'
            25.10.1999 evaluation function _wdiff added
            15.09.2000 some assertions added
            02.12.2000 memory alloc. improved, function ft_copy added
            03.12.2000 table access optimized (concerning index -1)
            02.03.2001 evaluation measure FEM_INFGBAL added
            26.05.2001 computation of ln(\Gamma(n)) improved
            26.09.2001 bug in clean up in ft_create removed
            02.01.2002 measure FEM_SPCGBAL added, FEM_INFGBAL corrected
            06.01.2002 switched to sorting functions from vecops
            11.01.2002 measure FEM_CHI2NRM added
            22.01.2002 computations in _wdiff improved
            31.01.2002 computation of Gini index measures improved
            02.02.2002 BD and description length measures reprogrammed
            04.02.2002 quadratic information measures added
            04.07.2002 bug in function _bdm fixed (equiv. sample size)
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "vecops.h"
#include "gamma.h"
#include "frqtab.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#ifdef FT_EVAL
#define	M_PI         3.14159265358979323846  /* \pi   */
#define LN_2         0.69314718055994530942  /* ln(2) */
#define EPSILON      1e-12      /* to handle roundoff errors */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef double EVALFN (FRQTAB* ftab, int measure, double *params);

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char* mnames[FEM_UNKNOWN+1] = {
  /* FEM_NONE      0 */ "no measure",
  /* FEM_INFGAIN   1 */ "information gain",
  /* FEM_INFGBAL   2 */ "balanced information gain",
  /* FEM_INFGR     3 */ "information gain ratio",
  /* FEM_INFSGR1   4 */ "symmetric information gain ratio 1",
  /* FEM_INFSGR2   5 */ "symmetric information gain ratio 2",
  /* FEM_QIGAIN    1 */ "quadratic information gain",
  /* FEM_QIGBAL    2 */ "balanced quadratic information gain",
  /* FEM_QIGR      3 */ "quadratic information gain ratio",
  /* FEM_QISGR1    4 */ "symmetric quadratic information gain ratio 1",
  /* FEM_QISGR2    5 */ "symmetric quadratic information gain ratio 2",
  /* FEM_GINI      6 */ "Gini index",
  /* FEM_GINISYM   7 */ "symmetric Gini index",
  /* FEM_GINIMOD   8 */ "modified Gini index",
  /* FEM_RELIEF    9 */ "relief measure",
  /* FEM_WDIFF    10 */ "sum of weighted differences",
  /* FEM_CHI2     11 */ "chi^2 measure",
  /* FEM_CHI2NRM  12 */ "normalized chi^2 measure",
  /* FEM_WEVID    13 */ "weight of evidence",
  /* FEM_RELEV    14 */ "relevance",
  /* FEM_BDM      15 */ "Bayesian-Dirichlet / K2 metric",
  /* FEM_BDMOD    16 */ "modified Bayesian-Dirichlet / K2 metric",
  /* FEM_RDLREL   17 */ "reduction of description length (rel. freq.)",
  /* FEM_RDLABS   18 */ "reduction of description length (abs. freq.)",
  /* FEM_STOCO    19 */ "stochastic complexity",
  /* FEM_SPCGAIN  20 */ "specificity gain",
  /* FEM_SPCGAIN  21 */ "balanced specificity gain",
  /* FEM_SPCGR    22 */ "specificity gain ratio",
  /* FEM_SPCSGR1  23 */ "symmetric specificity gain ratio 1",
  /* FEM_SPCSGR2  24 */ "symmetric specificity gain ratio 2",
  /* FEM_UNKNOWN  25 */ "<unknown measure>",
};                              /* names of evaluation measures */

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static double _nsp (double *dist, int cnt)
{                               /* --- compute nonspecificity */
  double nsp  = 0;              /* nonspecificity */
  double prec = 0;              /* preceding frequency */
  double t;                     /* temporary buffer */

  assert(dist && (cnt >= 0));   /* check the function arguments */
  v_dblsort(dist, cnt);         /* sort the frequencies */
  for ( ; cnt > 1; cnt--) {     /* and then traverse them */
    t = *dist -prec; prec = *dist++;
    if (t > 0) nsp += t *log(cnt);
  }                             /* calculate and return the */
  return nsp;                   /* nonspecificity of the distribution */
}  /* _nsp() */

/*----------------------------------------------------------------------
  Evaluation Functions
----------------------------------------------------------------------*/

static double _info (FRQTAB *ftab, int measure, double *params)
{                               /* --- Shannon information measures */
  int    x, y;                  /* loop variables */
  double **c;                   /* to traverse the table columns */
  double *fx, *fy, *fxy;        /* to traverse the frequencies */
  double s_x, s_y, s_xy;        /* sums for entropy computation */
  double info, t;               /* information gain (ratio), buffer */

  assert(ftab);                 /* check the function argument */
  if (ftab->known < EPSILON) return 0;
  s_x = s_y = s_xy = 0;         /* process the row distribution */
  for (fy = ftab->frq_y +(y = ftab->ycnt); --y >= 0; )
    if (*--fy > 0) s_y += *fy *log(*fy);
  c = ftab->frq_xy +ftab->xcnt; /* process the column distribution */
  for (fx = ftab->frq_x +(x = ftab->xcnt); --x >= 0; --c) {
    if (*--fx <= 0) continue;   /* skip empty and combined columns */
    s_x += *fx *log(*fx);       /* process the column distribution */
    t = 0;                      /* and a conditional distribution */
    for (fxy = *c +(y = ftab->ycnt); --y >= 0; )
      if (*--fxy > 0) t += *fxy *log(*fxy);
    s_xy += t;                  /* process columns individually and */
  }                             /* sum the results (higher accuracy) */
  t = ftab->known; t *= log(t); /* compute N *log(N) only once */
  s_x  = t -s_x;                /* N H_x  = -N sum_x  p_x  *log(p_x)  */
  s_y  = t -s_y;                /* N H_y  = -N sum_y  p_y  *log(p_y)  */
  s_xy = t -s_xy;               /* N H_xy = -N sum_xy p_xy *log(p_xy) */
  info = s_x +s_y -s_xy;        /* compute information gain *N *ln(2) */
  switch (measure & 0xff) {     /* evaluate the measure code */
    case FEM_INFGBAL: info /= log(ftab->xcnt) *ftab->known; break;
    case FEM_INFGR  : if (s_x      <= 0) return 0;
                      info /= s_x;                          break;
    case FEM_INFSGR1: if (s_xy     <= 0) return 0;
                      info /= s_xy;                         break;
    case FEM_INFSGR2: if (s_x +s_y <= 0) return 0;
                      info /= s_x +s_y;                     break;
    default:          info /= LN_2 *ftab->known;            break;
  }                             /* form requested entropy ratio */
  if (measure & FEF_WGTD) return info *(ftab->known/ftab->frq);
  return info;                  /* return the information measure */
}  /* _info() */

/*--------------------------------------------------------------------*/

static double _quad (FRQTAB *ftab, int measure, double *params)
{                               /* --- quadratic information measures */
  int    x, y;                  /* loop variables */
  double **c;                   /* to traverse the table columns */
  double *fx, *fy, *fxy;        /* to traverse the frequencies */
  double s_x, s_y, s_xy;        /* sum of squared frequencies */
  double quad, t;               /* information gain (ratio), buffer */

  assert(ftab);                 /* check the function argument */
  if (ftab->known < EPSILON) return 0;
  s_y = s_x = s_xy = 0;         /* process the row distribution */
  for (fy = ftab->frq_y +(y = ftab->ycnt); --y >= 0; ) {
    --fy; s_y += *fy * *fy; }   /* compute sum_y N(y)^2 */
  c = ftab->frq_xy +ftab->xcnt; /* process the joint distribution */
  for (fx = ftab->frq_x +(x = ftab->xcnt); --x >= 0; --c) {
    if (*--fx <= 0) continue;   /* skip empty and combined columns */
    s_x += *fx * *fx;           /* process the column distribution */
    t = 0;                      /* and a conditional distribution */
    for (fxy = *c +(y = ftab->ycnt); --y >= 0; ) {
      --fxy; t += *fxy * *fxy; }
    s_xy += t;                  /* compute sum_xy N(x,y)^2 */
  }
  t = ftab->known; t *= t;      /* compute N^2 only once */
  s_x  = t -s_x;                /* N^2/2 H^2_i  = N^2 -sum_i  N_i^2  */
  s_y  = t -s_y;                /* N^2/2 H^2_j  = N^2 -sum_j  N_j^2  */
  s_xy = t -s_xy;               /* N^2/2 H^2_ij = N^2 -sum_ij N_ij^2 */
  quad = s_x +s_y -s_xy;        /* compute information gain *N *ln(2) */
  switch (measure & 0xff) {     /* evaluate the measure code */
    case FEM_QIGBAL: quad /= t *(1 -1/ftab->xcnt); break;
    case FEM_QIGR  : if (s_x      <= 0) return 0;
                     quad /= s_x;                  break;
    case FEM_QISGR1: if (s_xy     <= 0) return 0;
                     quad /= s_xy;                 break;
    case FEM_QISGR2: if (s_x +s_y <= 0) return 0;
                     quad /= s_x +s_y;             break;
    default:         quad /= 0.5 *t;               break;
  }                             /* form requested entropy ratio */
  if (measure & FEF_WGTD) return quad *(ftab->known/ftab->frq);
  return quad;                  /* return the information measure */
}  /* _quad() */

/*--------------------------------------------------------------------*/

static double _gini (FRQTAB *ftab, int measure, double *params)
{                               /* --- Gini index/relief measure */
  int    x, y;                  /* loop variables */
  double **c;                   /* to traverse the table columns */
  double *fx, *fy, *fxy;        /* to traverse the frequencies */
  double s_x, s_y, s_xy;        /* sum of squared frequencies */
  double w_xy, w_yx;            /* weighted sum of squared freq. */
  double gini, t;               /* Gini index / relief measure */

  assert(ftab);                 /* check the function argument */
  if (ftab->known < EPSILON) return 0;

  s_y = s_xy = w_yx = 0;        /* process the row distribution */
  for (fy = ftab->frq_y +(y = ftab->ycnt); --y >= 0; ) {
    --fy; s_y += *fy * *fy; }   /* compute sum_y N(y)^2 */
  c = ftab->frq_xy +ftab->xcnt; /* process the joint distribution */
  for (fx = ftab->frq_x +(x = ftab->xcnt); --x >= 0; --c) {
    if (*--fx <= 0) continue;   /* skip empty and combined columns */
    t = 0;                      /* process a conditional distribution */
    for (fxy = *c +(y = ftab->ycnt); --y >= 0; ) {
      --fxy; t += *fxy * *fxy; }
    s_xy += t;                  /* compute sum_xy N(x,y)^2 and */
    w_yx += t / *fx;            /* sum_x 1/N(x) sum_y N(x,y)^2 */
  }
  if ((measure & 0xff) == FEM_GINI) {
    return (w_yx -s_y /ftab->known)
         / ((measure & FEF_WGTD) ? ftab->frq : ftab->known);
  }                             /* compute and return the Gini index */

  s_x = w_xy = 0;               /* process the column distribution */
  for (fx = ftab->frq_x +(x = ftab->xcnt); --x >= 0; ) {
    --fx; s_x += *fx * *fx; }   /* compute sum_x N(x)^2 */
  for (fy = ftab->frq_y +(y = ftab->ycnt); --y >= 0; ) {
    if (*--fy <= 0) continue;   /* skip empty rows */
    t = 0;                      /* process a conditional distribution */
    c = ftab->frq_xy +ftab->xcnt;
    for (fx = ftab->frq_x +(x = ftab->xcnt); --x >= 0; --c)
      if (*--fx > 0) t += (*c)[y]*(*c)[y];
    w_xy += t / *fy;            /* sum_y 1/N(y) sum_x N(x,y)^2 */
  }

  t = ftab->known; t *= t;      /* compute N^2 */
  switch (measure & 0xff) {     /* evaluate the measure code */
    case FEM_GINIMOD:           /* modified Gini index */
      if (s_x <= 0) return 0;   /* check the denominator */
      gini = s_xy / s_x - s_y / t;                     break;
    case FEM_RELIEF:            /* relief measure */
      if ((s_y <= 0) || (t -s_y <= 0)) return 0;
      gini = s_xy / s_y - (s_x -s_xy) / (t -s_y);      break;
    default:                    /* symmetric Gini index */
      t = 2*t -s_x -s_y;        /* compute and */
      if (t <= 0) return 0;     /* check the denominator */
      gini = ((w_xy +w_yx) *ftab->known -s_x -s_y) /t; break;
  }                             /* compute requested measure */
  if (measure & FEF_WGTD) return gini *(ftab->known/ftab->frq);
  return gini;                  /* return Gini index/relief measure */
}  /* _gini() */

/*--------------------------------------------------------------------*/

static double _wdiff (FRQTAB *ftab, int measure, double *params)
{                               /* --- weighted sum of differences */
  int    x, y, m;               /* loop variables, integer exponent */
  double **c;                   /* to traverse the table columns */
  double *fx, *fy, *fxy;        /* to traverse the frequencies */
  double e, s, t;               /* exponent of difference, buffers */
  double sum;                   /* sum of weighted differences */

  assert(ftab);                 /* check the function argument */
  if (ftab->known < EPSILON) return 0;
  e = (params) ? *params : 1.0; /* get the sensitivity parameter */
  if      (e <= 0.0) e = 1.0;   /* an exponent less than 0 is useless */
  if      (e == 1.0) m = 1;     /* (exponent of absolute difference) */
  else if (e == 2.0) m = 2;     /* and create a flag to distinguish */
  else               m = 0;     /* the two simplest special cases */
  sum = 0;                      /* traverse the joint distribution */
  c = ftab->frq_xy +ftab->xcnt; /* and the column distribution */
  for (fx = ftab->frq_x +(x = ftab->xcnt); --x >= 0; --c) {
    if (*--fx <= 0) continue;   /* skip empty and combined columns */
    s   = 0;                    /* traverse the row distribution */
    fxy = *c +ftab->ycnt;       /* and the rows of the column */
    fy  = ftab->frq_y +(y = ftab->ycnt);
    switch (m) {                /* distinguish special cases */
      case  1: while (--y >= 0){/* if exponent for difference is 1 */
                 t  = *fx * *--fy -*--fxy *ftab->known;
                 s += *fxy *fabs(t);
               } break;         /* sum weighted absolute differences */
      case  2: while (--y >= 0){/* if exponent for difference is 2 */
                 t  = *fx * *--fy -*--fxy *ftab->known;
                 s += *fxy *t *t;
               } break;         /* sum weighted squared differences */
      default: while (--y >= 0){/* if any other exponent */
                 t  = *fx * *--fy -*--fxy *ftab->known;
                 s += *fxy *pow(fabs(t), e);
               } break;         /* raise the differences to */
    }                           /* the given power and sum them */
    sum += s;                   /* sum the results for the columns */
  }
  t = ftab->known;              /* normalize the sum */
  return sum /(pow(t*t, e) *((measure & FEF_WGTD) ? ftab->frq : t));
}  /* _wdiff() */               /* return sum of weighted differences */

/*--------------------------------------------------------------------*/

static double _chi2 (FRQTAB *ftab, int measure, double *params)
{                               /* --- chi^2 measure */
  int    x, y;                  /* loop variables */
  double **c;                   /* to traverse the table columns */
  double *fx, *fy, *fxy;        /* to traverse the frequencies */
  double s, t, p;               /* temporary buffers */
  double chi2;                  /* chi^2 measure */

  assert(ftab);                 /* check the function argument */
  if (ftab->known < EPSILON) return 0;
  chi2 = 0;                     /* traverse the joint distribution */
  c = ftab->frq_xy +ftab->xcnt; /* and the column distribution */
  for (fx = ftab->frq_x +(x = ftab->xcnt); --x >= 0; --c) {
    if (*--fx <= 0) continue;   /* skip empty and combined columns */
    s   = 0;                    /* traverse the rows of the column */ 
    fxy = *c +ftab->ycnt;       /* and the row distribution */
    for (fy = ftab->frq_y +(y = ftab->ycnt); --y >= 0; ) {
      --fxy;                    /* go to the next row */
      if (*--fy <= 0) continue; /* traverse non-empty rows */
      p = *fx * *fy;            /* compute (N(x)N(y) - N N(x,y))^2 */
      t = p -*fxy *ftab->known; /*       / (N(x)N(y))              */
      s += t *t /p;             /* and sum these values */
    }
    chi2 += s;                  /* sum the results for the columns */
  }                             /* (results in chi^2 *N) */
  if ((measure & 0xff) == FEM_CHI2NRM) {
    t = (ftab->xcnt-1) *(ftab->ycnt-1);
    if (t > 0) chi2 /= t;       /* normalize the chi^2 measure by */
  }                             /* dividing by the degrees of freedom */
  return chi2 /((measure & FEF_WGTD) ? ftab->frq : ftab->known);
}  /* _chi2() */                /* return the chi^2 measure */

/*--------------------------------------------------------------------*/

static double _wevid (FRQTAB *ftab, int measure, double *params)
{                               /* --- weight of evidence */
  int    x, y;                  /* loop variables */
  double **c;                   /* to traverse the table columns */
  double *fx, *fy, *fxy;        /* to traverse the frequencies */
  double outer, inner;          /* inner and outer sum */
  double a, b, z;               /* temporary buffers */

  assert(ftab);                 /* check the function argument */
  if (ftab->known < EPSILON) return 0;
  outer = 0;                    /* traverse the joint distribution */
  c = ftab->frq_xy +ftab->xcnt; /* and the column distribution */
  for (fx = ftab->frq_x +(x = ftab->xcnt); --x >= 0; --c) {
    if (*--fx <= 0) continue;   /* skip empty and combined columns */
    inner = 0;                  /* traverse the rows of the column */
    fxy   = *c +ftab->ycnt;     /* and the row distribution */
    for (fy = ftab->frq_y +(y = ftab->ycnt); --y >= 0; ) {
      z = *--fy * *--fxy;       /* compute the common term */
      a = *fxy* ftab->known -z; /* and the terms of */
      b = *fx * *fy         -z; /* the odds fraction */
      if ((a >= EPSILON) && (b >= EPSILON))
        inner += *fy *fabs(log(a/b));
    }                           /* compute the inner sum */
    outer += *fx *inner;        /* compute the outer sum */
  }
  outer /= LN_2 *ftab->known;   /* scale and return the result */
  return outer /((measure & FEF_WGTD) ? ftab->frq : ftab->known);
}  /* _wevid() */

/*--------------------------------------------------------------------*/

static double _relev (FRQTAB *ftab, int measure, double *params)
{                               /* --- relevance */
  int    x, y;                  /* loop variables */
  double **c;                   /* to traverse the table columns */
  double *fx, *fy, *fxy;        /* to traverse the frequencies */
  double max, tmp;              /* (maximal) frequency ratio */
  double relev = 0, sum;        /* relevance measure, buffer */

  assert(ftab);                 /* check the function argument */
  if (ftab->known < EPSILON) return 0;
  c = ftab->frq_xy +ftab->xcnt; /* traverse the joint distribution */
  for (fx = ftab->frq_x +(x = ftab->xcnt); --x >= 0; --c) {
    if (*--fx <= 0) continue;   /* skip empty and combined columns */
    max = sum = 0;              /* traverse the rows of the column */
    fxy = *c +ftab->ycnt;       /* and the row distribution */
    for (fy = ftab->frq_y +(y = ftab->ycnt); --y >= 0; ) {
      --fy;                     /* go to the next row */
      if (*--fxy <= 0) continue;/* skip zero frequencies */
      tmp = *fxy / *fy;         /* compute the frequency ratio */
      if (tmp <= max) sum += tmp;
      else          { sum += max; max = tmp; }
    }                           /* sum ratios with the exception */
    relev += sum;               /* of the/one maximal ratio */
  }
  relev = 1 -relev /(ftab->ycnt-1);
  if (measure & FEF_WGTD) return relev *(ftab->known/ftab->frq);
  return relev;                 /* return the relevance measure */
}  /* _relev() */

/*--------------------------------------------------------------------*/

static double _bdm (FRQTAB *ftab, int measure, double *params)
{                               /* --- Bayesian-Dirichlet metric */
  int    x, y;                  /* loop variables */
  double **c;                   /* to traverse the table columns */
  double *fx, *fy, *fxy;        /* to traverse the frequencies */
  double p, s, a;               /* (sum of) prior(s), sensitivity */
  double bdm, r, t, z;          /* Bayesian-Dirichlet metric, buffers */

  assert(ftab);                 /* check the function argument */
  if (ftab->known < EPSILON) return 0;
  a = (params && (params[0]+1 > 0)) ? params[0]+1 : 1;
  a *= a;                       /* get the sensitivity parameter */
  p = (params && (params[1] !=  0)) ? params[1]   : 1;
  if (p < 0) p /= -ftab->ycnt;  /* get the prior/equiv. sample size */
  for (r = 0, fy = ftab->frq_y +(y = ftab->ycnt); --y >= 0; )
    r += logGa(*--fy *a +p);    /* process the class distribution */
  s  =  ftab->ycnt *p;          /* and compute the reference value */
  r += -ftab->ycnt *logGa(p) +(logGa(s) -logGa(ftab->known *a +s));
  if (params && (params[1] < 0))/* adapt prior for l. equiv. version */
    s = ftab->ycnt *(p /= ftab->xcnt);
  z = logGa(s) -ftab->ycnt *logGa(p);
  c = ftab->frq_xy +ftab->xcnt; /* process the joint distribution */
  for (bdm = 0, fx = ftab->frq_x +(x = ftab->xcnt); --x >= 0; --c) {
    if (*--fx < 0) continue;    /* skip empty and combined columns */
    for (t = 0, fxy = *c +(y = ftab->ycnt); --y >= 0; )
      t += logGa(*--fxy *a +p); /* process a conditional distribution */
    bdm += t +z -logGa(*fx *a +s);
  }                             /* sum the results over the columns */
  bdm -= r;                     /* and return the weighted BD metric */
  return bdm /(LN_2 *((measure & FEF_WGTD) ? ftab->frq : ftab->known));
}  /* _bdm() */

/*--------------------------------------------------------------------*/

static double _bdmod (FRQTAB *ftab, int measure, double *params)
{                               /* --- modified BD / K2 metric 2 */
  int    x, y;                  /* loop variables */
  double **c;                   /* to traverse the table columns */
  double *fx, *fy, *fxy;        /* to traverse the frequencies */
  double p, a;                  /* prior and sensitivity */
  double bdm, r, t, z;          /* Bayesian-Dirichlet metric, buffers */

  assert(ftab);                 /* check the function argument */
  if (ftab->known < EPSILON) return 0;
  a = (params && (params[0] >= 0)) ? params[0] : 0;
  if (a < 0) a = 0;             /* get the sensitivity parameter */
  p = (params && (params[1] != 0)) ? params[1] : 1;
  if (p < 0) p /= -ftab->ycnt;  /* get the prior/equiv. sample size */
  for (r = 0, fy = ftab->frq_y +(y = ftab->ycnt); --y >= 0; ) {
    z = *--fy *a +p; r += logGa(z +*fy) -logGa(z); }
  z  = ftab->known *a +ftab->ycnt *p;    /* process the */
  r += logGa(z) -logGa(z +ftab->known);  /* class distribution */
  if (params && (params[1] < 0))/* adapt the prior for the */
    p /= ftab->xcnt;            /* likelihood equivalent version */
  c = ftab->frq_xy +ftab->xcnt; /* and the column distribution */
  for (bdm = 0, fx = ftab->frq_x +(x = ftab->xcnt); --x >= 0; --c) {
    if (*--fx < 0) continue;    /* skip empty and combined columns */
    for (t = 0, fxy = *c +(y = ftab->ycnt); --y >= 0; ) {
      z = *--fxy *a +p; t += logGa(z +*fxy) -logGa(z); }
    z = *fx *a +ftab->ycnt *p;  /* process a conditional distribution */
    bdm += t +(logGa(z) -logGa(z +*fx));
  }                             /* sum the results over the columns */
  bdm -= r;                     /* and return the weighted BD metric */
  return bdm /(LN_2 *((measure & FEF_WGTD) ? ftab->frq : ftab->known));
}  /* _bdmod() */

/*--------------------------------------------------------------------*/

static double _rdlen (FRQTAB *ftab, int measure, double *params)
{                               /* --- reduction of desc. length */
  int    x, y;                  /* loop variables */
  double **c;                   /* to traverse the table columns */
  double *fx, *fy, *fxy;        /* to traverse the frequencies */
  double lGy;                   /* buffer for log_2(Gamma(ycnt)) */
  double a;                     /* sensitivity parameter */
  double dat, mod, s, t;        /* description lengths, buffer */

  assert(ftab);                 /* check the function argument */
  if (ftab->known < EPSILON) return 0;
  lGy = logGa(ftab->ycnt);      /* note \ln\Gamma(number of classes)) */
  a   = (params || (params[0]+1 > 0)) ? params[0]+1 : 1;
  dat = mod = 0;                /* get the sensitivity parameter */
  c   = ftab->frq_xy +ftab->xcnt;

  /* --- relative frequency coding --- */
  if ((measure & 0xff) == FEM_RDLREL) {
    for (fx = ftab->frq_x +(x = ftab->xcnt); --x >= 0; --c) {
      if (*--fx <= 0) continue; /* skip empty and combined columns */
      s = 0;                    /* process a conditional distribution */
      for (fxy = *c +(y = ftab->ycnt); --y >= 0; )
        if (*--fxy > 0) s += *fxy *log(*fxy);
      dat -= *fx *log(*fx) -s;  /* compute and sum cond. entropy */
      mod -= logGa(*fx +ftab->ycnt) -logGa(*fx +1) -lGy;
    }                           /* sum conditional model terms */
    s = 0;                      /* process the row distribution */
    for (fy = ftab->frq_y +(y = ftab->ycnt); --y >= 0; )
      if (*--fy > 0) s += *fy *log(*fy);
    dat += ftab->known *log(ftab->known) -s;
    mod += logGa(ftab->known +ftab->ycnt) -logGa(ftab->known +1) -lGy;
  }                             /* compute red. of desc. lengths */

  /* --- absolute frequency coding --- */
  else {                        /* ((measure & 0xff) == FEM_RDLABS) */
    for (fx = ftab->frq_x +(x = ftab->xcnt); --x >= 0; --c) {
      if (*--fx <= 0) continue; /* skip empty and combined columns */
      s = 0;                    /* process a conditional distribution */
      for (fxy = *c +(y = ftab->ycnt); --y >= 0; )
        s += logGa(*--fxy +1);  /* compute data description length */
      dat -= (t = logGa(*fx +1)) -s;         /* and sum the result */
      mod -= logGa(*fx +ftab->ycnt) -t -lGy;
    }                           /* sum model description lengths */
    s = 0;                      /* process the row distribution */
    for (fy = ftab->frq_y +(y = ftab->ycnt); --y >= 0; )
      s += logGa(*--fy +1);     /* compute data description length */
    dat += (t = logGa(ftab->known +1)) -s;
    mod += logGa(ftab->known +ftab->ycnt) -t -lGy;
  }                             /* compute red. of desc. lengths */

  return (dat +mod/a)           /* return the weighted measure */
       / (LN_2 *((measure & FEF_WGTD) ? ftab->frq : ftab->known));
}  /* _rdlen() */

/*--------------------------------------------------------------------*/

static double _stoco (FRQTAB *ftab, int measure, double *params)
{                               /* --- stochastic complexity */
  int    x;                     /* loop variable */
  double *fx;                   /* to traverse the frequencies */
  double stc;                   /* stochastic complexity */

  assert(ftab);                 /* check the function argument */
  if (ftab->known < EPSILON) return 0;
  stc = log(0.5 *ftab->known);  /* compute log(n../2) */
  for (fx = ftab->frq_x +(x = ftab->xcnt); --x >= 0; ) {
    if (*--fx > 0) stc -= log(0.5 * *fx);
  }                             /* compute (r_y -1)/2 *[ log(n../2) */
  stc *= 0.5 *(ftab->ycnt-1);   /*    - sum_{j=1}{r_x} log(n.j/2) ] */
  stc -= (ftab->xcnt-1) *(log(pow(M_PI, 0.5 *ftab->ycnt))
                        - logGa(0.5 *ftab->ycnt));
  stc /= LN_2 *ftab->known;     /* scale and normalize the length and */
  stc += _info(ftab, FEM_INFGAIN, NULL);  /* add the information gain */
  if (measure & FEF_WGTD) return stc *(ftab->known/ftab->frq);
  return stc;                   /* return the stochastic complexity */
}  /* _stoco() */

/*--------------------------------------------------------------------*/

static double _spec (FRQTAB *ftab, int measure, double *params)
{                               /* --- specificity measures */
  int    x, y;                  /* loop variables */
  double **c;                   /* to traverse the table columns */
  double *fx, *fxy;             /* to traverse the frequencies */
  double *bx, *by, *bxy;        /* to traverse the buffers */
  double s_x, s_y, s_xy;        /* nonspecificities */
  double spec;                  /* specificity gain (ratio) */

  assert(ftab);                 /* check the function argument */
  if ((ftab->xcnt <= 1) || (ftab->ycnt <= 1)
  ||  (ftab->known < EPSILON)) return 0;
  for (bx = ftab->buf_x  +(x = ftab->xcnt); --x >= 0; ) *--bx = 0;
  for (by = ftab->buf_y  +(y = ftab->ycnt); --y >= 0; ) *--by = 0;
  bxy = ftab->buf_xy;           /* clear the buffers */
  c  = ftab->frq_xy+ftab->xcnt; /* traverse the joint distribution */
  bx = ftab->buf_x +ftab->xcnt; /* and the frequency buffers */
  for (fx = ftab->frq_x +(x = ftab->xcnt); --x >= 0; --c) {
    --bx;                       /* go to the next column */
    if (*--fx <= 0) continue;   /* skip empty and combined columns */
    by = ftab->buf_y +ftab->ycnt;
    for (fxy = *c +(y = ftab->ycnt); --y >= 0; ) {
      --by; --fxy;              /* traverse the rows of the column */
      if (*fxy > 0)   *bxy++ = *fxy;
      if (*fxy > *by) *by    = *fxy;
      if (*fxy > *bx) *bx    = *fxy;
    }                           /* compute the joint and marginal */
  }                             /* possibility distributions */
  s_x  = _nsp(ftab->buf_x,  ftab->xcnt);  /* compute the */
  s_y  = _nsp(ftab->buf_y,  ftab->ycnt);  /* nonspecificities */
  s_xy = _nsp(ftab->buf_xy, (int)(bxy -ftab->buf_xy));
  spec = s_x +s_y -s_xy;        /* compute the specificity gain */
  switch (measure & 0xff) {     /* evaluate the measure code */
    case FEM_SPCGBAL: spec /= log(ftab->xcnt); break;
    case FEM_SPCGR  : if (s_x      <= 0) return 0;
                      spec /= s_x;          break;
    case FEM_SPCSGR1: if (s_xy     <= 0) return 0;
                      spec /= s_xy;         break;
    case FEM_SPCSGR2: if (s_x +s_y <= 0) return 0;
                      spec /= s_x +s_y;     break;
    default:          spec /= LN_2;         break;
  }                             /* form requested nonspec. ratio */
  if (measure & FEF_WGTD) return spec *(ftab->known/ftab->frq);
  return spec;                  /* return the specificity measure */
}  /* _spec() */

/*----------------------------------------------------------------------
  Table of Evaluation Functions
----------------------------------------------------------------------*/
static EVALFN *_evalfn[] = {    /* evaluation functions */
  /* FEM_NONE      0 */ (EVALFN*)0,
  /* FEM_INFGAIN   1 */ _info,
  /* FEM_INFGBAL   2 */ _info,
  /* FEM_INFGR     3 */ _info,
  /* FEM_INFSGR1   4 */ _info,
  /* FEM_INFSGR2   5 */ _info,
  /* FEM_QIGAIN    6 */ _quad,
  /* FEM_QIGBAL    7 */ _quad,
  /* FEM_QIGR      8 */ _quad,
  /* FEM_QISGR1    9 */ _quad,
  /* FEM_QISGR2   10 */ _quad,
  /* FEM_GINI     11 */ _gini,
  /* FEM_GINISYM  12 */ _gini,
  /* FEM_GINIMOD  13 */ _gini,
  /* FEM_RELIEF   14 */ _gini,
  /* FEM_WDIFF    15 */ _wdiff,
  /* FEM_CHI2     16 */ _chi2,
  /* FEM_CHI2NRM  17 */ _chi2,
  /* FEM_WEVID    18 */ _wevid,
  /* FEM_RELEV    19 */ _relev,
  /* FEM_BDM      20 */ _bdm,
  /* FEM_BDMOD    21 */ _bdmod,
  /* FEM_RDLEN1   22 */ _rdlen,
  /* FEM_RDLEN2   23 */ _rdlen,
  /* FEM_STOCO    24 */ _stoco,
  /* FEM_SPCGAIN  25 */ _spec,
  /* FEM_SPCGBAL  26 */ _spec,
  /* FEM_SPCGR    27 */ _spec,
  /* FEM_SPCSGR1  28 */ _spec,
  /* FEM_SPCSGR2  29 */ _spec,
  /* FEM_UNKNOWN  30 */ (EVALFN*)0,
};

#endif  /* #ifdef FT_EVAL */
/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

FRQTAB* ft_create (int xsize, int ysize)
{                               /* --- create a frequency table */
  int    i;                     /* loop variable, temporary buffer */
  FRQTAB *ftab;                 /* created frequency table */
  double *p;                    /* to traverse the table elements */

  if (xsize < 1) xsize = 1;     /* check and correct */
  if (ysize < 1) ysize = 1;     /* the function arguments */
  i = xsize*ysize +xsize+ysize; /* compute number of buffer elements */
  ftab = (FRQTAB*)calloc(1, sizeof(FRQTAB) +xsize *sizeof(double*));
  if (!ftab) return NULL;       /* allocate frequency table body */
  ftab->xsize = xsize++;        /* and set the number of classes and */
  ftab->ysize = ysize++;        /* the maximal number of att. values */
  ftab->dsts  = (int*)malloc(xsize *sizeof(int));
  if (!ftab->dsts) { free(ftab); return NULL; }
  ftab->dsts++;                 /* allocate destination vector */
  i += xsize*ysize+xsize+ysize; /* add number of table elements */
  p = (double*)malloc(i *sizeof(double));
  if (!p) { free(ftab->dsts-1); free(ftab); return NULL; }
  ftab->frq_x = p += 1;         /* allocate table and buffer */
  ftab->frq_y = p += xsize;     /* elements and organize them */
  for (i = xsize; --i >= 0; )
    ftab->frq_xy[i] = p += ysize;
  ftab->buf_x  = p += ysize -1;
  ftab->buf_y  = p += xsize -1;
  ftab->buf_xy = p += ysize -1;
  return ftab;                  /* return the created structure */
}  /* ft_create() */

/*--------------------------------------------------------------------*/

void ft_delete (FRQTAB* ftab)
{                               /* --- delete a frequency table */
  assert(ftab);                 /* check the function argument */
  if (ftab->frq_x) free(ftab->frq_x -1);
  if (ftab->dsts)  free(ftab->dsts  -1);
  free(ftab);                   /* delete table entries, vectors, */
}  /* ft_delete() */            /* and the table body */

/*--------------------------------------------------------------------*/

int ft_init (FRQTAB *ftab, int xcnt, int ycnt)
{                               /* --- initialize a frequency table */
  int    x, y;                  /* loop variables */
  double *fxy;                  /* to traverse the frequencies */
  int    *d;                    /* to traverse the destination refs. */

  assert(ftab                   /* check the function arguments */
      && (xcnt <= ftab->xsize) && (ycnt <= ftab->ysize));
  if (xcnt >= 0) ftab->xcnt = xcnt;  /* note the number of */
  if (ycnt >= 0) ftab->ycnt = ycnt;  /* columns and rows */
  for (x = ftab->xcnt +1; --x >= 0; )
    for (fxy = ftab->frq_xy[x] +(y = ftab->ycnt), ++y; --y >= 0; )
      *--fxy = 0;               /* clear the joint frequencies */
  for (d = ftab->dsts +(x = ftab->xcnt), ++x; --x >= 0; )
    *--d = -1;                  /* clear the column destination refs. */
  return 0;                     /* return 'ok' */
}  /* ft_init() */

/*--------------------------------------------------------------------*/

void ft_copy (FRQTAB *dst, const FRQTAB *src)
{                               /* --- copy a frequency table */
  int    x, y;                  /* loop variable */
  double *df; const double *sf; /* to traverse the table elements */
  int    *di; const int    *si; /* to traverse the dest. vectors */
  
  assert(dst && src             /* check the function arguments */
      && (dst->xsize >= src->xcnt)
      && (dst->ysize >= src->ycnt));
  dst->xcnt = src->xcnt;        /* copy the current size */
  dst->ycnt = src->ycnt;        /* of the frequency table */
  for (x = dst->xcnt +1; --x >= 0; ) {
    sf = src->frq_xy[x] +dst->ycnt;
    df = dst->frq_xy[x] +dst->ycnt;
    for (y = dst->ycnt +1; --y >= 0; )
      *--df = *--sf;            /* traverse and copy the elements */
  }                             /* of the joint frequency table */
  si = src->dsts  +dst->xcnt;   /* traverse and copy the */
  di = dst->dsts  +dst->xcnt;   /* destination indicators */
  for (x = dst->xcnt +1; --x >= 0; ) *--di = *--si;
  sf = src->frq_x +dst->xcnt;   /* traverse and copy the */
  df = dst->frq_x +dst->xcnt;   /* marginal x-frequencies */
  for (x = dst->xcnt +1; --x >= 0; ) *--df = *--sf;            
  sf = src->frq_y +dst->ycnt;   /* traverse and copy the */
  df = dst->frq_y +dst->ycnt;   /* marginal y-frequencies */
  for (y = dst->ycnt +1; --y >= 0; ) *--df = *--sf;
  dst->known = src->known;      /* copy number of cases */
  dst->frq   = src->frq;        /* with known value and */
}  /* ft_copy() */              /* the total number of cases */

/*--------------------------------------------------------------------*/

void ft_marg (FRQTAB *ftab)
{                               /* --- marginalize a frequency table */
  int    x, y;                  /* loop variables */
  int    *d;                    /* to traverse destination references */
  double **p;                   /* to traverse joint table columns */
  double *fxy, *fx, *fy;        /* to traverse frequencies */

  assert(ftab);                 /* check the function arguments */
  ftab->known = 0;              /* clear the total for knowns */
  for (fx = ftab->frq_x +(x = ftab->xcnt), ++x; --x >= 0; )
    *--fx = 0;                  /* clear the marginal x-frequencies */
  for (fy = ftab->frq_y +(y = ftab->ycnt), ++y; --y >= 0; )
    *--fy = 0;                  /* clear the marginal y-frequencies */
  fx = ftab->frq_x  +ftab->xcnt;
  p  = ftab->frq_xy +ftab->xcnt +1;
  for (d = ftab->dsts +(x = ftab->xcnt); --x >= 0; ) {
    --p; --fx;                  /* traverse the other columns */
    if (*--d >= 0) {            /* skip combined columns and */
      *fx = -1; continue; }     /* invalidate their marginals */
    fy  = ftab->frq_y +ftab->ycnt;
    fxy = *p +ftab->ycnt;       /* traverse the rows of the column */
    for (y = ftab->ycnt; --y >= 0; ) {
      *--fy += *--fxy;          /* marginalize for known y */
      *  fx += *  fxy;          /* and         for known x */
    }                           /* (sum rows and columns) */
    *--fy += *--fxy;            /* marginalize for unknown y */
    ftab->known += *fx;         /* compute total for knowns */
  }
  for (--fx, fxy = *--p +(y = ftab->ycnt); --y >= 0; )
    *fx += *--fxy;              /* marginalize for unknown x */
  ftab->frq = ftab->known       /* compute the overall total */
            + ftab->frq_x    [-1]
            + ftab->frq_y    [-1]
            + ftab->frq_xy[0][-1];
}  /* ft_marg() */

/*--------------------------------------------------------------------*/

void ft_addwm (FRQTAB *ftab, int x, int y, double frq)
{                               /* --- add frequency with marginals */
  assert(ftab                   /* check the function arguments */
     &&  (x <= ftab->xcnt) && (y <= ftab->ycnt));
  ftab->frq_xy[x+1][y] += frq;  /* add the frequency */
  ftab->frq_x [x]      += frq;  /* to the joint and */
  ftab->frq_y [y]      += frq;  /* to the marginal tables */
}  /* ft_addwm() */

/*--------------------------------------------------------------------*/

void ft_move (FRQTAB *ftab, int xsrc, int xdst, int y, double frq)
{                               /* --- move frequency */
  assert(ftab                   /* check the function arguments */
      && (xsrc >= 0) && (xsrc < ftab->xcnt)
      && (xdst >= 0) && (xdst < ftab->xcnt)
      && (y    >= 0) && (y    < ftab->ycnt));
  ftab->frq_xy[xsrc+1][y] -= frq;  /* move the frequency */
  ftab->frq_xy[xdst+1][y] += frq;
  ftab->frq_x [xsrc]      -= frq;
  ftab->frq_x [xdst]      += frq;
}  /* ft_move() */

/*--------------------------------------------------------------------*/

void ft_comb (FRQTAB *ftab, int xsrc, int xdst)
{                               /* --- combine two columns */
  int    i;                     /* loop variable */
  double *src, *dst;            /* to traverse the frequencies */
  double marg;                  /* marginal frequency */

  assert(ftab                   /* check the function arguments */
      && (xsrc >= 0) && (xsrc < ftab->xcnt)
      && (xdst >= 0) && (xdst < ftab->xcnt)
      && (ftab->dsts[xsrc] < 0));
  ftab->dsts[xsrc] = xdst;      /* note reference to destination */
  src  = ftab->frq_x +xsrc;     /* get pointer to source marginal */
  marg = *src; *src = -1;       /* and note and invalidate it */
  do {                          /* traverse destination columns */
    i   = ftab->ycnt +1;        /* get pointers to the table columns */
    src = ftab->frq_xy[xsrc+1] +i; /* traverse the joint frequencies */
    dst = ftab->frq_xy[xdst+1] +i; /* and sum them in the destination */
    while (--i >= 0) *--dst += *--src;
    ftab->frq_x[xdst] += marg;  /* sum marginal frequencies */
    xdst = ftab->dsts[xdst];    /* get next destination column */
  } while (xdst >= 0);          /* loop, while next column found */
}  /* ft_comb() */

/*--------------------------------------------------------------------*/

void ft_uncomb (FRQTAB *ftab, int xsrc)
{                               /* --- uncombine a column */
  int    i, xdst;               /* loop variable, buffer */
  double *src, *dst;            /* to traverse the frequencies */
  double marg = 0;              /* marginal frequency */

  assert(ftab                   /* check the function arguments */
      && (xsrc >= 0) && (xsrc < ftab->xcnt));
  xdst = ftab->dsts[xsrc];      /* get the first destination column */
  if (xdst < 0) return;         /* if column is not combined, abort */
  ftab->dsts[xsrc] = -1;        /* clear the destination reference */
  for (src = ftab->frq_xy[xsrc+1] +(i = ftab->ycnt); --i >= 0; )
    marg += *--src;             /* sum the joint frequencies */
  ftab->frq_x[xsrc] = marg;     /* and store this marginal */
  do {                          /* traverse destination columns */
    i   = ftab->ycnt +1;        /* get pointers to the table columns */
    src = ftab->frq_xy[xsrc+1] +i; /* traverse the joint frequencies */
    dst = ftab->frq_xy[xdst+1] +i; /* and subtract them */
    while (--i >= 0) *--dst -= *--src;
    ftab->frq_x[xdst] -= marg;  /* get pointer to dest. marginal */
    xdst = ftab->dsts[xdst];    /* get next destination column */
  } while (xdst >= 0);          /* loop, while next column found */
}  /* ft_uncomb() */

 /*--------------------------------------------------------------------*/

int ft_dest (FRQTAB *ftab, int x)
{                               /* --- get the destination of a column */
  int t;                        /* temporary buffer */

  assert(ftab                   /* check the function arguments */
      && (x >= 0) && (x < ftab->xcnt));
  while ((t = ftab->dsts[x]) >= 0) x = t;
  return x;                     /* follow references and return id */
}  /* ft_dest() */

/*--------------------------------------------------------------------*/

void ft_alldst (FRQTAB *ftab, int *dsts)
{                               /* --- get all column destinations */
  int i;                        /* loop variable */
  int x, t;                     /* temporary buffers */

  assert(ftab && dsts);         /* check the function arguments */
  for (i = 0; i < ftab->xcnt; i++) {
    x = i;                      /* default: index of source column */
    while ((t = ftab->dsts[x]) >= 0) x = t;
    *dsts++ = x;                /* follow references and */
  }                             /* set id of destination column */
}  /* ft_alldst() */

/*--------------------------------------------------------------------*/
#ifdef FT_EVAL

double ft_eval (FRQTAB *ftab, int measure, double *params)
{                               /* --- evaluate a frequency table */
  int m = measure & 0xff;       /* flagless measure */

  assert(ftab);                 /* check the function arguments */
  if ((m <= FEM_NONE) || (m >= FEM_UNKNOWN))
    return 0;                   /* check the measure code */
  return _evalfn[m](ftab, measure, params);
}  /* ft_eval() */              /* return the evaluation result */

/*--------------------------------------------------------------------*/

const char* ft_mname (int measure)
{                               /* --- get name of evaluation measure */
  static char buf[64];          /* buffer for measure name */
  int m = measure & 0xff;       /* index in names vector */

  if ((m < FEM_NONE) || (m > FEM_UNKNOWN))
    m = FEM_UNKNOWN;            /* check measure code */
  strcpy(buf, (measure & FEF_WGTD) ? "weighted " : "");
  strcat(buf, mnames[m]);       /* build measure name */
  return buf;                   /* and return it */
}  /* ft_mname() */

#endif
/*--------------------------------------------------------------------*/
#ifndef NDEBUG

void ft_show (FRQTAB *ftab)
{                               /* --- show a frequency table */
  int    x, y;                  /* loop variables */
  double **p;                   /* to traverse joint table columns */
  double *fx, *fy;              /* to traverse frequencies */

  assert(ftab);                 /* check the function argument */
  p = ftab->frq_xy;             /* get a pointer to the table and */
  printf("%6.1f |", p[0][-1]);  /* print the frequency of unknowns */
  for (x = ftab->xcnt; --x >= 0; ) /* print the frequency */
    printf(" %6.1f", (*++p)[-1]);  /* of unknown x and y */
  fy = ftab->frq_y;             /* get a pointer to the row dist. */
  printf(" | %6.1f\n", fy[-1]); /* and print the row marginal */
  for (x = ftab->xcnt; --x >= 0; ) printf("-------");
  printf("-----------------\n");/* print a separator */

  for (y = ftab->ycnt; --y >= 0; ) {
    p = ftab->frq_xy;           /* get a pointer to the table */
    printf("%6.1f |", p[0][y]); /* print the frequency of unknowns */
    for (x = ftab->xcnt; --x >= 0; )  /* traverse the columns and */
      printf(" %6.1f", (*++p)[y]); /* print the joint frequencies */
    printf(" | %6.1f\n", fy[y]);   /* and the row marginal */
  }
  for (x = ftab->xcnt; --x >= 0; ) printf("-------");
  printf("-----------------\n");/* print separator */

  fx = ftab->frq_x -1;          /* get a pointer to the column dist. */
  printf("%6.1f |", *fx);       /* print marginal for unknowns */
  for (x = ftab->xcnt; --x >= 0; ) /* traverse marginal frequencies */
    printf(" %6.1f", *++fx);       /* for knowns and print them */
  printf(" | %6.1f\n", ftab->known);
}  /* ft_show() */              /* finally print the total freq. */

#endif
