/*----------------------------------------------------------------------
  File    : rsx.c
  Contents: rule set execution
  Authors : Christian Borgelt
  History : 06.07.2002 file created from file dtx.c
            23.04.2003 missing AS_MARKED added for second reading
            16.08.2003 slight changes in error message output
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#ifndef SC_SCAN
#define SC_SCAN
#endif
#include "scan.h"
#ifndef AS_RDWR
#define AS_RDWR
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#ifndef RS_DESC
#define RS_DESC
#endif
#ifndef RS_PARSE
#define RS_PARSE
#endif
#include "io.h"
#include "rules.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "rsx"
#define DESCRIPTION "rule set execution"
#define VERSION     "version 1.3 (2003.09.25)         " \
                    "(c) 2002-2003   Christian Borgelt"

/* --- error codes --- */
#define OK            0         /* no error */
#define E_NONE        0         /* no error */
#define E_NOMEM     (-1)        /* not enough memory */
#define E_FOPEN     (-2)        /* cannot open file */
#define E_FREAD     (-3)        /* read error on file */
#define E_FWRITE    (-4)        /* write error on file */
#define E_OPTION    (-5)        /* unknown option */
#define E_OPTARG    (-6)        /* missing option argument */
#define E_ARGCNT    (-7)        /* wrong number of arguments */
#define E_STDIN     (-8)        /* double assignment of stdin */
#define E_PARSE     (-9)        /* parse error */
#define E_EMPTY    (-10)        /* empty rule set */
#define E_HEAD     (-11)        /* more than one head attribute */
#define E_CLASS    (-10)        /* missing class (head attribute) */
#define E_UNKNOWN  (-12)        /* unknown error */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- prediction result --- */
  ATT  *att;                    /* target attribute */
  int   type;                   /* type of the target attribute */
  char *n_pred;                 /* name  of prediction column */ 
  int   w_pred;                 /* width of prediction column */
  INST  pred;                   /* predicted value */
  char *n_supp;                 /* name  of support column */
  int   w_supp;                 /* width of support column */
  float supp;                   /* support of prediction */
  char *n_conf;                 /* name  of confidence column */
  int   w_conf;                 /* width of confidence column */
  float conf;                   /* confidence of prediction */
} RESULT;                       /* (prediction result) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
const char *errmsgs[] = {       /* error messages */
  /* E_NONE      0 */  "no error\n",
  /* E_NOMEM    -1 */  "not enough memory\n",
  /* E_FOPEN    -2 */  "cannot open file %s\n",
  /* E_FREAD    -3 */  "read error on file %s\n",
  /* E_FWRITE   -4 */  "write error on file %s\n",
  /* E_OPTION   -5 */  "unknown option -%c\n",
  /* E_OPTARG   -6 */  "missing option argument\n",
  /* E_ARGCNT   -7 */  "wrong number of arguments\n",
  /* E_STDIN    -8 */  "double assignment of standard input\n",
  /* E_PARSE    -9 */  "parse error(s) on file %s\n",
  /* E_EMPTY   -10 */  "rule set is empty\n",
  /* E_HEAD    -11 */  "more than one head attribute in rule set\n",
  /* E_CLASS   -12 */  "missing attribute \"%s\" in file %s\n",
  /* E_UNKNOWN -13 */  "unknown error\n"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
const  char    *prgname = NULL; /* program name for error messages */
static SCAN    *scan    = NULL; /* scanner */
static RULESET *ruleset = NULL; /* rule set to execute */
static ATTSET  *attset  = NULL; /* attribute set */
static FILE    *in      = NULL; /* input  file */
static FILE    *out     = NULL; /* output file */
static RESULT  res      = {     /* prediction result */
  NULL, AT_SYM,                 /* target attribute and its type */
  "rs", 0, { 0 },               /* data for prediction column */
  NULL, 0, 0.0F,                /* data for support    column */
  NULL, 0, 0.0F };              /* data for confidence column */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- print error message */
  va_list    args;              /* list of variable arguments */
  const char *msg;              /* error message */

  assert(prgname);              /* check the program name */
  if (code < E_UNKNOWN) code = E_UNKNOWN;
  if (code < 0) {               /* if to report an error, */
    msg = errmsgs[-code];       /* get the error message */
    if (!msg) msg = errmsgs[-E_UNKNOWN];
    fprintf(stderr, "\n%s: ", prgname);
    va_start(args, code);       /* get variable arguments */
    vfprintf(stderr, msg, args);/* print error message */
    va_end(args);               /* end argument evaluation */
  }
  #ifndef NDEBUG
  if (ruleset) rs_delete(ruleset, 0);
  if (attset)  as_delete(attset);  /* clean up memory */
  if (scan)    sc_delete(scan);    /* and close files */
  if (in  && (in  != stdin))  fclose(in);
  if (out && (out != stdout)) fclose(out);
  #endif
  #ifdef STORAGE
  showmem("at end of program"); /* check memory usage */
  #endif
  exit(code);                   /* abort the program */
}  /* error() */

/*--------------------------------------------------------------------*/

static void infout (ATTSET *set, FILE *file, int mode, CCHAR *seps)
{                               /* --- write additional information */
  static char *cfmt = "%.2f";   /* format for confidence output */
  int   i, k;                   /* loop variables, buffers */
  CCHAR *pred;                  /* name of prediction column/value */
  char  *supp, *conf;           /* support and confidence */
  char  b0[32], b1[32], b2[32]; /* buffers for output */

  if (mode & AS_ATT) {          /* if to write header */
    cfmt = (res.type == AT_SYM) ? "%.2f" : "%g";
    pred = res.n_pred;          /* get names of prediction column, */
    supp = res.n_supp;          /* support column, and */
    conf = res.n_conf;          /* confidence column */
    if (mode & AS_ALIGN) {      /* if to align fields */
      if ((mode & AS_WEIGHT) || supp || conf) {
        i = att_valwd(res.att, 0);
        k = (int)strlen(pred); res.w_pred = (i > k) ? i : k;
      }                         /* compute width of class column */
      if (supp && ((mode & AS_WEIGHT) || conf)) {
        k = (int)strlen(supp); res.w_supp = (k > 4) ? k : 4;
      }                         /* compute width of support column */
      if (conf && (mode & AS_WEIGHT)) {
        k = (int)strlen(conf); res.w_conf = (k > 4) ? k : 4; }
    } }                         /* compute width of errors column */
  else {                        /* if to write normal record */
    if (res.type == AT_SYM) {   /* if the target att. is symbolic */
      if (res.pred.i <= UV_SYM) pred = "?";
      else pred = att_valname(res.att, res.pred.i); }
    else {                      /* if the target att. is numeric */
      if (res.pred.f <= UV_FLT) pred = "?";
      else { sprintf(b0, "%g", res.pred.f); pred = b0; }
    }                           /* print the predicted value */
    if (res.n_supp) sprintf(b1, "%.1f", res.supp);
    if (res.n_conf) sprintf(b2, cfmt,   res.conf);
    supp = b1; conf = b2;       /* get and format field contents */
  }                             /* and get pointers to buffers */
  k = fprintf(file, pred);      /* write the prediction result */
  for (i = res.w_pred -k; --i >= 0; )
    fputc(seps[0], file);       /* if to align, pad with blanks */
  if (res.n_supp) {             /* if to write the support */
    fputc(seps[1], file);       /* write a field separator and */
    k = fprintf(file, supp);    /* the support of the prediction */
    for (i = res.w_supp -k; --i >= 0; )
      fputc(seps[0], file);     /* if to align, pad with blanks */
  }
  if (res.n_conf) {             /* if to write the confidence */
    fputc(seps[1], file);       /* write a field separator and the */
    k = fprintf(file, conf);    /* confidence of the prediction */
    for (i = res.w_conf -k; --i >= 0; )
      fputc(seps[0], file);     /* if to align, pad with blanks */
  }
}  /* infout() */

/*--------------------------------------------------------------------*/

int main (int argc, char* argv[])
{                               /* --- main function */
  int    i, k = 0;              /* loop variables, counter */
  char   *s;                    /* to traverse options */
  char   **optarg = NULL;       /* option argument */
  char   *fn_hdr  = NULL;       /* name of table header file */
  char   *fn_tab  = NULL;       /* name of table file */
  char   *fn_rs   = NULL;       /* name of rule set file */
  char   *fn_out  = NULL;       /* name of output file */
  char   *blanks  = NULL;       /* blanks */
  char   *fldseps = NULL;       /* field  separators */
  char   *recseps = NULL;       /* record separators */
  char   *uvchars = NULL;       /* unknown value characters */
  int    inflags  = 0, f;       /* table file read  flags */
  int    outflags = AS_ATT;     /* table file write flags */
  int    adapt    = 0;          /* flag for rule set adaptation */
  int    tplcnt   = 0;          /* number of tuples */
  double tplwgt   = 0.0;        /* weight of tuples */
  double errcnt   = 0.0;        /* number of misclassifications */
  int    maxlen   = 0;          /* maximal output line length */
  int    attid;                 /* loop variable for attributes */
  float  wgt;                   /* tuple/instantiation weight */
  RULE   *rule;                 /* to traverse the rules */
  INST   *inst;                 /* instance of target attribute */
  double diff, t;               /* difference to true value */
  TFSERR *err;                  /* error information */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument is given */
    printf("usage: %s [options] rsfile "
                     "[-d|-h hdrfile] tabfile [outfile]\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-x       adapt rule set support/confidence "
                    "(default: classify cases)\n");
    printf("-p#      prediction field name "
                    "(default: %s)\n", res.n_pred);
    printf("-s#      support    field name "
                    "(default: no support    field)\n");
    printf("-c#      confidence field name "
                    "(default: no confidence field)\n");
    printf("-a       align fields of output table "
                    "(default: do not align)\n");
    printf("-w       do not write field names to output file\n");
    printf("-b/f/r#  blank characters, field and record separators\n"
           "         (default: \" \\t\\r\", \" \\t\", \"\\n\")\n");
    printf("-u#      unknown value characters (default: \"?\")\n");
    printf("-n       number of tuple occurrences in last field\n");
    printf("rsfile   file containing rule set description\n");
    printf("-d       use default header "
                    "(field names = field numbers)\n");
    printf("-h       read table header (field names) from hdrfile\n");
    printf("hdrfile  file containing table header (field names)\n");
    printf("tabfile  table file to read "
                    "(field names in first record)\n");
    printf("outfile  file to write output table/rule set to "
                    "(optional)\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case 'x': adapt     = 1;                     break;
          case 'p': optarg    = &res.n_pred;           break;
          case 's': optarg    = &res.n_supp;           break;
          case 'c': optarg    = &res.n_conf;           break;
          case 'a': outflags |= AS_ALIGN;              break;
          case 'w': outflags &= ~AS_ATT;               break;
          case 'b': optarg    = &blanks;               break;
          case 'f': optarg    = &fldseps;              break;
          case 'r': optarg    = &recseps;              break;
          case 'u': optarg    = &uvchars;              break;
          case 'n': outflags |= AS_WEIGHT;
                    inflags  |= AS_WEIGHT;             break;
          case 'l': maxlen    = (int)strtol(s, &s, 0); break;
          case 'd': inflags  |= AS_DFLT;               break;
          case 'h': optarg    = &fn_hdr;               break;
          default : error(E_OPTION, *--s);             break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_rs  = s;      break;
        case  1: fn_tab = s;      break;
        case  2: fn_out = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check option argument */
  if ((k < 2) || (k > 3)) error(E_ARGCNT);
  if (fn_hdr && (strcmp(fn_hdr, "-") == 0))
    fn_hdr = "";                /* convert "-" to "" */
  i = (!fn_rs || !*fn_rs) ? 1 : 0;
  if  (fn_tab && !*fn_tab) i++;
  if  (fn_hdr && !*fn_hdr) i++; /* check assignments of stdin: */
  if (i > 1) error(E_STDIN);    /* stdin must not be used twice */
  if (fn_hdr)                   /* set the header file flag */
    inflags = AS_ATT | (inflags & ~AS_DFLT);
  if ((outflags & AS_ATT) && (outflags & AS_ALIGN))
    outflags |= AS_ALNHDR;      /* set align to header flag */

  /* --- read rule set --- */
  scan = sc_create(fn_rs);      /* create a scanner */
  if (!scan) error((!fn_rs || !*fn_rs) ? E_NOMEM : E_FOPEN, fn_rs);
  attset = as_create("domains", att_delete);
  if (!attset) error(E_NOMEM);  /* create an attribute set */
  fprintf(stderr, "\nreading %s ... ", sc_fname(scan));
  if ((sc_nexter(scan)   <  0)  /* start scanning (get first token) */
  ||  (as_parse(attset, scan, AT_ALL) != 0)
  ||  (as_attcnt(attset) <= 0))     /* parse attribute set */
    error(E_PARSE, sc_fname(scan));
  ruleset = rs_parse(attset, scan); /* parse the rule set */
  if (!ruleset || !sc_eof(scan)) error(E_PARSE, fn_rs);
  sc_delete(scan); scan = NULL; /* delete the scanner */
  i = rs_rulecnt(ruleset);      /* check whether the rule set */
  if (i < 0) error(E_EMPTY);    /* contains at least one rule */
  attid = r_headatt(rs_rule(ruleset, 0));
  while (--i > 0) {             /* check for a unique head attribute */
    if (r_headatt(rs_rule(ruleset, i)) != attid) break; }
  if (i > 0) error(E_HEAD);     /* (otherwise execution is imposs.) */
  fprintf(stderr, "[%d attribute(s), %d rule(s)] done.\n",
          as_attcnt(attset), rs_rulecnt(ruleset));
  if (adapt) {                  /* if to adapt support and confidence */
    for (i = rs_rulecnt(ruleset); --i >= 0; ) {
      rule = rs_rule(ruleset,i);/* traverse the rules */
      r_setsupp(rule, 0); r_setconf(rule, 0);
    }                           /* clear support and confidence */
  }                             /* of each rule */

  /* --- read table header --- */
  res.att  = as_att(attset, attid);
  res.type = att_type(res.att); /* get the target attribute */
  for (attid = as_attcnt(attset); --attid >= 0; )
    att_setmark(as_att(attset, attid), 1);
  att_setmark(res.att, 0);      /* mark all attribs. except the class */
  as_chars(attset, blanks, fldseps, recseps, uvchars);
  in = io_hdr(attset, fn_hdr, fn_tab, inflags|AS_MARKED, 1);
  if (!in) error(1);            /* read the table header */

  /* --- process tuples --- */
  if ((att_getmark(res.att) < 0)/* either the class must be present */
  &&  (k <= 2))                 /* or an output file must be written */
    error(E_CLASS, att_name(res.att), fn_tab);
  if ((k > 2) && !adapt) {      /* if to write an output table */
    if ((outflags & AS_ALIGN)   /* if to align output file */
    &&  (in != stdin)) {        /* and not to read from stdin */
      i = AS_INST | (inflags & ~(AS_ATT|AS_DFLT));
      while (as_read(attset, in, i) == 0);
      fclose(in);               /* determine the column widths */
      in = io_hdr(attset, fn_hdr, fn_tab, inflags|AS_MARKED, 1);
      if (!in) error(1);        /* reread the table header */
    }                           /* (necessary because of first tuple) */
    if (fn_out && *fn_out)      /* if an output file name is given, */
      out = fopen(fn_out, "w"); /* open the output file for writing */
    else {                      /* if no output file name is given, */
      out = stdout; fn_out = "<stdout>"; }       /* write to stdout */
    if (!out) error(E_FOPEN, fn_out);
    k = AS_MARKED|AS_INFO1|AS_RDORD|outflags;
    if (outflags & AS_ATT)      /* if to write table header */
      as_write(attset, out, k, infout);
    k = AS_INST|(k & ~AS_ATT);  /* write the attribute names */
  }                             /* to the output file */
  f = AS_INST | (inflags & ~(AS_ATT|AS_DFLT));
  i = ((inflags & AS_DFLT) && !(inflags & AS_ATT))
    ? 0 : as_read(attset, in, f);
  inst = att_inst(res.att);     /* get the instance of the target */
  while (i == 0) {              /* record read loop */
    i = rs_exec(ruleset);       /* find the first applicable rule */
    if (i >= 0) {               /* if a rule is applicable, */
      rule = rs_rule(ruleset,i);/* get and execute this rule */
      res.pred = *r_headval(rule);
      res.supp = r_getsupp(rule);
      res.conf = r_getconf(rule); }
    else {                      /* if no rule is applicable */
      rule = NULL;              /* use the rule variable as a flag */
      if (res.type == AT_SYM) res.pred.i = UV_SYM;
      else                    res.pred.f = UV_FLT;
      res.supp = res.conf = 0;  /* set an unknown prediction and */
    }                           /* clear the support and confidence */
    wgt = as_getwgt(attset);    /* classify tuple */
    tplwgt += wgt; tplcnt++;    /* count tuple and sum its weight */
    if (adapt && rule) r_setsupp(rule, res.supp +wgt);
    if (res.type == AT_SYM) {   /* if the target att. is symbolic */
      if ((inst->i > UV_SYM) && (res.pred.i != inst->i))
        errcnt += wgt;          /* count classification errors */
      else if (adapt && rule)   /* sum the correct classifications */
        r_setconf(rule, res.conf +wgt); }
    else {                      /* if the target att. is numeric */
      if (res.type == AT_INT)   /* if it is integer-valued */
        diff = (inst->i > UV_INT) ? res.pred.f -inst->i : 0;
      else                      /* if it is float-valued */
        diff = (inst->f > UV_FLT) ? res.pred.f -inst->f : 0;
      errcnt += t = diff *diff; /* sum the squared differences */
      if (adapt && rule)        /* for all tuples and for the rule */
        r_setconf(rule, res.conf +t);
    }
    if (out && (as_write(attset, out, k, infout) != 0))
      error(E_FWRITE, fn_out);  /* write tuple to output file */
    i = as_read(attset, in, f); /* try to read the next record */
  }
  if (i < 0) {                  /* if an error occurred, */
    err = as_err(attset);       /* get the error information */
    tplcnt += (inflags & (AS_ATT|AS_DFLT)) ? 1 : 2;
    io_error(i, fn_tab, tplcnt, err->s, err->fld, err->exp);
    error(1);                   /* print an error message */
  }                             /* and abort the program */
  if (in != stdin) fclose(in);  /* close the table file and */
  in = NULL;                    /* clear the file variable */
  if (out && (out != stdout)) { /* if an output file exists, */
    i = fclose(out); out = NULL;/* close the output file */
    if (i) error(E_FWRITE, fn_out);
  }                             /* print a success message */
  fprintf(stderr, "[%d/%g tuple(s)] done.\n", tplcnt, tplwgt);

  /* --- write the adapted rule set --- */
  if (adapt) {                  /* if to adapt support and confidence */
    for (i = rs_rulecnt(ruleset); --i >= 0; ) {
      rule = rs_rule(ruleset,i);/* traverse the rules */
      wgt  = r_getsupp(rule);   /* get and check */
      if (wgt <= 0) continue;   /* the new support of each rule */
      wgt = r_getconf(rule) /wgt;
      r_setconf(rule, (res.type == AT_SYM) ? wgt : sqrt(wgt));
    }                           /* compute and set the new confidence */
    if (!fn_out || !*fn_out) {  /* if no proper file name is given, */
      out = stdout; fn_out = "<stdout>"; }         /* write to stdout */
    else {                      /* if a table file is given, */
      fprintf(stderr, "writing %s ... ", fn_out);
      out = fopen(fn_out, "w"); /* open output file for writing */
      if (!out) error(E_FOPEN, fn_out);
    }
    outflags = RS_TITLE|RS_INFO|RS_CONF|RS_SUPP;
    if ((as_desc(attset, out, AS_TITLE|AS_MARKED, maxlen) != 0)
    ||  (rs_desc(ruleset, out, outflags, maxlen) != 0))
      error(E_FWRITE, fn_out);  /* describe domains and rule set */
    if (out != stdout) {        /* if not written to standard output, */
      i = fclose(out); out = NULL;           /* close the output file */
      if (i != 0) error(E_FWRITE, fn_rs);
    }                           /* print a success message */
    fprintf(stderr, "[%d attribute(s), %d rule(s)] done.\n",
                    as_attcnt(attset), rs_rulecnt(ruleset));
  }

  /* --- print the error statistics --- */
  if (att_getmark(res.att) >= 0) { /* if the target is present */
    if (res.type != AT_SYM)     /* if the target att. is numeric */
      fprintf(stderr, "root of mean squared error: %g\n",
              (tplwgt > 0) ? sqrt(errcnt /tplwgt) : 0);
    else                        /* if the target att. is symbolic */
      fprintf(stderr, "%g error(s) (%.2f%%)\n", errcnt,
              (tplwgt > 0) ? 100*(errcnt /tplwgt) : 0);
  }                             /* print number of misclass. */

  /* --- clean up --- */
  #ifndef NDEBUG
  rs_delete(ruleset, 1);        /* delete the rule set */
  #endif                        /* and the attribute set */
  #ifdef STORAGE
  showmem("at end of program"); /* check memory usage */
  #endif
  return 0;                     /* return 'ok' */
}  /* main() */
