/*----------------------------------------------------------------------
  File    : rules.c
  Contents: rule and rule set management
  Author  : Christian Borgelt
  History : 17.05.1998 file created
            26.05.1998 function rs_desc added
            27.05.1998 comparison operators added
            23.06.1998 adapted to modified attset functions
            13.08.1998 parameter delas added to function rs_delete
            24.08.1998 management of subsets of values added
            26.08.1998 function rs_ruleexg added
            29.09.1998 assertions added, minor improvements
            19.07.2001 global variables removed, rs_desc improved
            06.06.2002 rule and rule set parsing functions added
            25.09.2003 bug in function r_check fixed (condition dir)
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <assert.h>
#include "rules.h"
#include "vecops.h"
#ifndef RS_PARSE
#include "scan.h"
#endif
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define BLKSIZE  32             /* block size for rule vector */
#define MARK     0x80           /* mark for conditions to adapt */
#define MSBIT    31             /* most significant bit */

#ifdef RS_PARSE
/* --- error codes --- */
#define E_CHREXP    (-16)       /* character expected */
#define E_STREXP    (-17)       /* string expected */
#define E_ATTEXP    (-18)       /* attribute expected */
#define E_VALEXP    (-19)       /* attribute value expected */
#define E_NUMEXP    (-20)       /* number expected */
#define E_ILLNUM    (-21)       /* illegal number */
#define E_UNKATT    (-22)       /* unknown attribute */
#define E_UNKVAL    (-23)       /* unknown attribute value */
#define E_ILLOP     (-24)       /* illegal comparison operator */

/* --- functions --- */
#define ERROR(c)    _paerr(scan, c,        -1, NULL)
#define ERR_CHR(c)  _paerr(scan, E_CHREXP,  c, NULL)
#define ERR_STR(s)  _paerr(scan, E_STREXP, -1, s)
#define GET_TOK()   if (sc_next(scan) < 0) \
                      return sc_error(scan, sc_token(scan))
#define GET_CHR(c)  if (sc_token(scan) != (c)) return ERR_CHR(c); \
                    else GET_TOK();
#endif  /* #ifdef RS_PARSE */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
#ifdef RS_DESC
typedef struct {                /* --- description information --- */
  FILE   *file;                 /* output file */
  int    pos;                   /* position in output line */
  int    max;                   /* maximal length of output line */
  ATTSET *attset;               /* attribute set */
} DESC;                         /* (description information) */

#endif
/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#if defined(RS_DESC) || defined(RS_PARSE)
static const char *_opsigns[] = {
  NULL, "=", "<", "<=", ">", ">=", "!=", NULL, "in", "in" };
#endif                          /* signs for comparison operators */

#ifdef RS_PARSE
#ifdef GERMAN                     /* deutsche Texte */
static const char *errmsgs[] = {  /* Fehlermeldungen */
  /* E_CHREXP  -16 */  "\"%c\" erwartet statt %s",
  /* E_STREXP  -17 */  "\"%s\" erwartet statt %s",
  /* E_ATTEXP  -18 */  "Attribut erwartet statt %s",
  /* E_VALEXP  -19 */  "Attributwert erwartet statt %s",
  /* E_NUMEXP  -20 */  "Zahl erwartet statt %s",
  /* E_ILLNUM  -21 */  "ungltige Zahl %s",
  /* E_UNKATT  -22 */  "unbekanntes Attribut %s",
  /* E_UNKVAL  -23 */  "unbekannter Attributwert %s",
  /* E_ILLOP   -24 */  "ungltiger Vergleichsoperator %s",
};
#else                             /* English texts */
static const char *errmsgs[] = {  /* error messages */
  /* E_CHREXP  -16 */  "\"%c\" expected instead of %s",
  /* E_STREXP  -17 */  "\"%s\" expected instead of %s",
  /* E_ATTEXP  -18 */  "attribute expected instead of %s",
  /* E_VALEXP  -19 */  "attribute value expected instead of %s",
  /* E_NUMEXP  -20 */  "number expected instead of %s",
  /* E_ILLNUM  -21 */  "illegal number %s",
  /* E_UNKATT  -22 */  "unknown attribute %s",
  /* E_UNKVAL  -23 */  "unknown attribute value %s",
  /* E_ILLOP   -24 */  "illegal comparison operator %s",
};
#endif
#define MSGCNT  (int)(sizeof(errmsgs)/sizeof(const char*))

#endif  /* #ifdef RS_PARSE */
/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static int _valcmp (const void *p1, const void *p2)
{                               /* --- compare two attribute values */
  if (*(const int*)p1 < *(const int*)p2) return -1;
  if (*(const int*)p1 > *(const int*)p2) return  1;
  return 0;                     /* return sign of difference */
}  /* _valcmp() */

/*--------------------------------------------------------------------*/

static int _condset (COND *cond, int att, int op, INST *val)
{                               /* --- set a rule condition */
  int i;                        /* loop variable, number of values */
  int *sv, *dv, *tmp;           /* to traverse attribute values */
  unsigned int bits;            /* bitflag rep. of a value set */

  assert(cond && val);          /* check the function arguments */
  if (op != R_IN) {             /* if there is no value vector, */
    cond->op  = op;             /* simply copy the operator */
    cond->val = *val; }         /* and the attribute value */
  else {                        /* if there is a value vector, */
    sv = val->p; i = *sv;       /* get vector and number of values */
    if (i == 1) {               /* if there is only one value */
      cond->op = R_EQ; cond->val.i = sv[1]; }
    else {                      /* if there is more than one value */
      for (bits = 0, sv++; --i >= 0; ) {
        if (*sv > MSBIT) break; /* traverse the value vector */
        bits |= 1 << *sv++;     /* and try to convert it to */
      }                         /* a bitflag representation */
      if (i < 0) {              /* if conversion was successful */
        cond->op = R_BF; cond->val.i = bits; }
      else {                    /* if conversion was not successful */
        sv = val->p; i = *sv+1; /* get vector and number of values */
        dv = tmp = malloc(i *sizeof(int));
        if (!dv) return -1;     /* allocate a value vector */
        do { *dv++ = *sv++; } while (--i >= 0);
        qsort(tmp+1, *tmp, sizeof(int), _valcmp);
        cond->op = R_IN; cond->val.p = tmp;
      }                         /* copy and sort the given */
    }                           /* value vector and set it */
  }                             /* as the value of the condition */
  cond->att = att;              /* set attribute identifier */
  return 0;                     /* return 'ok' */
}  /* _condset() */

/*--------------------------------------------------------------------*/

static int _condcopy (COND *dst, COND *src)
{                               /* --- set a rule condition */
  int i;                        /* loop variable, number of values */
  int *sv, *dv;                 /* to traverse attribute values */

  assert(dst && src);           /* check the function arguments */
  if (src->op != R_IN) {        /* if there is no value vector, */
    *dst = *src; return 0; }    /* simply copy the condition */
  sv = src->val.p; i = *sv;     /* get vector and number of values */
  dv = malloc((i+1) *sizeof(int));
  if (!dv) return -1;           /* allocate a value vector */
  dst->att   = src->att;        /* set attribute identifier, */
  dst->op    = src->op;         /* comparison operator, */
  dst->val.p = dv;              /* and new value vector */
  do { *dv++ = *sv++; } while (--i >= 0);
  return 0;                     /* copy value vector and return 'ok' */
}  /* _condcopy() */

/*--------------------------------------------------------------------*/

static int _condcmp (const void *p1, const void *p2)
{                               /* --- compare two conditions */
  const COND *c1 = p1, *c2 = p2;/* conditions to compare */

  if (c1->att < c2->att) return -1;
  if (c1->att > c2->att) return  1;
  if (c1->op  < c2->op)  return  1;
  if (c1->op  > c2->op)  return -1;
  return 0;                     /* compare attribute and operator */
}  /* _condcmp() */

/*--------------------------------------------------------------------*/

static int _isect (COND *dst, COND *src)
{                               /* --- intersect two conditions */
                                /*     with a subset in one of them */
  int sn,  dn;                  /* numbers of values in subsets */
  int *sv, *dv;                 /* to traverse value vectors */
  unsigned int bits;            /* bitflag rep. of a value set */

  assert(dst && src);           /* check the function arguments */

  /* --- only source contains a subset --- */
  if (dst->op == R_EQ) {        /* if dest. equal to some value */
    if (src->op == R_BF)        /* if bitflag rep. in source, */
      return (src->val.i & (1 << dst->val.i));   /* check bit */
    sv = src->val.p; sn = *sv++;/* if vector rep. in source */
    return (bsearch(&dst->val.i, sv, sn, sizeof(int),_valcmp) != NULL);
  }                             /* search for the value */
  if (dst->op == R_NE) {        /* if dest. not equal to some value */
    if (src->op == R_BF) {      /* if bitflag rep. in source, */
      dst->op    = R_BF;        /* remove value from the value set */
      dst->val.i = src->val.i & ~(1 << dst->val.i);
      return (dst->val.i != 0); /* return whether the value set */
    }                           /* without the value is not empty */
    sv = src->val.p;            /* if vector rep. in source */
    return ((*sv > 1) || (sv[1] != dst->val.i));
  }                             /* check for remaining values */

  /* --- only destination contains a subset --- */
  if (src->op == R_EQ) {        /* if source equal to some value */
    if (dst->op == R_BF) {      /* if bitflag rep. in destination */
      dst->val.i &= 1 << src->val.i;     /* remove other values from */
      if (dst->val.i == 0) return 0; }   /* the set and check result */
    else {                      /* if vector rep. in destination */
      dv = dst->val.p;          /* search the vector for the value */
      if (bsearch(&src->val.i, dv+1, *dv, sizeof(int),_valcmp) == NULL)
        return *dv = 0;         /* if value not found, empty */
      free(dv);                 /* the value vector and abort, */
    }                           /* otherwise delete the value vector */
    dst->op    = R_EQ;          /* replace the destination condition */
    dst->val.i = src->val.i;    /* by the source condition, which is */
    return 1;                   /* stronger, and return 'compatible' */
  }
  if (src->op == R_NE) {        /* if source not equal to some value */
    if (dst->op == R_BF) {      /* if bitflag rep. in destination */
      dst->val.i &= ~(1 << src->val.i);
      return (dst->val.i != 0); /* remove value from the set */
    }                           /* and check for an empty set */
    dv = dst->val.p; dn = *dv;  /* search the value vector */
    sv = bsearch(&src->val.i, dv+1, dn, sizeof(int), _valcmp);
    if (!sv) return 1;          /* if value not found, abort */
    for (dn -= (int)(sv -dv); --dn >= 0; sv++)
      *sv = sv[1];              /* remove value from the vector */
    return (--(*dv) > 0);       /* return whether the resulting */
  }                             /* value set is not empty */

  /* --- both conditions contain subsets --- */
  if ((src->op == R_BF)         /* if bitflag representation */
  ||  (dst->op == R_BF)) {      /* in source or in destination */
    if (dst->op == R_IN) {      /* if vector rep. in destination, */
      dv = src->val.p;          /* traverse value vector of dest. */
      for (bits = 0, dn = *dv++; --dn >= 0; )
        bits |= 1 << *dv++;     /* set bits corresponding to values */
      dst->op = R_BF; dst->val.i = bits;
    }                           /* (convert dest. to bitflag rep.) */
    if (src->op == R_BF)        /* if bitflag rep. in source, */
      bits = src->val.i;        /* get the bitflag representation */
    else {                      /* if vector rep. in source, */
      sv = src->val.p;          /* traverse value vector of source */
      for (bits = 0, sn = *sv++; --sn >= 0; )
        bits |= 1 << *sv++;     /* set bits corresponding to values */
    }                           /* (convert source to bitflag rep.) */
    dst->val.i &= bits;         /* intersect bitflag representations */
    return (dst->val.i != 0);   /* return whether the intersection */
  }                             /* of the value sets is not empty */

  sv = src->val.p; sn = *sv++;  /* vectors in source and destination */
  dv = dst->val.p; dn = *dv++;  /* traverse both value vectors */
  if ((dn <= 0) || (sn <= 0))   /* if either set is empty, */
    return *--dv = 0;           /* the intersection is empty */
  while (1) {                   /* intersection loop */
    if      (*sv < *dv) {       /* source value not in destination */
      sv++;       if (--sn <= 0) break; }  /* skip source value */
    else if (*sv > *dv) {       /* destination value not in source */
      *dv++ = -1; if (--dn <= 0) break; }  /* mark dest. value */
    else {                      /* value in source and destination */
      sv++, dv++; if ((--dn <= 0) || (--sn <= 0)) break; }
  }                             /* skip both values */
  sv = dv = dst->val.p;         /* traverse destination again */
  for (dn = *sv++ -= dn; --dn >= 0; sv++)
    if (*sv >= 0) *++dv = *sv;  /* group together remaining values, */
  sv  = dst->val.p;             /* set new number of values, and */
  *sv = (int)(dv -sv);          /* return whether the resulting */
  return (*sv > 0);             /* value set is not empty */
}  /* _isect() */

/*----------------------------------------------------------------------
  Rule Functions
----------------------------------------------------------------------*/

RULE* r_create (ATTSET *attset, int maxcond)
{                               /* --- create a rule */
  RULE *rule;                   /* created rule */

  assert(attset && (maxcond >= 0)); /* check the function arguments */
  rule = (RULE*)malloc(sizeof(RULE) +(maxcond-1) *sizeof(COND));
  if (!rule) return NULL;       /* allocate memory and */
  rule->attset   = attset;      /* initialize fields */
  rule->set      = NULL; rule->id = -1;
  rule->supp     = rule->conf = 0;
  rule->condvsz  = maxcond;
  rule->condcnt  =  0;
  rule->head.att = -1;          /* invalidate head attribute */
  rule->head.op  =  0;          /* and comparison operator */
  return rule;                  /* return the created rule */
}  /* r_create() */

/*--------------------------------------------------------------------*/

void r_delete (RULE *rule)
{                               /* --- delete a rule */
  assert(rule);                 /* check the function argument */
  if (rule->set)                /* remove rule from containing set */
    rs_rulerem(rule->set, rule->id);
  if (rule->head.op == R_IN)    /* delete value vetors */
    free(rule->head.val.p);     /* from the head and */
  r_condrem(rule, -1);          /* from all conditions */
  free(rule);                   /* delete rule body */
}  /* r_delete() */

/*--------------------------------------------------------------------*/

int r_copy (RULE *dst, RULE *src)
{                               /* --- copy a rule */
  int  i;                       /* loop variable */
  COND *sc, *dc;                /* to traverse the conditions */

  assert(src && dst             /* check the function arguments */
     && (dst->condvsz >= src->condcnt));
  if (dst->head.op == R_IN)     /* delete value vectors */
    free(dst->head.val.p);      /* from the head and */
  r_condrem(dst, -1);           /* from all conditions */
  if (_condcopy(&dst->head, &src->head) != 0)
    return -1;                  /* copy the rule head */
  dst->supp    = src->supp;     /* copy rule support */
  dst->conf    = src->conf;     /* and confidence */
  dst->condcnt = src->condcnt;  /* traverse conditions */
  sc = src->conds; dc = dst->conds;
  for (i = src->condcnt; --i >= 0; dc++, sc++)
    if (_condcopy(dc, sc) != 0) { dst->condcnt -= i+1; return -1; }
  return 0;                     /* copy conditions and return 'ok' */
}  /* r_copy() */

/*--------------------------------------------------------------------*/

int r_headset (RULE *rule, int att, int op, INST *val)
{                               /* --- set the head of a rule */
  void *vec;                    /* buffer for value vector */
  INST dummy;                   /* dummy for head deletion */

  assert(rule                   /* check the function arguments */
     && (att >= 0) && (att < as_attcnt(rule->attset)));
  if (att < 0) { op = 0; val = &dummy; }
  vec = (rule->head.op == R_IN) ? rule->head.val.p : NULL;
  if (_condset(&rule->head, att, op, val) != 0)
    return -1;                  /* set new rule head and */
  if (vec) free(vec);           /* delete value vector (if any) */
  return 0;                     /* and return 'ok' */
}  /* r_headset() */

/*--------------------------------------------------------------------*/

int r_condadd (RULE *rule, int att, int op, INST *val)
{                               /* --- add a condition to a rule */
  int  i;                       /* loop variable */
  int  min, max;                /* min./max. condition to adapt */
  COND new;                     /* new condition (to add to the rule) */
  COND *cond, *dst;             /* to traverse conditions */
  int  cmp;                     /* result of value comparison */
  int  *set;                    /* buffer for value subset */
  unsigned int bits;            /* buffer for bitflags */

  assert(rule                   /* check the function arguments */
     && (att >= 0) && (att < as_attcnt(rule->attset)));
  if (_condset(&new, att, op, val) != 0)
    return -1;                  /* create new condition */

  /* --- check compatibility of new condition --- */
  min = -1; max = -2;           /* initialize condition indices */
  cond = rule->conds;           /* traverse conditions */
  for (i = 0; i < rule->condcnt; i++, cond++) {
    if (cond->att != att)       /* if condition refers to a */
      continue;                 /* different attribute, skip it */
    if ((new.op   >= R_IN)      /* if there is a subset operator */
    ||  (cond->op >= R_IN)) {   /* in one of the conditions */
      if (_isect(&new, cond) == 0) break; }
    else {                      /* if there is no subset operator */
      cmp = (new.val.f > cond->val.f) ? R_GT
          : (new.val.f < cond->val.f) ? R_LT : R_EQ;
      if (cmp == R_EQ) {        /* if values in conditions are equal, */
        new.op &= cond->op;     /* combine condition operators */
        if (new.op <= 0) break; }      /* and check the result */
      else {                    /* if values in conditions differ */
        if (cond->op & cmp) {   /* if new value in same direction */
          if (!(new.op & cmp))  /* if new operator has opp. dir., */
            continue; }         /* condition merging is not possible */
        else {                  /* if new value in opp. direction */
          if (new.op & cmp)     /* if new operator has opp. dir., */
            break;              /* the conditions are incompatible */
          new.op  = cond->op;   /* if the new operator has same dir., */
          new.val = cond->val;  /* the old condition is stronger, */
        }                       /* so replace the new condition */
      }                         /* (the new condition is modified to */
    }                           /* represent the combined conditions) */
    cond->op |= MARK;           /* mark condition to replace */
    if (min < 0) min = i;       /* and note condition index for */
    max = i;                    /* later adaption of the conditions */
  }
  if (i < rule->condcnt) {      /* if new condition is not compatible */
    cond = rule->conds;         /* remove all condition markers */
    for (i = min; i <= max; i++) (cond++)->op &= ~MARK;
    if (new.op == R_IN) free(new.val.p);
    return -2;                  /* delete a value vector rep. */
  }                             /* and abort the function */

  /* --- simplify new condition --- */
  if (new.op == R_BF) {         /* if bitflag rep. in new condition, */
    for (i = 0, bits = new.val.i; bits; i++, bits >>= 1)
      if (bits & 1) break;      /* find the index of the first value */
    if (bits && !(bits >> 1)) { /* if there is only one value */
      new.op = R_EQ; new.val.i = i; }
  }                             /* simplify condition */
  if (new.op == R_IN) {         /* if vector rep. in new condition, */
    set = new.val.p;            /* check the value vector */
    if (*set == 1) {            /* if there is only one value, */
      new.op = R_EQ; new.val.i = set[1];
      free(set); }              /* simplify cond. and delete vector */
    else {                      /* if there is more than one value, */
      for (bits = 0, i = *set++; --i >= 0; ) {
        if (*set > MSBIT) break;/* traverse the value vector */
        bits |= 1 << *set++;    /* and try to convert it to */
      }                         /* a bitflag representation */
      if (i < 0) {              /* if the conversion was successful, */
        free(new.val.p);        /* delete the value vector */
        new.op = R_BF; new.val.i = bits;
      }                         /* replace the condition by */
    }                           /* a subset test that uses */
  }                             /* a bitflag representation */

  /* --- add new condition --- */
  if (min < 0) {                /* if no adaptable condition found */
    if (rule->condcnt >= rule->condvsz)
      return -3;                /* check the number of conditions */
    rule->conds[rule->condcnt] = new;
    return rule->condcnt++;     /* insert new condition into the rule */
  }                             /* and return the condition index */

  /* --- adapt old condition(s) --- */
  cond = rule->conds +min;      /* get a pointer to the condition */
  cond->op &= ~MARK;            /* to adapt and remove the marker */
  if (new.op != R_NE) {         /* if to adapt condition, */
    if (cond->op == R_IN) free(cond->val.p);
    *cond = new;                /* delete an existing value vector */
  }                             /* and replace the condition */
  if (max <= min) return min;   /* if only one cond. to adapt, abort */
  dst = ++cond;                 /* traverse remaining conditions */
  for (i = rule->condcnt -min; --i > 0; cond++) {
    if (!(cond->op & MARK))     /* group together all */
      *dst++ = *cond;           /* unmarked conditions */
    else if ((cond->op & ~MARK) == R_IN)
      free(cond->val.p);        /* delete an existing value vector */
  }                             /* from the marked conditions */
  rule->condcnt = (int)(dst -rule->conds);
  return min;                   /* set new number of conditions and */
}  /* r_condadd() */            /* return index of changed condition */

/*--------------------------------------------------------------------*/

void r_condrem (RULE *rule, int index)
{                               /* --- remove a cond. from a rule */
  int  i;                       /* loop variable */
  COND *cond;                   /* to traverse the conditions */

  assert(rule                   /* check the function arguments */
     && (index < rule->condcnt));
  if (index < 0) {              /* if to remove all conditions, */
    cond = rule->conds;         /* delete value vectors */
    for (i = rule->condcnt; --i >= 0; cond++)
      if (cond->op == R_IN) free(cond->val.p);
    rule->condcnt = 0; }        /* clear condition counter */
  else {                        /* if to remove one condition */
    cond = rule->conds +index;  /* get the condition and delete */
    if (cond->op == R_IN) free(cond->val.p);    /* value vector */
    for (i = --rule->condcnt -index; --i >= 0; cond++)
      *cond = cond[1];          /* shift conditions to fill the gap */
  }                             /* left by the deleted condition */
}  /* r_condrem() */

/*--------------------------------------------------------------------*/

void r_condsort (RULE *rule)
{                               /* --- sort rule conditions */
  assert(rule);                 /* check the function argument */
  qsort(rule->conds, rule->condcnt, sizeof(COND), _condcmp);
}  /* r_condsort() */           /* sort the rule conditions */

/*--------------------------------------------------------------------*/

int r_check (RULE *rule)
{                               /* --- check applicability of a rule */
  int    i, t;                  /* loop variable, attribute type */
  COND   *cond;                 /* to traverse the conditions */
  ATT    *att;                  /* to traverse the attributes */
  INST   *val;                  /* to traverse the values */
  double f;                     /* buffer for comparisons */
  int    *p;                    /* to traverse an identifier vector */

  assert(rule);                 /* check the function argument */
  for (cond = rule->conds, i = rule->condcnt; --i >= 0; cond++) {
    att = as_att(rule->attset, cond->att);
    val = att_inst(att);        /* traverse the conditions and */
    t   = att_type(att);        /* get the corresponding attribute */
    if (t == AT_SYM) {          /* if the attribute is symbolic */
      if      (cond->op == R_EQ) {  /* equality to a value */
        if (cond->val.i != val->i) return i; }
      else if (cond->op == R_BF) {  /* bit flag represented set */
        if ((val->i >= 32) || !(cond->val.i & (1 << val->i)))
          return i; }           /* check whether the bit is set */
      else {                    /* identifier vector represented set */
        for (t = *(p = cond->val.p); --t >= 0; )
          if (*++p == val->i) break;
        if (t < 0) return i;    /* traverse the identifiers and */
      } }                       /* try to find the value */
    else {                      /* if the attribute is numeric */
      if (t == AT_INT) f = (double)val->i;  /* get the value from */
      else             f = (double)val->f;  /* the attribute set */
      switch (cond->op) {       /* evaluate the comparison operator */
        case R_EQ: if (f != cond->val.f) return i; break;
        case R_LT: if (f >= cond->val.f) return i; break;
        case R_GT: if (f <= cond->val.f) return i; break;
        case R_LE: if (f >  cond->val.f) return i; break;
        case R_GE: if (f <  cond->val.f) return i; break;
        default  : if (f == cond->val.f) return i; break;
      }                         /* if the comparison fails, */
    }                           /* return the index of */
  }                             /* the condition that fails */
  return -1;                    /* return 'rule is applicable' */
}  /* r_check() */

/*--------------------------------------------------------------------*/
#ifdef RS_PARSE

static int _paerr (SCAN *scan, int code, int c, const char *s)
{                               /* --- report a parse error */
  char src[  AS_MAXLEN+1];      /* buffer for string to format */
  char dst[4*AS_MAXLEN+4];      /* buffer for fomatted string */

  assert(scan);                 /* check the function arguments */
  strncpy(src, sc_value(scan), AS_MAXLEN); src[AS_MAXLEN] = '\0';
  sc_format(dst, src, 1);       /* copy and format the token value */
  if      (code == E_CHREXP) return sc_error(scan, code, c, dst);
  else if (code == E_STREXP) return sc_error(scan, code, s, dst);
  else                       return sc_error(scan, code,    dst);
}  /* _paerr() */               /* print an error message */

/*--------------------------------------------------------------------*/

static int _valset (ATT *att, SCAN *scan, RULE *rule)
{                               /* --- read a set of values */
  int  t;                       /* buffer for a token value */
  int  id, op;                  /* identifier and comparison operator */
  INST val;                     /* to represent the condition value */
  int  *set = NULL;             /* set of values read */

  assert(att && scan && rule);  /* check the function arguments */
  t  = att_valcnt(att);         /* get the number of values and */
  if (t <= 32) {                /* decide on the set representation */
    op  = R_BF; val.i = 0; }    /* if to use bit flags, clear bits */
  else {                        /* if to use a vector of values, */
    op  = R_IN;                 /* allocate an identifier vector */
    set = (int*)malloc((t+1) *sizeof(int));
    if (!set) return ERROR(E_NOMEM);
    set[0] = 0; val.p = set;    /* initialize the number of values */
  }                             /* (vector format: [cnt v1 v2 ... ]) */
  GET_CHR('{');                 /* consume '{' */
  while (1) {                   /* value read loop */
    t = sc_token(scan);         /* check for an attribute value */
    if ((t != T_ID) && (t != T_NUM)) return ERROR(E_VALEXP);
    id = att_valid(att, sc_value(scan));
    if (id < 0) return ERROR(E_UNKVAL);
    GET_TOK();                  /* get, check, and consume the value */
    if (op == R_BF)             /* if the set is rep. by bit flags, */
      val.i |= 1 << id;         /* set the bit corresp. to the value */
    else {                      /* if the set is rep. by a vector, */
      for (t = *set; --t >= 0;) /* check whether the value exists */
        if (set[t+1] == id) break;
      if (t < 0) { t = *set++; set[t] = id; }
    }                           /* add only new values */
    if (sc_token(scan) != ',')  /* add a condition to the rule */
      break;                    /* and check for another value */
    GET_TOK();                  /* consume ',' */
  }                           
  GET_CHR('}');                 /* consume '}' */
  r_condadd(rule, att_id(att), op, &val);
  if (set) free(set);           /* add the condition, delete the */
  return 0;                     /* identifier vector, and return 'ok' */
}  /* _valset() */

/*--------------------------------------------------------------------*/

static int _rule (ATTSET *attset, SCAN *scan, RULE *rule)
{                               /* --- parse a rule */
  int    id, op;                /* identifier and comp. operator */
  ATT    *att;                  /* attribute of a condition */
  INST   val;                   /* attribute value */
  int    head = 1;              /* flag for the rule head */
  double supp, conf;            /* support and confidence */
  int    t;                     /* buffer for a token value */

  assert(attset && scan && rule);  /* check the function arguments */
  while (1) {                   /* condition read loop */
    t = sc_token(scan);         /* check for an attribute */
    if ((t != T_ID) && (t != T_NUM)) return ERROR(E_ATTEXP);
    id = as_attid(attset, sc_value(scan));
    if (id < 0) return ERROR(E_UNKATT);    /* get the attribute id */
    att = as_att(attset, id);   /* and the corresponding attribute */
    GET_TOK();                  /* consume the attribute name */
    t = sc_token(scan);         /* check for a comparison operator */
    if ((t != '=')   && (t != '<') && (t != '>')
    &&  (t != T_CMP) && (t != T_ID)) ERR_CHR('=');
    for (op = R_IN +1; --op >= 0; )
      if (_opsigns[op] && (strcmp(_opsigns[op], sc_value(scan)) == 0))
        break;                  /* encode the comparison operator */
    if ((op < 0) || (head && (op != R_EQ)))
      ERR_CHR('=');             /* check for a valid comp. operator */
    if (att_type(att) == AT_SYM) { /* if the attribute is symbolic */
      if ((op != R_IN) && (op != R_EQ)) return ERROR(E_ILLOP);
      GET_TOK();                /* consume the comparison operator */
      if (op == R_IN) {         /* if to read a set of values */
        t = _valset(att, scan, rule); if (t) return t; }
      else {                    /* if to read a single value */
        t = sc_token(scan);     /* check for a symbolic value */
        if ((t != T_ID) && (t != T_NUM)) return ERROR(E_VALEXP);
        val.i = att_valid(att, sc_value(scan));
        if (val.i < 0) return ERROR(E_UNKVAL);
        GET_TOK();              /* get, check, and consume the value */
        if (head) r_headset(rule, id, op, &val);/* set the rule head */
        else      r_condadd(rule, id, op, &val);/* or add a condition */
      } }
    else {                      /* if the attribute is numeric */
      if (op >= R_IN) return ERROR(E_ILLOP);
      GET_TOK();                /* consume the comparison operator */
      if (sc_token(scan) != T_NUM) return ERROR(E_NUMEXP);
      val.f = atof(sc_value(scan));
      GET_TOK();                /* get, check, and consume the number */
      if (head) r_headset(rule, id, op, &val); /* set the rule head */
      else      r_condadd(rule, id, op, &val); /* or add a condition */
    }
    if ((sc_token(scan) == ';') || (sc_token(scan) == '['))
      break;                    /* check for the end of the rule */
    if (head) { if (sc_token(scan) != T_LFT) ERR_STR("<-"); }
    else      { if (sc_token(scan) != '&')   ERR_CHR('&');  }
    GET_TOK();                  /* consume implication or conjunction */
    head = 0;                   /* the first round reads the head, */
  }                             /* all following only conditions */
  if (sc_token(scan) == '[') {  /* if support/confidence follows */
    GET_TOK(); t = 0;           /* consume '[' */
    if (sc_token(scan) == '~') { GET_TOK(); t = 1; }
    if (sc_token(scan) != T_NUM) return ERROR(E_NUMEXP);
    supp = atof(sc_value(scan));/* get the support/confidence */
    conf = 0;                   /* (set default confidence first) */
    GET_TOK();                  /* consume the number */
    if ((t == 0) && (sc_token(scan) == '%')) t = 2;
    if (t != 0) {               /* if there is no support value, */
      conf = supp; supp = 0;    /* set and check the confidence */
      if ((conf < 0) || (conf > 100)) {
        sc_back(scan); return ERROR(E_ILLNUM);}
      if (t == 2) GET_TOK(); }  /* consume '%' */
    else if (sc_token(scan) == '/') {
      GET_TOK();                /* consume the supp./conf. separator */
      if (sc_token(scan) == '~') { GET_TOK(); t = 1; }
      if (sc_token(scan) != T_NUM)    return ERROR(E_NUMEXP);
      conf = atof(sc_value(scan));
      if ((conf < 0) || (conf > 100)) return ERROR(E_ILLNUM);
      GET_TOK();                /* get, check and consume the conf. */
      if ((t == 0) && (sc_token(scan) == '%')) { GET_TOK(); t = 2; }
    }                           /* consume '%' */
    GET_CHR(']');               /* consume ']' */
    rule->supp = supp; rule->conf = (t == 2) ? 0.01 *conf : conf;
  }                             /* set support and confidence */
  GET_CHR(';');                 /* consume ';' */
  return 0;                     /* return 'ok' */
}  /* _rule() */

/*--------------------------------------------------------------------*/

RULE* r_parse (ATTSET *attset, SCAN *scan)
{                               /* --- parse a rule */
  RULE *rule;                   /* created rule */

  assert(attset && scan);       /* check the function arguments */
  sc_errmsgs(scan, errmsgs, MSGCNT);
  rule = r_create(attset, 2 *as_attcnt(attset));
  if (!rule) return NULL;       /* create a rule of maximum size */
  if (_rule(attset, scan, rule) != 0) {
    r_delete(rule); return NULL; }  /* parse a rule */
  rule = (RULE*)realloc(rule, sizeof(RULE)
                             +(rule->condcnt-1) *sizeof(COND));
  rule->condvsz = rule->condcnt;/* try to shrink the rule */
  return rule;                  /* return the created rule */
}  /* r_parse() */

#endif
/*----------------------------------------------------------------------
  Rule Set Functions
----------------------------------------------------------------------*/

RULESET* rs_create (ATTSET *attset, RULE_DELFN delfn)
{                               /* --- create a ruleset */
  RULESET *set;                 /* created rule set */

  assert(attset && delfn);      /* check arguments */
  set = (RULESET*)malloc(sizeof(RULESET));
  if (!set) return NULL;        /* allocate memory and */
  set->attset  = attset;        /* initialize fields */
  set->delfn   = delfn;
  set->rulevsz = set->rulecnt = 0;
  set->rules   = NULL;
  return set;                   /* return the created rule set */
}  /* rs_create() */

/*--------------------------------------------------------------------*/

void rs_delete (RULESET *set, int delas)
{                               /* --- delete a ruleset */
  int  i;                       /* loop variable */
  RULE **p;                     /* to traverse the rules */

  assert(set);                  /* check the function argument */
  p = set->rules +set->rulecnt; /* traverse rules */
  for (i = set->rulecnt; --i >= 0; ) {
    (*--p)->set = NULL; (*p)->id = -1;
    set->delfn(*p);             /* remove and delete all rules */
  }                             /* and delete the rule vector */
  if (set->rules) free(set->rules);
  if (delas) as_delete(set->attset);
  free(set);                    /* delete the rule set body */
}  /* rs_delete() */

/*--------------------------------------------------------------------*/

int rs_ruleadd (RULESET *set, RULE *rule)
{                               /* --- add a rule to a ruleset */
  int  vsz;                     /* size of rule vector */
  RULE **p;                     /* (new) rule vector */

  assert(set && rule);          /* check the function arguments */
  if (set->rulecnt >= set->rulevsz) {
    vsz = set->rulevsz          /* if the rule vector is full */
        + ((set->rulevsz > BLKSIZE) ? (set->rulevsz >> 1) : BLKSIZE);
    p = (RULE**)realloc(set->rules, vsz *sizeof(RULE*));
    if (!p) return -1;          /* allocate a (new) rule vector */
    set->rules = p; set->rulevsz = vsz;
  }                             /* set rule vector and its size */
  rule->set = set;              /* note rule set reference */
  rule->id  = set->rulecnt;     /* and set rule identifier */
  set->rules[set->rulecnt++] = rule;
  return 0;                     /* add the rule and return 'ok' */
}  /* rs_ruleadd() */

/*--------------------------------------------------------------------*/

RULE* rs_rulerem (RULESET *set, int ruleid)
{                               /* --- remove a rule from a ruleset */
  int  i;                       /* loop variable */
  RULE **p;                     /* to traverse the rule vector */
  RULE *rule;                   /* buffer for removed rule */

  assert(set && (ruleid < set->rulecnt));   /* check arguments */

  /* --- delete all rules --- */
  if (ruleid < 0) {             /* if no rule identifier given */
    for (p = set->rules +(i = set->rulecnt); --i >= 0; ) {
      (*--p)->set = NULL; (*p)->id = -1;
      set->delfn(*p);           /* remove and delete all rules */
    }
    if (set->rules) { free(set->rules); set->rules = NULL; }
    set->rulecnt = set->rulevsz = 0;
    return NULL;                /* delete the rule vector */
  }                             /* and abort the function */

  /* --- delete one rule --- */
  rule = set->rules[ruleid];    /* note the rule to remove */
  p    = set->rules +ruleid;    /* traverse the successor rules */
  for (i = --set->rulecnt -ruleid; --i >= 0; ) {
    *p = p[1]; (*p++)->id--; }  /* shift rules and adapt identifiers */
  rule->set = NULL;             /* clear the rule set reference */
  rule->id  = -1;               /* and the rule identifier */
  return rule;                  /* and return the removed rule */
}  /* rs_rulerem() */

/*--------------------------------------------------------------------*/

void rs_ruleexg (RULESET *set, int ruleid1, int ruleid2)
{                               /* --- exchange two rules */
  RULE **p1, **p2, *rule;       /* exchange buffer */

  assert(set && (ruleid1 >= 0) && (ruleid1 < set->rulecnt)
             && (ruleid2 >= 0) && (ruleid2 < set->rulecnt));
  p1 = set->rules +ruleid1; rule = *p1;
  p2 = set->rules +ruleid2;     /* get pointers to attributes, */
  *p1 = *p2; (*p1)->id = ruleid1;  /* exchange them, and */
  *p2 = rule; rule->id = ruleid2;  /* set new identifiers */
}  /* rs_ruleexg() */

/*--------------------------------------------------------------------*/

void rs_rulemove (RULESET *set, int off, int cnt, int pos)
{                               /* --- move some rules */
  int  n;                       /* temporary buffer */
  RULE **p;                     /* to traverse rules */

  assert(set);                  /* check for a valid attribute set */
  n = set->rulecnt;             /* check and adapt insert position */
  if (pos > n)      pos = n;    /* and number of attributes */
  if (cnt > n -off) cnt = n -off;
  assert((cnt >= 0) && (off >= 0)
      && (pos >= 0) && ((pos <= off) || (pos >= off +cnt)));
  v_move(set->rules, off, cnt, pos, (int)sizeof(RULE*));
  if (pos <= off) {             /* move rules in vector */
    cnt += off; off = pos; pos = cnt; }
  p = set->rules +off;          /* set new rule identifiers */
  while (off < pos) (*p++)->id = off++;
}  /* rs_rulemove() */

/*--------------------------------------------------------------------*/
#if 0

int rs_rulecut (RULESET *dst, RULESET *src, int mode, ...)
{                               /* --- cut/copy rules */
  return 0;                     /* not implemented yet */
}  /* rs_rulecut() */

#endif
/*--------------------------------------------------------------------*/

void rs_sort (RULESET *set, RULE_CMPFN cmpfn, void *data)
{                               /* --- sort a ruleset */
  int  i;                       /* loop variable */
  RULE **p;                     /* to traverse rules */

  assert(set && cmpfn);         /* check arguments */
  v_sort(set->rules, set->rulecnt, (VCMPFN*)cmpfn, data);
  p = set->rules +set->rulecnt; /* sort rules and set new identifiers */
  for (i = set->rulecnt; --i >= 0; ) (*--p)->id = i;
}  /* rs_sort() */

/*--------------------------------------------------------------------*/

int rs_exec (RULESET *set)
{                               /* --- find first applicable rule */
  int i;                        /* loop variable */

  assert(set);                  /* check the function argument */
  for (i = 0; i < set->rulecnt; i++)
    if (r_check(set->rules[i]) < 0) return i;
  return -1;                    /* traverse and check the rules */
}  /* rs_exec() */

/*--------------------------------------------------------------------*/
#ifdef RS_DESC

static void _condout (DESC *desc, COND *cond, int op)
{                               /* --- print a condition */
  ATT  *att;                    /* attribute of the condition */
  int  i, bit;                  /* loop variable, bit index */
  int  *val;                    /* to traverse attribute values */
  int  len;                     /* length of attribute/value name */
  char buf[8 *AS_MAXLEN +64];   /* output buffer */
  char *s = buf;                /* to traverse the output buffer */
  const char *t;                /* to traverse the operator sign */

  if      (op >  0) { *s++ = ' '; *s++ = '&';             *s++ = ' '; }
  else if (op == 0) { *s++ = ' '; *s++ = '<'; *s++ = '-'; *s++ = ' '; }
                                /* store logical operator (if any) */
  att  = as_att(desc->attset, cond->att);
  s   += sc_format(s, att_name(att), 0);
  *s++ = ' ';                   /* get and format attribute name */
  for (t = _opsigns[cond->op]; *t; )
    *s++ = *t++;                /* store the condition operator */
  *s++ = ' ';                   /* and a separating blank */
  if (att_type(att) != AT_SYM)  /* if the attribute is numeric, */
    s += sprintf(s, "%g",  cond->val.f);     /* store its value */
  else if (cond->op >= R_IN) {  /* if subset test on symbolic att., */
    *s++ = '{'; *s = '\0'; }    /* start an attribute value set */
  else {                        /* if simple symbolic comparison */
    s += sc_format(s, att_valname(att, cond->val.i), 0);
  }                             /* get and format attribute value */
  if ((desc->pos > 4)           /* if the line would get too long, */
  &&  (desc->pos +(s -buf) >= desc->max)) {    /* start a new line */
    fputs((op > 0) ? "\n  " : "\n ", desc->file); desc->pos = 1; }
  fputs(buf, desc->file);       /* print output buffer contents */
  desc->pos += (int)(s -buf);   /* compute new position */
  if (cond->op < R_IN) return;  /* if this is no subset test, abort */

  s = NULL; buf[0] = ' ';       /* use string as a first value flag */
  if (cond->op == R_IN) {       /* if vector representation */
    val = cond->val.p;          /* traverse value vector */
    for (i = *val++; --i >= 0; val++) {
      if (s) {                  /* if this is not the first value, */
        fputc(',', desc->file); desc->pos++;} /* print a separator */
      len = sc_format(s = buf+1, att_valname(att, *val), 0);
      if ((desc->pos > 4)       /* if the line would get too long, */
      &&  (desc->pos +len +3 > desc->max)) {   /* start a new line */
        fputs("\n    ", desc->file); desc->pos = 4; }
      fputs(buf, desc->file);   /* print attribute name */
      desc->pos += len +1;      /* and compute new position */
    } }
  else {                        /* if bitflag representation */
    for (bit = 1, i = 0; bit; i++, bit <<= 1) {
      if (!(cond->val.i & bit)) /* traverse bit flags */
        continue;               /* and skip cleared bits */
      if (s) {                  /* if this is not the first value, */
        fputc(',', desc->file); desc->pos++;} /* print a separator */
      len = sc_format(s = buf+1, att_valname(att, i), 0);
      if ((desc->pos > 4)       /* if the line would get too long, */
      &&  (desc->pos +len +3 > desc->max)) {  /* start a new line */
        fputs("\n    ", desc->file); desc->pos = 4; }
      fputs(buf, desc->file);   /* print attribute name */
      desc->pos += len +1;      /* and compute new position */
    }
  }
  fputs(" }", desc->file);      /* terminate value set */
  desc->pos += 2;               /* and compute new position */
}  /* _condout() */

/*--------------------------------------------------------------------*/

int rs_desc (RULESET *set, FILE *file, int mode, int maxlen)
{                               /* --- describe a rule set */
  int  i, k;                    /* loop variables */
  DESC desc;                    /* description info */
  RULE **rule;                  /* to traverse the rule vector */
  COND *cond;                   /* to traverse a condition vector */
  int  op;                      /* logical operator */
  char buf[64];                 /* output buffer */
  char *s;                      /* to traverse the output buffer */

  assert(set && file);          /* check arguments */

  /* --- print header (as a comment) --- */
  if (mode & RS_TITLE) {        /* if the title flag is set */
    i = k = (maxlen > 0) ? maxlen -2 : 70;
    fputs("/*", file); while (--i >= 0) fputc('-', file);
    fprintf(file, "\n  rules\n");
    while (--k >= 0) fputc('-', file); fputs("*/\n", file);
  }                             /* print title header */
  desc.file   = file;           /* note the output file */
  desc.attset = set->attset;    /* and the attribute set */
  desc.max    = (maxlen <= 0) ? INT_MAX : maxlen;

  /* --- print a list of rules --- */
  rule = set->rules;            /* traverse the rules */
  for (i = set->rulecnt; --i >= 0; rule++) {
    if ((*rule)->head.att < 0) {/* if there is no rule head, */
      fputc('#', file); desc.pos = 1; } /* print a special mark */
    else {                      /* if there is a rule head */
      _condout(&desc, &(*rule)->head,-1);
      if (mode & RS_CONDLN) { fputs("\n ", file); desc.pos = 1; }
    }                           /* print head condition */
    op   = 0;                   /* set implication first */
    cond = (*rule)->conds;      /* traverse conditions */
    for (k = (*rule)->condcnt; --k > 0; cond++) {
      _condout(&desc, cond, op);
      op = 1;                   /* print a condition, switch to and */
      if (mode & RS_CONDLN) { fputs("\n  ", file); desc.pos = 2; }
    }                           /* in line mode start a new line */
    if (k >= 0)                 /* print last condition (if any, */
      _condout(&desc, cond, op);/* there could be no cond. at all) */
    if (mode & (RS_SUPP|RS_CONF)) {
      s = buf;                  /* if to print support or confidence */
      *s++ = ' '; *s++ = '[';   /* start additional information */
      if (mode & RS_SUPP)       /* if to print the support */
        s += sprintf(s, "%g", (*rule)->supp);
      if (mode & RS_CONF) {     /* if to print the confidence */
        if (mode & RS_SUPP) *s++ = '/';
        k = (*rule)->head.att;  /* check the head attribute */
        if ((k >= 0) && (att_type(as_att(set->attset, k)) != AT_SYM))
          s += sprintf(s, "~%g",    (*rule)->conf);
        else                    /* numeric: print rmse as confidence */
          s += sprintf(s, "%.1f%%", (*rule)->conf *100);
      }                         /* symbolic: print conf. in percent */
      *s++ = ']'; *s = '\0';    /* terminate additional information */
      k = (int)(s -buf);        /* and compute its length */
      if (desc.pos+k > desc.max)/* if the line would get too long, */
        fputs("\n  ", file);    /* start a new line and indent */
      fputs(buf, file);         /* print support and confidence */
    }
    fputs(";\n", file);         /* terminate the output line */
    desc.pos = 0;               /* and reset the position */
  }

  /* --- print additional information (as a comment) --- */
  if (mode & RS_INFO) {         /* if the add. info. flag is set */
    i = k = (maxlen > 0) ? maxlen -2 : 70;
    fputs("\n/*", file); while (--i >= 0) fputc('-', file);
    i = as_attcnt(set->attset);
    fprintf(file, "\n  number of attributes: %d",   i);
    fprintf(file, "\n  number of rules     : %d\n", set->rulecnt);
    while (--k >= 0) fputc('-', file); fputs("*/\n", file);
  }                             /* print add. rule set information */
  return ferror(file) ? -1 : 0; /* check for a write error */
}  /* rs_desc() */

#endif  /* #ifdef RS_DESC */
/*--------------------------------------------------------------------*/
#ifdef RS_PARSE

RULESET* rs_parse (ATTSET *attset, SCAN *scan)
{                               /* --- parse a rule set */
  RULESET *set;                 /* created rule set */
  RULE    *rule;                /* to read/traverse the rules */
  int     err = 0;              /* error flag */

  assert(attset && scan);       /* check the function arguments */
  set = rs_create(attset, r_delete);
  if (!set) return NULL;        /* create a rule set */
  while (sc_token(scan) == T_ID) {    /* rule read loop */
    rule = r_parse(attset, scan);     /* parse a rule and */
    if (rule) rs_ruleadd(set, rule);  /* add it to the rule set */
    else { err = 1;             /* try to recover after an error */
      if (sc_recover(scan, ';', 0, 0, 0) == T_EOF) break; }
  }                             
  if (err) { rs_delete(set, 0); return NULL; }
  return set;                   /* return the created rule set */
}  /* rs_parse() */

#endif  /* #ifdef RS_PARSE */
