#!/usr/bin/env python
#
# Gnome Data Miner: GNOME interface to Borgelt's apriori
#
# Copyright (c) 2003-2004 Graham J Williams
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version. See the file gpl-license.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Authors: Graham.Williams@togaware.com
#

"""A Gnome GUI to interface the command line APRIORI associations tool.

Essentially, the aim is to capture in the interface all options
available with the APRIORI program. You can then also see the
resulting command line through viewing the log file. All options are
also documented through the interface.

Command line usage:

  python gdmapriori.py [options] [<datafile>]

  -h|--help	Print this help message.
  -v|--version	Print the current version number.

"""

########################################################################
#
# Libraries
#
import os
import sys
import popen2
import tempfile
try:
    import gtk
    import gtk.glade
except ImportError, e:
    print e
    print "Please install python-glade2 and python-gnome2."
    sys.exit(1)
import glutil
########################################################################
#
# Global Constants
#
from const import version

measure_map = {
    "Absolute Confidence Difference to Prior"			: "d",
    "Difference of Confidence Quotient to 1"			: "q",
    "Absolute Difference of Improvement Value to 1"		: "a",
    "Information Difference to Prior"				: "i",
    "Normalized Chi^2 Measure"					: "c"
    }
########################################################################
#
# Apriori App
#
class AprioriApp(glutil.GnomeApp):

    def __init__(self, trans="", app=""):
        glutil.GnomeApp.__init__(self,
                                 "gdmaprior",
                                 "Gnome Data Miner: Apriori",
                                 version,
                                 glutil.appdir("gdmapriori.glade"),
                                 "apriori_app")
        
        self.default_data_ext  = "tab"
        #
        # Temporary Files
        #
        self.rul_file = tempfile.mktemp(".rul")
        self.log_file = tempfile.mktemp(".gdmapriori-log")

        if trans != "":
            if os.path.splitext(trans)[1] == "" \
               and not os.path.exists(trans) \
               and os.path.exists(trans + ".tab"):
                trans += ".tab"
                                  
            self.trans_file_entry.gtk_entry().set_text(trans)
            self.trans_file_entry.gnome_entry().append_history(1, trans)

        if app != "":
            self.app_file_entry.gtk_entry().set_text(app)
            self.app_file_entry.gnome_entry().append_history(1, app)

    #
    # Command line generation and execution
    #

    def get_dataset_options(self):
        command = ""
        #
        # Characters to be treated as blank
        #
        blanks = self.blank_entry.get_text()
        if blanks != "": command += "-b\"" + blanks + "\" "
        #
        # Dataset field separator
        #
        field_separator = self.field_separator_entry.get_text()
        if field_separator != "": command += "-f\"" + field_separator + "\" "
        #
        # Dataset record separator
        #
        record_separator = self.record_separator_entry.get_text()
        if record_separator != "": command += "-r\"" + record_separator + "\" "
        #
        # Comment marker
        #
        comment = self.comment_entry.get_text()
        if comment != "": command += "-i\"" + comment + "\" "
        #
        # All done
        #
        return(command)
        

    def execute_apriori(self):
        #
        # Relevant files
        #
        trans_file = self.trans_file_entry.gtk_entry().get_text()
        app_file = self.app_file_entry.gtk_entry().get_text()

        if trans_file == "":
            glutil.WarningDialog(self,
                          "Please enter the name of a transaction file first.")
        elif not os.path.exists(trans_file):
            glutil.ErrorDialog(self,
                        "Transaction file could not be found: " + trans_file)
        elif app_file != "" and not os.path.exists(app_file):
            glutil.ErrorDialog(self,
                        "Appearances file could not be found: " + app_file)
        else:
            #
            # BUILD THE COMMAND
            #
            command = "apriori "
            #
            # Add dataset options
            #
            command += self.get_dataset_options()
            #
            # Print absolute support (number of transactions)
            #
            if self.support_count_checkbutton.get_active(): command += "-a "
            #
            # Minimum confidence
            #
            command += "-c%d " % self.confidence_spinbutton.get_value()
            #
            # Determine whether an extra measure is to be used.
            #
            extra_measure = self.extra_combo.entry.get_text()
            if extra_measure != "":
                command += "-e" + measure_map[extra_measure] + " "
                command += "-d%d " % self.extra_value_spinbutton.get_value()
            #
            # Write output in scanable form quoting certain characters
            #
            if self.scanable_checkbutton.get_active(): command += "-g "
            #
            # Do not load transactions into memory (work on input file)
            #
            if not self.load_transactions_checkbutton.get_active():
                command += "-l "
            #
            # Minimum number of items
            #
            command += "-m%d " % self.min_items_spinbutton.get_value()
            #
            # Maximum number of items
            #
            command += "-n%d " % self.max_items_spinbutton.get_value()
            #
            # Use original definition of the support of a rule (body & head)
            #
            if self.original_support_checkbutton.get_active(): command += "-o "
            #
            # Output format
            #
            output_format = self.output_format_entry.get_text()
            if output_format != "":
                command += "-p\"%s\" " % output_format
            #
            # Sort
            #
            if self.descend_radiobutton.get_active(): command += "-q1 "
            if self.ascend_radiobutton.get_active(): command += "-q-1 "
            if self.none_radiobutton.get_active(): command += "-q0 "
            #
            # Minimum support
            #
            command += "-s%.2f " % self.support_spinbutton.get_value()
            #
            # Target
            #
            if self.rules_radiobutton.get_active(): command += "-tr "
            if self.itemsets_radiobutton.get_active(): command += "-ts "
            if self.hyperedges_radiobutton.get_active(): command += "-th "
            #
            # Write additional measure value
            #
            if self.evaluation_output_checkbutton.get_active():
                command += "-v "
            #
            # Write extended support output
            #
            if self.extended_support_checkbutton.get_active(): command += "-x "
            #
            # Write extended support output
            #
            if self.memory_checkbutton.get_active(): command += "-z "
            #
            # Add file names to command line
            #
            command += trans_file + " " + self.rul_file + " " + app_file
            #
            # Trace of command being executed
            #
            glutil.start_log(self.log_file, command)
            #
            # Redirect output to a file
            #
            command += " 2>> " + self.log_file
            #
            # RUN THE COMMAND
            #
            # print command
            os.system(command)
            glutil.end_log(self.log_file)
            #
            # Display the output
            #
            glutil.FileViewWindow(self.rul_file,
                           "Associtaions View",
                           os.path.splitext(trans_file)[0] + ".rul")

    def on_view_log_button_clicked(self, *extra):
        if not os.path.exists(self.log_file):
            glutil.WarningDialog(self,
                                "No log file yet.")
        else:
            glutil.FileViewWindow(self.log_file, "Log File", "", clearable=1)

    def cleanup(self):
        if os.path.isfile(self.rul_file): os.remove(self.rul_file)
        if os.path.isfile(self.log_file): os.remove(self.log_file)

    #
    # global
    #
    def on_help_about_activate(self, *extra):
        aboutxml = gtk.glade.XML(glutil.appdir("gdmapriori.glade"),"about")
        about = aboutxml.get_widget("about")
        about.set_property("name", "Gnome Data Miner: Apriori")
        about.set_property("version", version)
        about.show()
    def on_quit_activate(self, *extra):
        self.cleanup()
        self.quit()
    def on_quit_activate(self, *extra):
        self.cleanup()
        self.quit()

    #
    # implementation
    #

    def on_execute_button_clicked(self, *extra):
        self.execute_apriori()

    #
    # Buttons
    #

    def on_view_trans_button_clicked(self, *extra):
        trans_file = self.trans_file_entry.gtk_entry().get_text()
        if trans_file == "":
            glutil.WarningDialog(self,
                          "Please specify the filename for " +\
                          "the transaction file first. Often they have " +\
                          "a .tab extension")
        elif not os.path.exists(trans_file):
            glutil.WarningDialog(self, "The specified transaction file does not exist. " +\
                          "Please supply an existing transaction file. Often " +\
                          "they have a .tab extension")
        else:
            glutil.FileViewWindow(trans_file, "Transaction File", trans_file, saveable=0)

        
    def on_view_app_button_clicked(self, *extra):
        app_file = self.app_file_entry.gtk_entry().get_text()
        if app_file == "":
            glutil.WarningDialog(self,
                          "Please specify the filename for " +\
                          "the appearances file first. Often they have " +\
                          "a .app extension")
        elif not os.path.exists(app_file):
            glutil.WarningDialog(self, "The specified appearances file does not exist. " +\
                          "Please supply an existing appearances file. Often " +\
                          "they have a .app extension")
        else:
            glutil.FileViewWindow(app_file, "Appearances File",
                                 app_file, saveable=1, editable=1)


    def on_new_button_clicked(self, *extra):
        self.reset_defaults()
        self.trans_file_entry.gtk_entry().set_text("")
        self.app_file_entry.gtk_entry().set_text("")

    def on_reset_button_clicked(self, *extra):
        self.reset_defaults()

    def on_extra_combo_entry_changed(self, *extra):
        #
        # Reset parameter to default whenever change this
        #
        self.extra_value_spinbutton.set_value(10)
        if self.extra_combo.entry.get_text() == "":
            self.evaluation_output_checkbutton.set_active(gtk.FALSE)
            self.evaluation_output_checkbutton.set_sensitive(gtk.FALSE)
        else:
            self.evaluation_output_checkbutton.set_sensitive(gtk.TRUE)


    def reset_defaults(self):
        self.support_count_checkbutton.set_active(gtk.FALSE)
        self.rules_radiobutton.set_active(gtk.TRUE)
        self.min_items_spinbutton.set_value(1)
        self.max_items_spinbutton.set_value(5)
        self.support_spinbutton.set_value(10)
        self.confidence_spinbutton.set_value(80)
        self.extra_combo.entry.set_text("")
        self.descend_radiobutton.set_active(gtk.TRUE)
        self.output_format_entry.set_text("")
        self.scanable_checkbutton.set_active(gtk.FALSE)
        self.extended_support_checkbutton.set_active(gtk.FALSE)
        self.original_support_checkbutton.set_active(gtk.FALSE)
        self.memory_checkbutton.set_active(gtk.FALSE)
        self.load_transactions_checkbutton.set_active(gtk.TRUE)

        
    def on_save_button_clicked(self, *extra):
        #
        # This is a convenient place to try things out
        #
        return

    def on_trans_file_entry_activate(self, *extra):
        trans_entry =  self.trans_file_entry.gtk_entry()
        filename = trans_entry.get_text()
        (path,extension) = os.path.splitext(filename)
        if extension == "" \
           and not os.path.exists(path) \
           and os.path.exists(path + ".tab"):
            trans_entry.set_text(path + ".tab")

    def on_app_file_entry_activate(self, *extra):
        app_entry =  self.app_file_entry.gtk_entry()
        filename = app_entry.get_text()
        (path,extension) = os.path.splitext(filename)
        if extension == "" \
           and not os.path.exists(path) \
           and os.path.exists(path + ".app"):
            app_entry.set_text(path + ".app")

    #
    # dummies
    #
    def on_not_implemented(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_save1_activate(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_preferences1_activate(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_cut1_activate(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_open1_activate(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_paste1_activate(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_clear1_activate(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_new1_activate(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_save_as1_activate(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_properties1_activate(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_copy1_activate(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_new_button_clicked(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_open_button_clicked(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_save_button_clicked(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")

########################################################################
#
# AppStatusBar : Try figure it out
#
########################################################################

class AppStatusBar:

    def __init__(self, appbar):
        self.appbar = appbar
        self.statusmessages = []
        self.statuscount = []

    def add_status(self, status, timeout=4000, allow_duplicate=0):
        if not allow_duplicate:
            try:
                dup = self.statusmessages.index(status)
            except ValueError:
                pass
            else:
                self.statuscount[dup] += 1
                gtk.timeout_add(timeout, lambda x: self._remove_status(status), 0)
                return

        self.statusmessages.append(status)
        message = self._get_status_message()
        if len(self.statusmessages)==1:
            self.appbar.push(message)
        else:
            self.appbar.set_status(message)
        gtk.timeout_add(timeout, lambda x: self._remove_status(status), 0)
        self.statuscount.append(1)

    def _get_status_message(self):
        return "[%s]" % "] [".join(self.statusmessages)

    def _remove_status(self, status):
        i = self.statusmessages.index(status)
        if self.statuscount[i] == 1:
            self.statusmessages.pop(i)
            self.statuscount.pop(i)
            if len(self.statusmessages)==0:
                self.appbar.pop()
            else:
                message = self._get_status_message()
                self.appbar.set_status(message)
        else:
            self.statuscount[i] -= 1

########################################################################
#
# Main
#
########################################################################

def main():
    if len(sys.argv) == 1:
        AprioriApp().mainloop()
    elif len(sys.argv) == 2:
        AprioriApp(sys.argv[1]).mainloop()

if __name__ == "__main__":
    main()
