#! /usr/bin/env python
#
# Gnome Data Miner: GNOME interface to Borgelt's dtree
#
# Copyright (c) Graham J Williams
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version. See the file gpl-license.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# TODO
#
#	Experiment with the new logging package
#
#	Complete File Options
#
#	Open => prompt for a project to load
#	Save => Collect set up and save as a project
#	Save As => Collect set up and save as a project
#	Edit => Remove copy, cut, paste clear.
#	Edit -> Properties => Location of programs
#	Edit -> Preferences => ?
#
#	Introduce concept of project: saves file names and all options.
#
#	How to add tooltips to Measure items?
#
#	Why don't the tooltips on the toolbar work?
#
#	There are too many literals throughout.  Name the constants (like
#	default classificaiton measure) and use them instead.
#
#	Plenty of opportunity to tidy up the code.
#
# Style Guide (see http://www.python.org/doc/essays/styleguide.html)
#
#    Use the "_" convention for function/variable naming (e.g., main_window).
#	Consistent with glade naming. No conflict with Class/Module names.
#    Use Python doc strings to comment each method (triple quotes).
#       Refer to http://www.python.org/peps/pep-0257.html
#    Document functions in docstrings, not in comments, so they are available.
#	Automated tools can extract the docstrings, but not the comments.
#    Use same names in Python as for the Glade objects.
#	No need to remember two names for everything.
#    Classes and Modules generally use the CapWords naming style.
#
# Authors: Graham Williams
#

########################################################################
#
# Libraries
#
import os
import sys
import popen2
import tempfile
try:
    import gtk
    import gtk.glade
except ImportError, e:
    print e
    print "Please install python-glade2 and python-gnome2."
    sys.exit(1)
import glutil
########################################################################
#
# Global Constants
#
########################################################################

from const import version

weights_map = {
    "Lower"			: "l",
    "Boost"			: "b",
    "Shift Tuple Weights"	: "s"
    }

prune_methods = {
    "Confidence Level"		: "clvl",
    "Pessimistic"		: "pess",
    "None"			: "none"
    }

classification_map = [
    ("Information Gain Ratio", "infgr",
     "This is the default method. It is commonly used in the " +\
     "machine learning community, while the Gini index is often " +\
     " used by the statistical community."),
    ("Information Gain", "infgain"),
    ("Balanced Info Gain", "infgbal"),
    ("Symmetric Info Gain Ratio1", "infsgr1"),
    ("Symmetric Info Gain Ratio 2", "infsgr2"),
    ("Quadratic Info Gain", "qigain"),
    ("Balanced Quadratic Info Gain", "qigbal"),
    ("Quadratic Info Gain Ratio", "qigr"),
    ("Symmetric Quadratic Information Gain Ratio 1" , "qisgr1"),
    ("Symmetric Quadratic Information Gain Ratio 2" , "qisgr2"),
    ("Gini Index" , "gini"),
    ("Symmetric Gini Index" , "ginisym"),
    ("Modified Gini Index" , "ginimod"),
    ("Relief Measure" , "relief"),
    ("Sum Of Weighted Differences" , "wdiff"),
    ("Chi^2 Measure" , "chi2"),
    ("Normalized Chi^2 Measure" , "chi2nrm"),
    ("Weight Of Evidence" , "wevid"),
    ("Relevance" , "relev"),
    ("Bayesian-Dirichlet / K2 Metric" , "bdm"),
    ("Modified Bayesian-Dirichlet / K2 Metric" , "bdmod"),
    ("Reduction Of Description Length (rel. freq.)" , "rdlrel"),
    ("Reduction Of Description Length (abs. freq.)" , "rdlabs"),
    ("Stochastic Complexity" , "stoco"),
    ("Specificity Gain" , "spcgain"),
    ("Balanced Specificity Gain" , "spcgbal"),
    ("Specificity Gain Ratio" , "spcgr"),
    ("Symmetric Specificity Gain Ratio 1" , "spcsgr1"),
    ("Symmetric Specificity Gain Ratio 2" , "spcsgr2"),
    ("No measure", "none",
     "Don't use any measure!"),
    ]

regression_map = [
    ("Reduction Of Square Root Of Mean Squared Error", "rmse",
     ""),
    ("Reduction Of Sum Of Squared Errors", "sse",
     ""),
    ("Reduction Of Mean Squared Error", "mse",
     ""),
    ("Reduction Of Variance (Unbiased Estimator)", "var",
     ""),
    ("Reduction Of Standard Deviation (From Variance)", "sd",
     ""),
    ("No Measure", "none",
     ""),
  ]


########################################################################
#
# Decision Tree App
#
########################################################################

class DecisionTreeApp(glutil.GnomeApp):

    def __init__(self, dom="", tab=""):
        glutil.GnomeApp.__init__(self,
                                 "gdmdtree",
                                 "Gnome Data Miner: Decision Tree",
                                 version,
                                 glutil.appdir("gdmdtree.glade"),
                                 "dtree_app")
        
        self.default_data_ext  = "tab"
        self.default_names_ext = "dom"

        #
        # Temporary Files
        #
        self.dt_file = tempfile.mktemp(".dt")
        self.dtp_file = tempfile.mktemp(".dtp")
        self.rs_file = tempfile.mktemp(".rs")
        self.log_file = tempfile.mktemp(".gdmdtree-log")

        if tab != "":
            self.dataset_file_entry.gtk_entry().set_text(tab)
            self.dataset_file_entry.gnome_entry().append_history(1, tab)

        if dom != "":
            self.description_file_entry.gtk_entry().set_text(dom)
            self.description_file_entry.gnome_entry().append_history(1, dom)

        self.set_measure_classification()

    def set_measure_classification(self):
        item_list = []
        for k in classification_map:
            item_list.append(k[0])
        self.selection_combo.set_popdown_strings(item_list)

    def set_measure_regression(self):
        item_list = []
        for k in regression_map:
            item_list.append(k[0])
        self.selection_combo.set_popdown_strings(item_list)

    #
    # Command line generation and execution
    #

    def get_dataset_options(self):
        command = ""
        #
        # Characters to be treated as blank
        #
        blanks = self.blank_entry.get_text()
        if blanks != "": command += "-b\"" + blanks + "\" "
        #
        # Use default header (column count)
        #
        if self.number_columns_checkbutton.get_active(): command += "-d "
        #
        # Dataset field separator
        #
        field_separator = self.field_separator_entry.get_text()
        if field_separator != "": command += "-f\"" + field_separator + "\" "
        #
        # Include a record count
        #
        if self.record_count_checkbutton.get_active(): command += "-n "
        #
        # Dataset record separator
        #
        record_separator = self.record_separator_entry.get_text()
        if record_separator != "": command += "-r\"" + record_separator + "\" "
        #
        # Unknown value character
        #
        unknown = self.unknown_entry.get_text()
        if unknown != "": command += "-u\"" + unknown + "\" "
        #
        # All done
        #
        return(command)
        

    def execute_dti(self):
        #
        # What do we generate: Tree, Rules, Both
        #
        generate_tree  = self.tree_checkbutton.get_active()
        prune_tree = self.prune_checkbutton.get_active()
        generate_rules = self.rules_checkbutton.get_active()
        #
        # Relevant files
        #
        data_file = self.dataset_file_entry.gtk_entry().get_text()
        desc_file = self.description_file_entry.gtk_entry().get_text()

        if not (generate_tree or prune_tree or generate_rules):
            glutil.WarningDialog(self,
                                "Nothing to do. Enable at least one of " +\
                                "Display Decision Tree, Tree Pruning, " +\
                                "and Generate Rule Set")
        elif data_file == "":
            glutil.WarningDialog(self,
                          "Please enter the name of a dataset file first.")
        elif desc_file == "":
            glutil.WarningDialog(self,
                          "Please enter the name of a description file first.")
        elif not os.path.exists(data_file):
            glutil.ErrorDialog(self,
                        "Dataset file could not be found: " + data_file)
        elif not os.path.exists(desc_file):
            glutil.ErrorDialog(self,
                        "Description file could not be found: " + desc_file)
        else:
            #
            # BUILD THE COMMAND
            #
            command = "dti "
            #
            # Add dataset options
            #
            command += self.get_dataset_options()
            #
            # Align the output
            #
            if self.tree_align_checkbutton.get_active(): command += "-a "
            #
            # Specify the target column.  If blank the the default
            # last column is used.
            #
            target_column = self.target_column_entry.get_text()
            if target_column != "": command += "-c" + target_column + " "
            #
            # Determine whether a non-default measure is used.
            #
            is_classification = self.classification_radiobutton.get_active()
            is_regression = self.regression_radiobutton.get_active()
            full_measure = self.selection_combo.entry.get_text()
            if is_classification:
                measure = filter(lambda x: x[0] == full_measure,
                                 classification_map)[0][1]
                if measure != "infgr":
                    command += "-e" + measure + " "
            elif is_regression:
                measure = filter(lambda x: x[0] == full_measure,
                                 regression_map)[0][1]
                if measure != "rmse":
                    command += "-e" + measure + " "
            #
            # Minimum value allowed for the feature selection measure.
            # A zero means to ignore it - no limit.
            #
            min_measure = self.tree_minmeasure_spinbutton.get_value()
            if min_measure != 0: command += "-i%f " % min_measure
            #
            # Maximum line length
            #
            command += "-l%d " % self.line_spinbutton.get_value()

            branch = self.tree_branch_spinbutton.get_value()
            command += "-m%d " % branch
            #
            #
            #
            if self.tree_balance_checkbutton.get_active():
                method = self.tree_balance_combo_entry.get_text()
                command += "-q" + weights_map[method] + " "
            if self.tree_subset_checkbutton.get_active(): command += "-s "
            #
            # Maximal tree height
            #
            levels = self.tree_levels_spinbutton.get_value()
            if levels > 0: command += "-t%d " % levels
            #
            # Output relative frequencies
            #
            if self.tree_frequency_checkbutton.get_active(): command += "-v "
            #
            # If we don't want to weight the feature selection measure then
            # we add -w.
            #
            if not self.tree_weights_checkbutton.get_active(): command += "-w "
            #
            # Add file names to command line
            #
            command += desc_file + " " + data_file + " " + self.dt_file
            #
            # Trace of command being executed
            #
            glutil.start_log(self.log_file, command)
            #
            # Redirect output to a file
            #
            command += " 2>> " + self.log_file
            #
            # RUN THE COMMAND
            #
            # print command
            os.system(command)
            glutil.end_log(self.log_file)
            #
            # Conditionally display the tree.
            #
            if generate_tree:
                glutil.FileViewWindow(self.dt_file,
                               "Decision Tree View",
                               os.path.splitext(data_file)[0] + ".dt")

            #
            # PRUNE THE TREE IF REQUESTED
            #
            if prune_tree:
                #
                # Build the command
                #
                command = "dtp "
                #
                # Add dataset options
                #
                command += self.get_dataset_options()
                #
                # Align the output
                #
                if self.tree_align_checkbutton.get_active(): command += "-a "
                #
                # Check largest branch.
                #
                if self.prune_largest_checkbutton.get_active():
                    command += "-k "
                #
                # Maximum line length
                #
                command += "-l%d " % self.line_spinbutton.get_value()
                #
                # Pruning method
                #
                method = self.prune_method_combo_entry.get_text()
                command += "-m" + prune_methods[method] + " "
                #
                # Pruning paramter
                #
                command += "-p%.2f " % self.prune_parameter_hscale.get_value()
                #
                # Balance class frequencies (weight tuples)
                #
                if self.tree_balance_checkbutton.get_active():
                    method = self.tree_balance_combo_entry.get_text()
                    command += "-q" + weights_map[method] + " "
                #
                # Maximal tree height
                #
                levels = self.tree_levels_spinbutton.get_value()
                if levels > 0: command += "-t%d " % levels
                #
                # Output relative frequencies
                #
                if self.tree_frequency_checkbutton.get_active():
                    command += "-v "
                #
                # Add the file arguments.
                #
                command += self.dt_file + " "
                command += self.dtp_file + " "
                #
                # Include the data file if requested.
                #
                if self.prune_data_checkbutton.get_active():
                    command += data_file + " "
                #
                # Trace of command being executed
                #
                glutil.start_log(self.log_file, command)
                #
                # Redirect output to the file
                #
                command += " 2>> " + self.log_file
                #
                # Run the command
                #
                os.system(command)
                glutil.end_log(self.log_file)
                #
                # Display the tree.
                #
                if not generate_rules:
                    glutil.FileViewWindow(self.dtp_file,
                                   "Pruned Decision Tree View",
                                   os.path.splitext(data_file)[0] + ".dtp")

            #
            # BUILD RULES IF REQUIRED
            #
            if generate_rules:
                command = "dtr "

                command += "-l%d " % self.line_spinbutton.get_value()

                if self.rules_support_checkbutton.get_active():
                    command += "-s "
                if self.rules_confidence_checkbutton.get_active():
                    command+="-c "
                if self.rules_onecond_checkbutton.get_active():
                    command+="-d "

                if prune_tree:
                    command += self.dtp_file + " "
                else:
                    command += self.dt_file + " "
                command += self.rs_file + " "
                #
                # Trace of command being executed
                #
                glutil.start_log(self.log_file, command)
                
                command += "2>> " + self.log_file
                os.system(command)
                glutil.end_log(self.log_file)
                #
                # Conditionally display the rules file.
                #
                glutil.FileViewWindow(self.rs_file,
                               "Rule Set View",
                               os.path.splitext(data_file)[0] + ".rs")

    def on_view_log_button_clicked(self, *extra):
        if not os.path.exists(self.log_file):
            glutil.WarningDialog(self,
                                "No log file yet.")
        else:
            glutil.FileViewWindow(self.log_file, "Log File", "", clearable=1)

    def cleanup(self):
        if os.path.isfile(self.dt_file): os.remove(self.dt_file)
        if os.path.isfile(self.rs_file): os.remove(self.rs_file)
        if os.path.isfile(self.log_file): os.remove(self.log_file)

    #
    # global
    #
    def on_help_about_activate(self, *extra):
        aboutxml = gtk.glade.XML(glutil.appdir("gdmdtree.glade"),"about")
        about = aboutxml.get_widget("about")
        about.set_property("name", "Gnome Data Miner: DTree")
        about.set_property("version", version)
        about.show()
    def on_quit_activate(self, *extra):
        self.cleanup()
        self.quit()
    def on_quit_activate(self, *extra):
        self.cleanup()
        self.quit()

    #
    # implementation
    #

    def on_execute_button_clicked(self, *extra):
        self.execute_dti()

    def on_generate_description_button_clicked(self, *extra):
        tabfile = self.dataset_file_entry.gtk_entry().get_text()
        domfile = self.description_file_entry.gtk_entry().get_text()
        if tabfile == "":
            glutil.WarningDialog(self,
                          "Please specify the filename for " +\
                          "the dataset file first. Often they have " +\
                          "a .tab extension")
        elif not os.path.exists(tabfile):
            glutil.WarningDialog(self, "The specified data file does not exist. " +\
                          "Please supply an existing data file. Often " +\
                          "they have a .tab extension")
        elif domfile == "":
            glutil.WarningDialog(self, "Please specify the filename for " +\
                          "writing the description to. Often this " +\
                          "will have an extension of .dom (for domain).")
        elif not os.path.exists(domfile) \
             or glutil.QuestionDialog(self, "This will overwrite " + domfile +\
                               ". Do you wish to continue?").get_answer():
            command = "dom "
            #
            # Add dataset options.
            #
            command += self.get_dataset_options()
            #
            # Automatic type determination.
            #
            command += "-a "
            #
            # Add the data files.
            #
            command += tabfile + " "
            command += domfile + " "
            #
            # Trace of command being executed
            #
            glutil.start_log(self.log_file, command)
            command += " 2>> " + self.log_file
            os.system(command)
            glutil.end_log(self.log_file)
            glutil.FileViewWindow(domfile,
                           "Domain File",
                           domfile, editable=1)

    #
    # Buttons
    #

    def on_view_dataset_button_clicked(self, *extra):
        tabfile = self.dataset_file_entry.gtk_entry().get_text()
        if tabfile == "":
            glutil.WarningDialog(self,
                          "Please specify the filename for " +\
                          "the dataset file first. Often they have " +\
                          "a .tab extension")
        elif not os.path.exists(tabfile):
            glutil.WarningDialog(self, "The specified data file does not exist. " +\
                          "Please supply an existing data file. Often " +\
                          "they have a .tab extension")
        else:
            glutil.FileViewWindow(tabfile, "Dataset File", tabfile, saveable=0)

        
    def on_view_description_button_clicked(self, *extra):
        domfile = self.description_file_entry.gtk_entry().get_text()
        if domfile == "":
            glutil.WarningDialog(self,
                          "Please specify the filename for " +\
                          "the description file first. Often they have " +\
                          "a .dom extension")
        elif not os.path.exists(domfile):
            glutil.WarningDialog(self, "The specified description file does not exist. " +\
                          "Please supply an existing description file. Often " +\
                          "they have a .dom extension")
        else:
            glutil.FileViewWindow(domfile, "Description File", domfile, saveable=0)


    def on_new_button_clicked(self, *extra):
        self.reset_defaults()
        self.dataset_file_entry.gtk_entry().set_text("")
        self.description_file_entry.gtk_entry().set_text("")

    def on_reset_button_clicked(self, *extra):
        self.reset_defaults()

    def reset_defaults(self):
        self.line_spinbutton.set_value(70)
        
        self.tree_checkbutton.set_active(gtk.TRUE)
        self.tree_align_checkbutton.set_active(gtk.TRUE)
        self.tree_subset_checkbutton.set_active(gtk.FALSE)
        self.tree_frequency_checkbutton.set_active(gtk.FALSE)
        self.tree_balance_checkbutton.set_active(gtk.FALSE)
        self.tree_balance_combo.set_sensitive(gtk.FALSE)
        self.tree_balance_combo_entry.set_text("Lower")
        self.tree_levels_spinbutton.set_value(0)
        self.tree_branch_spinbutton.set_value(2)

        self.classification_radiobutton.set_active(gtk.TRUE)
        self.selection_combo.entry.set_text(classification_map[0][0])
        self.tree_weights_checkbutton.set_active(gtk.TRUE)
        self.tree_minmeasure_spinbutton.set_value(0)

        self.prune_checkbutton.set_active(gtk.FALSE)
        self.prune_method_combo_entry.set_text("Confidence Level")
        self.prune_parameter_hscale.set_value(0.5)
        self.prune_data_checkbutton.set_active(gtk.FALSE)
        self.prune_largest_checkbutton.set_active(gtk.FALSE)

        self.rules_checkbutton.set_active(gtk.FALSE)
        self.rules_support_checkbutton.set_active(gtk.TRUE)
        self.rules_confidence_checkbutton.set_active(gtk.TRUE)
        self.rules_onecond_checkbutton.set_active(gtk.TRUE)
        
        self.target_column_entry.set_text("")

    def on_save_button_clicked(self, *extra):
        #
        # This is a convenient place to try things out
        #
        return
        #glutil.FileViewWindow("drug.dom", "Drug Data File")
        dlg=gtk.MessageDialog(self, gtk.DIALOG_DESTROY_WITH_PARENT,
                              gtk.MESSAGE_WARNING,
                              gtk.BUTTONS_OK,
                              "Please specify a dataset file name before executing")
        
        dlg.run()
        dlg.destroy()
        #MessageDialog("Please supply the name of a data file")
        #self.statusbar.add_status("Hi THere")

    def on_dataset_file_entry_activate(self, *extra):
        dataset_entry =  self.dataset_file_entry.gtk_entry()
        description_entry =  self.description_file_entry.gtk_entry()
        filename = dataset_entry.get_text()
        description = description_entry.get_text()
        (path,extension) = os.path.splitext(filename)
        if extension == "" \
           and not os.path.exists(path) \
           and os.path.exists(path + ".tab"):
            dataset_entry.set_text(path + ".tab")
            description_entry.set_text(path + ".dom")

    #
    # Decision Tree Options
    #
    def on_tree_balance_checkbutton_toggled(self, *extra):
        is_active = self.tree_balance_checkbutton.get_active()
        self.tree_balance_combo.set_sensitive(is_active)

    def on_classification_radiobutton_toggled(self, *extra):
        is_classification = self.classification_radiobutton.get_active()
        is_regression = self.regression_radiobutton.get_active()
        if is_classification:
            self.set_measure_classification()
        elif is_regression:
            self.set_measure_regression()


    #
    # Rule Options
    #
    def on_rules_checkbutton_toggled(self, *extra):
        is_active = self.rules_checkbutton.get_active()
        self.rules_support_checkbutton.set_sensitive(is_active)
        self.rules_confidence_checkbutton.set_sensitive(is_active)
        self.rules_onecond_checkbutton.set_sensitive(is_active)

    #
    # Prune Options
    #
    def on_prune_checkbutton_toggled(self, *extra):
        is_active = self.prune_checkbutton.get_active()
        self.prune_method_combo.set_sensitive(is_active)
        self.prune_method_label.set_sensitive(is_active)
        self.prune_parameter_hscale.set_sensitive(is_active)
        self.prune_parameter_label.set_sensitive(is_active)
        self.prune_largest_checkbutton.set_sensitive(is_active)
        self.prune_data_checkbutton.set_sensitive(is_active)

    #
    # dummies
    #
    def on_not_implemented(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_save1_activate(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_preferences1_activate(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_cut1_activate(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_open1_activate(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_paste1_activate(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_clear1_activate(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_new1_activate(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_save_as1_activate(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_properties1_activate(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_copy1_activate(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_new_button_clicked(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_open_button_clicked(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")
    def on_save_button_clicked(self, *extra):
        glutil.WarningDialog(self, "This function is not yet implemented.")

########################################################################
#
# AppStatusBar : Try figure it out
#
########################################################################

class AppStatusBar:

    def __init__(self, appbar):
        self.appbar = appbar
        self.statusmessages = []
        self.statuscount = []

    def add_status(self, status, timeout=4000, allow_duplicate=0):
        if not allow_duplicate:
            try:
                dup = self.statusmessages.index(status)
            except ValueError:
                pass
            else:
                self.statuscount[dup] += 1
                gtk.timeout_add(timeout, lambda x: self._remove_status(status), 0)
                return

        self.statusmessages.append(status)
        message = self._get_status_message()
        if len(self.statusmessages)==1:
            self.appbar.push(message)
        else:
            self.appbar.set_status(message)
        gtk.timeout_add(timeout, lambda x: self._remove_status(status), 0)
        self.statuscount.append(1)

    def _get_status_message(self):
        return "[%s]" % "] [".join(self.statusmessages)

    def _remove_status(self, status):
        i = self.statusmessages.index(status)
        if self.statuscount[i] == 1:
            self.statusmessages.pop(i)
            self.statuscount.pop(i)
            if len(self.statusmessages)==0:
                self.appbar.pop()
            else:
                message = self._get_status_message()
                self.appbar.set_status(message)
        else:
            self.statuscount[i] -= 1

########################################################################
#
# Main
#
########################################################################

def main():
    if len(sys.argv) == 1:
        DecisionTreeApp().mainloop()
    elif len(sys.argv) == 2:
        DecisionTreeApp(sys.argv[1] + ".dom", sys.argv[1] + ".tab").mainloop()

if __name__ == "__main__":
    main()
